#include "network-application.h"
#include "network-scheduler.h"

#include <sys/select.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <set>


void GetApplicationListWithDataReady(set<int>& set_data_ready, NetworkApplication *applications, int robot_num)
{
  static int ready_sockets;
  static fd_set readfds;
  static struct timeval no_timeout = {0, 0};
  static int nfds = (applications[robot_num-1].GetSocket()).sock + 1; // highest numbered socket fd + 1
  
  FD_ZERO(&readfds);
  for(int i=0; i < robot_num; i++)
    FD_SET((applications[i].GetSocket()).sock, &readfds);
  
  ready_sockets = select(nfds, &readfds, NULL, NULL, &no_timeout);
  
  if( set_data_ready.size() > 0 )
    set_data_ready.clear();

  for(int i=0; (i < robot_num) && (ready_sockets > 0); i++)
    {
      if( FD_ISSET((applications[i].GetSocket()).sock, &readfds) )
	{
	  ready_sockets--;
	  set_data_ready.insert(i);
	}
    }
}


int main(int argc, char *argv[])
{
   if (argc != 2) 
     {
       fprintf(stderr, "Usage: %s  <port_base>\n", argv[0]);
       exit(1);
     }

   /* Just to be sure that there's no effective limit on the number of open sockets */
   /*
   struct rlimit rlim;
      getrlimit(RLIMIT_FSIZE, &rlim);
   fprintf(stderr, "MAX OPEN SOCKETS: %lu - %lu\n", rlim.rlim_cur, rlim.rlim_max);
   */

   srand(12345);

   int port_base = atoi(argv[1]);

   int step = 0;

   // Create the scheduler object
   NetworkScheduler scheduler(port_base);
   
   // Open the socket with the robot simulation scheduler
   scheduler.OpenSocketWithRobotSimScheduler();

   // Get simulation and client information from the robot simulator
   scheduler.GetSimulationInformation();

   int robot_num = scheduler.GetRobotNum();
   Real simulation_step = scheduler.GetSimulationTimeStep();
   Real current_time_slice = scheduler.GetCurrentTimeSlice();

   fprintf(stderr, "Received basic information from robot simulator %s: %d robots, %g simulation step\n", 
	   scheduler.GetRobotSimulatorName(), scheduler.GetRobotNum(), scheduler.GetSimulationTimeStep());
   
   // start the applications and their connection sockets to the individual robot processes
   NetworkApplication *applications = new NetworkApplication[robot_num];
   for(int i=0; i < robot_num; i++)
     applications[i].InitSocketWithRobotApplication(i, port_base, MAX_MSG_SIZE);


   // Send back to the robot simulator an acknowledgement packet (with some basic info) to sync data exchange
   scheduler.SendNetworkSimInfo("NS3");

   // Get basic info from the robot apps, this is needed to get the addresses of the clients
   for(int i=0; i < robot_num; i++)
       applications[i].GetStartInfoFromRobotAppl();

   set<int> set_data_ready;

   while(true)
     {
       // Get positions and time updates from the robot simulator and send an ack back: these communications are blocking
       scheduler.SendNetworkAck(step);
       scheduler.GetNodePositionsUpdate();
       fprintf(stderr,"network time %g vs. robot time %g\n",current_time_slice,scheduler.GetCurrentTimeSlice());

       // Use the gathered info to update nodes' positions
       scheduler.UpdateNodePositions();

       // this is provided as an example: this function can be used to detect which subset of the nodes have data to be
       // read, in this way it would be pssible to avoid to loop over all nodes and/or generate reading events for all nodes
       /* Decomment to see how it works
       GetApplicationListWithDataReady(set_data_ready, applications, robot_num);

       set<int>::iterator i;
       for(i=set_data_ready.begin(); i != set_data_ready.end(); ++i) 
	 fprintf(stderr, "Application[%d] has data ready\n", *i);
       */

       // Actions at the application layers of the nodes
       vector< vector<string> > vecMsgs(robot_num);
       for(int i=0; i < robot_num; i++)
	 {
	   // Each application check for input data from the robot applications
	   vecMsgs[i] = applications[i].GetNewInputDataFromRobotApp(current_time_slice,robot_num);
	 }

       // Simulate that messages are being sent through the network and given back to the robots
       for(int i=0; i < robot_num; i++)
       {
          for(int j=0; j < robot_num; j++)
          {
	     // Each application sends to the destination robot the data received through NS3 communications
             if (vecMsgs[i][j].size() != 0)
             {
		printf("sending back to robot %d %d\n",i,j);
                applications[j].SendReceivedDataToRobotApp(current_time_slice+(simulation_step/2.0), i, vecMsgs[i][j]);
             }
          }
       }

       current_time_slice += simulation_step;
     }
}
