/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2010 UCM
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Author:  Delfín Rupérez Cañas <delfinrc@fdi.ucm.es>
 *          
 */
#include "antor-dlr-core.h"
//#include "ns3/cbr-client.h"
NS_LOG_COMPONENT_DEFINE ("AntOR-DLR");

namespace ns3 {

namespace antordlr {

NS_OBJECT_ENSURE_REGISTERED (AntORDLRCore);

/// Maximum allowed sequence number.
#define  ANTOR_DLR_MAX_SEQ_NUM	65535

#define ANYADDRESS Ipv4Address::GetAny()

#define ANTOR_RDLR_ALLOWED_SENT 5


///
/// \brief Gets the delay between a given time and the current time.
///
/// If given time is previous to the current one, then this macro returns
/// a number close to 0. This is used for scheduling events at a certain moment.
///
#define DELAY(time) ( (time < Simulator::Now () ) ? Seconds (0.000001) : \
                   (time - Simulator::Now () + Seconds (0.000001) ) )

///
/// \brief Gets the alternative delay between a given time and the current time

#define DELAY_ALTERNATIVE(timeNew , timeOld) timeNew - timeOld

/// UDP Port for AntOR-DLR control traffic
const uint32_t AntORDLRCore::ANTOR_DLR_PORT = 60001;

//-----------------------------------------------------------------------------
/// Tag used by AntOR DLR implementation
struct DeferredRouteOutputTag: public Tag {
	/// Positive if output device is fixed in RouteOutput
	int32_t oif;

	DeferredRouteOutputTag(int32_t o = -1) :
		Tag(), oif(o) {
	}

	static TypeId GetTypeId() {
		static TypeId
				tid =
						TypeId("ns3::antordlr::DeferredRouteOutputTag").SetParent<
								Tag> ();
		return tid;
	}

	TypeId GetInstanceTypeId() const {
		return GetTypeId();
	}

	uint32_t GetSerializedSize() const {
		return sizeof(int32_t);
	}

	void Serialize(TagBuffer i) const {
		i.WriteU32(oif);

	}

	void Deserialize(TagBuffer i) {
		oif = i.ReadU32();

	}

	void Print(std::ostream &os) const {
		os << "DeferredRouteOutputTag: output interface = " << oif;
	}
};

TypeId AntORDLRCore::GetTypeId(void) {
	static TypeId
			tid =
					TypeId("ns3::antordlr::AntORDLRCore") .SetParent<Ipv4RoutingProtocol> ()
							.AddConstructor<AntORDLRCore> ()
							.AddAttribute("HelloInterval","HELLO messages emission interval.",
									       TimeValue(Seconds(1)),
									       MakeTimeAccessor(&AntORDLRCore::m_helloInterval),
							MakeTimeChecker()) .AddAttribute("PFAInterval",
							"Proactive Forward Ant emission interval.",
							TimeValue(Seconds(2)), MakeTimeAccessor(
									&AntORDLRCore::m_PFAInterval),
							MakeTimeChecker()) .AddAttribute("ApplicationPort",
							"Application Port Number", UintegerValue(0),
							MakeUintegerAccessor(
									&AntORDLRCore::m_Application_Port),
							MakeUintegerChecker<uint32_t> ())
							.AddAttribute(
							"EnableBroadcast",
							"Indicates whether a broadcast data packets forwarding enable.",
							BooleanValue(true), MakeBooleanAccessor(
									&AntORDLRCore::EnableBroadcast),
							MakeBooleanChecker())
							.AddAttribute(
							"NumReportDestination",
							"Maximum number of Report Destination.",
							UintegerValue(10), MakeUintegerAccessor(
									&AntORDLRCore::num_report_destination),
							MakeUintegerChecker<uint32_t> ())
							.AddAttribute(
							"MaxBroadcastAllowed",
							"Maximum number of Broadcast message allowed.",
							UintegerValue(4), MakeUintegerAccessor(
									&AntORDLRCore::max_broadcast_allowed),
							MakeUintegerChecker<uint32_t> ());

	return tid;
}

void AntORDLRCore::SetMaxQueueLen(uint32_t len) {
	MaxQueueLen = len;
	m_queue.SetMaxQueueLen(len);
}
void AntORDLRCore::SetMaxQueueTime(Time t) {
	MaxQueueTime = t;
	m_queue.SetQueueTimeout(t);
}

AntORDLRCore::AntORDLRCore() :

			 NetDiameter(35), NodeTraversalTime(
					MilliSeconds(40)), NetTraversalTime(Time((2 * NetDiameter)
					* NodeTraversalTime)), PathDiscoveryTime(Time(2
					* NetTraversalTime)), m_helloInterval(
					Seconds(1)), AllowedHelloLoss(2), MaxQueueLen(64), MaxQueueTime(
					Seconds(30)), m_queue(MaxQueueLen,
					MaxQueueTime), m_rreqIdCache(
					PathDiscoveryTime), m_dpd(PathDiscoveryTime), m_nbs(
					m_helloInterval),
					m_helloTimer(Timer::CANCEL_ON_DESTROY),
					m_PFATimer(Timer::CANCEL_ON_DESTROY),
				    m_RRFATimer (Timer::CANCEL_ON_DESTROY),
			m_RFATimer(Timer::CANCEL_ON_DESTROY)
{
	//FIXME
	//LogComponentEnable("AntOR-DLR", LOG_LEVEL_ALL);
	//LogComponentEnable("Ipv4L3Protocol", LOG_LEVEL_ALL);
	m_nbs.SetCallback(MakeCallback(
			&AntORDLRCore::SendMessageWhenBreaksLinkToNextHop, this));
}

AntORDLRCore::~AntORDLRCore() {

}

void AntORDLRCore::SetIpv4(Ptr<Ipv4> ipv4) {
	NS_ASSERT(ipv4 != 0);
	//m_LFNTimer.SetFunction(&AntORDLRCore::LFNTimerExpire, this);
	m_RFATimer.SetFunction(&AntORDLRCore::RFATimerExpire, this);
	m_RRFATimer.SetFunction(&AntORDLRCore::RRFATimerExpire, this);
	m_PFATimer.SetFunction(&AntORDLRCore::PFATimerExpire, this);
	//m_helloTimer.SetFunction(&AntORDLRCore::HelloTimerExpire, this);
	//m_NeighborTimer.SetFunction(&AntORDLRCore::NeighborTimerExpire, this);
	m_messageSequenceNumber = ANTOR_DLR_MAX_SEQ_NUM;
	m_ipv4 = ipv4;
	// Create lo route. It is asserted that the only one interface up for now is loopback
	NS_ASSERT (m_ipv4->GetNInterfaces () == 1 && m_ipv4->GetAddress (0, 0).GetLocal () == Ipv4Address ("127.0.0.1"));
	m_lo = m_ipv4->GetNetDevice(0);
	NS_ASSERT (m_lo != 0);

	// Remember lo route
	RoutingTableEntry rt(/*device=*/m_lo, /*dst=*/Ipv4Address::GetLoopback(), /*know seqno=*/
	/*iface=*/Ipv4InterfaceAddress(Ipv4Address::GetLoopback(), Ipv4Mask(
			"255.0.0.0")),
	/*hops=*/1, /*next hop=*/Ipv4Address::GetLoopback());

	m_table.AddRoute(rt);

	// Start routing protocol
	Simulator::ScheduleNow(&AntORDLRCore::Start, this);

}

void AntORDLRCore::DoDispose() {
	/*
	 m_file.write(GetObject<Node> ()->GetId() + 1,
	 "    , Num   Hello Sent          ", m_stat.getHelloSent());
	 m_file.write(GetObject<Node> ()->GetId() + 1,
	 "    , Total Bytes Hello Sent    ", m_stat.getByteHelloSent());
	 m_file.write(GetObject<Node> ()->GetId() + 1,
	 "    , Num   Hello Received      ", m_stat.getHelloReceived());
	 m_file.write(GetObject<Node> ()->GetId() + 1,
	 "    , Total Bytes Hello Sent    ", m_stat.getByteHelloReceived());

	 if (m_stat.getRFASent() > 0) {
	 m_file.write(GetObject<Node> ()-> GetId() + 1,
	 "    , Total RFA Sent            ", m_stat.getRFASent());
	 m_file.write(GetObject<Node> ()->GetId() + 1,
	 "    , Total Bytes RFA Sent      ", m_stat.getBytesRFASent());
	 }
	 if (m_stat.getRFAReceived() > 0) {
	 m_file.write(GetObject<Node> ()-> GetId() + 1,
	 "    , Total RFA Received        ", m_stat.getRFAReceived());
	 m_file.write(GetObject<Node> ()->GetId() + 1,
	 "    , Total Bytes RFA Received  ", m_stat.getBytesRFAReceived());
	 }
	 if (m_stat.getRBASent() > 0) {
	 m_file.write(GetObject<Node> ()-> GetId() + 1,
	 "    , Total RBA Sent            ", m_stat.getRBASent());
	 m_file.write(GetObject<Node> ()->GetId() + 1,
	 "    , Total Bytes RBA Sent      ", m_stat.getBytesRBASent());
	 }
	 if (m_stat.getRBAReceived() > 0) {
	 m_file.write(GetObject<Node> ()-> GetId() + 1,
	 "    , Total RBA Received        ", m_stat.getRBAReceived());
	 m_file.write(GetObject<Node> ()->GetId() + 1,
	 "    , Total Bytes RBA Received  ", m_stat.getBytesRBAReceived());
	 }
	 if (m_stat.getPFASent() > 0) {
	 m_file.write(GetObject<Node> ()-> GetId() + 1,
	 "    , Total PFA Sent            ", m_stat.getPFASent());
	 m_file.write(GetObject<Node> ()->GetId() + 1,
	 "    , Total Bytes PFA Sent      ", m_stat.getBytesPFASent());
	 }
	 if (m_stat.getPFAReceived() > 0) {
	 m_file.write(GetObject<Node> ()-> GetId() + 1,
	 "    , Total PFA Received        ", m_stat.getPFAReceived());
	 m_file.write(GetObject<Node> ()->GetId() + 1,
	 "    , Total Bytes PFA Received  ", m_stat.getBytesPFAReceived());
	 }
	 if (m_stat.getPBASent() > 0) {
	 m_file.write(GetObject<Node> ()-> GetId() + 1,
	 "    , Total PBA Sent            ", m_stat.getPBASent());
	 m_file.write(GetObject<Node> ()->GetId() + 1,
	 "    , Total Bytes PBA Sent      ", m_stat.getBytesPBASent());
	 }
	 if (m_stat.getPBAReceived() > 0) {
	 m_file.write(GetObject<Node> ()-> GetId() + 1,
	 "    , Total RBA Received        ", m_stat.getPBAReceived());
	 m_file.write(GetObject<Node> ()->GetId() + 1,
	 "    , Total Bytes RBA Received  ", m_stat.getBytesPBAReceived());
	 }
	 if (m_stat.getLFNSent() > 0) {
	 m_file.write(GetObject<Node> ()-> GetId() + 1,
	 "    , Total LFN Sent            ", m_stat.getLFNSent());
	 m_file.write(GetObject<Node> ()->GetId() + 1,
	 "    , Total Bytes LFN Sent      ", m_stat.getBytesLFNSent());
	 }
	 if (m_stat.getLFNReceived() > 0) {
	 m_file.write(GetObject<Node> ()-> GetId() + 1,
	 "    , Total LFN Received        ", m_stat.getLFNReceived());
	 m_file.write(GetObject<Node> ()->GetId() + 1,
	 "    , Total Bytes LFN Received  ", m_stat.getBytesLFNReceived());
	 }
	 if (m_stat.getRRFASent() > 0) {
	 m_file.write(GetObject<Node> ()-> GetId() + 1,
	 "    , Total RRFA Sent            ", m_stat.getRRFASent());
	 m_file.write(GetObject<Node> ()->GetId() + 1,
	 "    , Total Bytes RRFA Sent      ", m_stat.getBytesRRFASent());
	 }
	 if ( m_stat.getRRFAReceived() > 0) {
	 m_file.write(GetObject<Node> ()-> GetId() + 1,
	 "    , Total RRFA Received        ", m_stat.getRRFAReceived());
	 m_file.write(GetObject<Node> ()->GetId() + 1,
	 "    , Total Bytes RRFA Received  ", m_stat.getBytesRRFAReceived());
	 }
	 if (m_stat.getRRBASent() > 0) {
	 m_file.write(GetObject<Node> ()-> GetId() + 1,
	 "    , Total RRBA Sent            ", m_stat.getRRBASent());
	 m_file.write(GetObject<Node> ()->GetId() + 1,
	 "    , Total Bytes RRBA Sent      ", m_stat.getBytesRRBASent());
	 }
	 if (m_stat.getRRBAReceived() > 0) {
	 m_file.write(GetObject<Node> ()-> GetId() + 1,
	 "    , Total RRBA Received        ", m_stat.getRRBAReceived());
	 m_file.write(GetObject<Node> ()->GetId() + 1,
	 "    , Total Bytes RRBA Received  ", m_stat.getBytesRRBAReceived());
	 }

	 m_file.write(GetObject<Node> ()->GetId() + 1,
	 "    , Num TOTAL PACKET SENT      ", m_stat.getPacketSent());
	 m_file.write(GetObject<Node> ()->GetId() + 1,
	 "    , TOTAL BYTES PACKET SENT    ", m_stat.getBytePacketSent());

	 m_file.write(GetObject<Node> ()->GetId() + 1,
	 "    , Num Neighbor              ", m_table.GetNeighbor().size());
	 if (m_mainAddress == m_table.FindSrcSession(m_mainAddress))
	 m_file.write(GetObject<Node> ()->GetId() + 1,
	 "    , Best Hop                  ", CalculateBestHop());
	 */
	m_ipv4 = 0;
	for (std::map<Ptr<Socket> , Ipv4InterfaceAddress>::iterator iter =
			m_socketAddresses.begin(); iter != m_socketAddresses.end(); iter++) {
		iter->first->Close();
	}
	m_socketAddresses.clear();

	Ipv4RoutingProtocol::DoDispose();

}

void AntORDLRCore::Start() {

	m_nbs.ScheduleTimer();

	m_helloTimer.SetFunction(&AntORDLRCore::HelloTimerExpire, this);
	Time t = Time(0.01 * MilliSeconds(UniformVariable().GetInteger(0, 100)));
	m_helloTimer.Schedule(t);

}

///
///  Sends a HELLO message and reschedules the HELLO timer.
///
///
void AntORDLRCore::HelloTimerExpire() {
	NS_LOG_FUNCTION (this);
	SendHello();
	m_helloTimer.Cancel();
	Time t = Time(0.01 * MilliSeconds(UniformVariable().GetInteger(0, 100)));
	m_helloTimer.Schedule(m_helloInterval - t);
}

void AntORDLRCore::RecvAntORDLR(Ptr<Socket> socket) {
	NS_LOG_FUNCTION (this << socket);
	Address sourceAddress;
	Ptr<Packet> receivedPacket = socket->RecvFrom(sourceAddress);
	InetSocketAddress inetSourceAddr = InetSocketAddress::ConvertFrom(
			sourceAddress);
	Ipv4Address sender = inetSourceAddr.GetIpv4();
	Ipv4Address receiver = m_socketAddresses[socket].GetLocal();
	Ipv4InterfaceAddress receiverIface = m_socketAddresses[socket];
	NS_LOG_INFO("AntOR-DLR node " << this << " received a AntOR-DLR packet from " << sender << " to " << receiver);
	// All routing messages are sent from and to port RT_PORT,
	// so we check it.
	NS_ASSERT(inetSourceAddr.GetPort() == ANTOR_DLR_PORT);
	Ptr<Packet> packet = receivedPacket->Copy();
	NS_LOG_DEBUG("AntOR-DLR node " << this << " received a AntOR-DLR packet from " << sender << " to " << receiver);
	ns3::antordlr::MessageHeader messageHeader;
	if (packet->RemoveHeader(messageHeader) == 0)
		NS_ASSERT(false);

	// If ttl is less than or equal to zero, or
	// the receiver is the same as the originator,
	// the message must be silently dropped
	if (messageHeader.GetTimeToLive() == 0) {
		return;
	}

	switch (messageHeader.GetMessageType()) {
	case ns3::antordlr::MessageHeader::HELLO_MESSAGE:
		NS_LOG_DEBUG(Simulator::Now().GetSeconds() << "s AntOR node "
				<< receiver << " received HELLO message of size "
				<< messageHeader.GetSerializedSize());

		ProcessHello(messageHeader, receiver, sender);

		break;
	case ns3::antordlr::MessageHeader::RFA_MESSAGE:
		NS_LOG_DEBUG(Simulator::Now().GetSeconds() << "s AntOR node "
				<< receiver << " received RFA message of size "
				<< messageHeader.GetSerializedSize());
		ProcessRFA(messageHeader, receiverIface, sender);
		break;
	case ns3::antordlr::MessageHeader::RBA_MESSAGE:
		NS_LOG_DEBUG(Simulator::Now().GetSeconds() << "s AntOR node "
				<< receiver << " received RBA message of size "
				<< messageHeader.GetSerializedSize());
		//std::cout << receiverIface.GetLocal() << " received RBA message \n";
		ProcessRBA(messageHeader, receiverIface, sender);
		break;
	case ns3::antordlr::MessageHeader::PFA_MESSAGE:
		NS_LOG_DEBUG(Simulator::Now().GetSeconds() << "s AntOR node "
				<< receiver << " received PFA message of size "
				<< messageHeader.GetSerializedSize());
		ProcessPFA(messageHeader, receiver, sender);
		break;
	case ns3::antordlr::MessageHeader::PBA_MESSAGE:
		NS_LOG_DEBUG(Simulator::Now().GetSeconds() << "s AntOR node "
				<< receiver << " received PBA message of size "
				<< messageHeader.GetSerializedSize());
		ProcessPBA(messageHeader, receiver, sender);
		break;
	case ns3::antordlr::MessageHeader::LFN_MESSAGE:
		NS_LOG_DEBUG(Simulator::Now().GetSeconds() << "s AntOR node "
				<< receiver << " received LFN message of size "
				<< messageHeader.GetSerializedSize());
		ProcessLFN(messageHeader, receiver, sender);
		break;
	case ns3::antordlr::MessageHeader::RRFA_MESSAGE:
		NS_LOG_DEBUG(Simulator::Now().GetSeconds() << "s AntOR node "
				<< receiver << " received RRFA message of size "
				<< messageHeader.GetSerializedSize());
		ProcessRRFA(messageHeader, receiverIface, sender);
		break;
	case ns3::antordlr::MessageHeader::RRBA_MESSAGE:
		NS_LOG_DEBUG(Simulator::Now().GetSeconds() << "s AntOR node "
				<< receiver << " received RRBA message of size "
				<< messageHeader.GetSerializedSize());
		ProcessRRBA(messageHeader, receiverIface, sender);
		break;
	case ns3::antordlr::MessageHeader::ULN_MESSAGE:
			NS_LOG_DEBUG(Simulator::Now().GetSeconds() << "s AntOR node "
					<< receiver << " received ULN message of size "
					<< messageHeader.GetSerializedSize());
			ProcessULN(messageHeader, receiver, sender);
			break;
	default:
		NS_LOG_DEBUG("AntOR message type "
				<< int(messageHeader.GetMessageType())
				<< " not implemented");
	}

}
void AntORDLRCore::SendULN ( Ipv4Address unDestination, Ipv4Address nextHop )

{
	    ns3::antordlr::MessageHeader msg;
		msg.SetMessageType(ns3::antordlr::MessageHeader::ULN_MESSAGE);
		msg.SetTimeToLive(255);
		msg.SetHopCount(0);
		msg.SetMessageSequenceNumber(GetMessageSequenceNumber());
		ns3::antordlr::MessageHeader::Uln uln = msg.GetUln();
		//FIXME
		Ptr<Ipv4L3Protocol> l3 = m_ipv4->GetObject<Ipv4L3Protocol> ();
		Ipv4InterfaceAddress iface = l3->GetAddress(1, 0);

		if (iface.GetLocal() == Ipv4Address("127.0.0.1"))
			return;
		uln.dst = unDestination;
		msg.GetSerializedSize();
		Ptr<Packet> packet = Create<Packet> ();
	    packet->AddHeader(msg);
	     Ptr<Socket> socket;
				socket = FindSocketWithInterfaceAddress(iface);
				int numBytes = socket->SendTo(packet, 0, InetSocketAddress(nextHop,
						ANTOR_DLR_PORT));
				if (numBytes != -1) {
	cout << "[ " << Simulator::Now().GetSeconds()
		 << " sg. ]  Sending first ULN message from interface "<<iface.GetLocal()<<" to "<<nextHop<<endl;
				}





}
void AntORDLRCore::SendLFN(vector<Ipv4Address> pre, vector<Ipv4Address> unDestination) {
/*
	NS_LOG_FUNCTION(this);

		if (unDestination.empty() || pre.empty()) {
			NS_LOG_LOGIC ("Neither unreachable destinations nor precesors");
			//cout <<" Neither unreachable destinations nor precesors\n";
			return;
		}
		ns3::antordlr::MessageHeader msg;
		msg.SetMessageType(ns3::antordlr::MessageHeader::LFN_MESSAGE);
		msg.SetTimeToLive(255);
		msg.SetHopCount(0);
		msg.SetMessageSequenceNumber(GetMessageSequenceNumber());
		ns3::antordlr::MessageHeader::Lfn &lfn = msg.GetLfn();
		//FIXME
		Ptr<Ipv4L3Protocol> l3 = m_ipv4->GetObject<Ipv4L3Protocol> ();
		Ipv4InterfaceAddress iface = l3->GetAddress(1, 0);

		if (iface.GetLocal() == Ipv4Address("127.0.0.1"))
			return;
		lfn.originator = iface.GetLocal();
		lfn.dest.clear();

		Ipv4Address destination;

	   for (uint32_t j = 0; j < unDestination.size(); j++) {
	   			 // cout << unDestination[j]<<" ";
	   		  	  //if (destination != unDestination[j]){
	   			   lfn.dest.push_back(unDestination[j]);
	   		  	  //}
	   		}//for
	   	  // cout <<"\n------\n";

		lfn.SetLdest(unDestination.size());
        Ipv4Address dir;
		uint32_t bytes = m_stat.getBytesLFNSent() + msg.GetSerializedSize();
		uint32_t num = m_stat.getLFNSent() + 1;
		m_stat.setBytesLFNSent(bytes);
		m_stat.setLFNSent(num);

	    Ptr<Packet> packet = Create<Packet> ();
	    packet->AddHeader(msg);
	     Ptr<Socket> socket;
				socket = FindSocketWithInterfaceAddress(iface);
				for  (uint32_t h=0; h < pre.size(); h++){
					dir = pre[h];
				  std::vector<RoutingTableEntry> rts;
						bool result = m_table.LookupAllRoutes(dir, rts);
						if (result) {
							dir = Ipv4Address("255.255.255.255");
							int numBytes = socket->SendTo(packet, 0, InetSocketAddress(dir,
									ANTOR_DLR_PORT));
							if (numBytes != -1) {

							//	cout << iface.GetLocal() << " sending first packet to " << dir
							//			<< " of bytes " << numBytes << endl;
							}
							break;
						}
										}

				//socket->SendTo(packet, 0, InetSocketAddress(Ipv4Address("255.255.255.255"),
				//									ANTOR_DLR_PORT));

*/
	NS_LOG_FUNCTION(this);

		if (unDestination.empty() || pre.empty()) {
			NS_LOG_LOGIC ("Neither unreachable destinations nor precesors");
			return;
		}
		ns3::antordlr::MessageHeader msg;
		msg.SetMessageType(ns3::antordlr::MessageHeader::LFN_MESSAGE);
		msg.SetTimeToLive(255);
		msg.SetHopCount(0);
		msg.SetMessageSequenceNumber(GetMessageSequenceNumber());
		ns3::antordlr::MessageHeader::Lfn &lfn = msg.GetLfn();
		//FIXME
		Ptr<Ipv4L3Protocol> l3 = m_ipv4->GetObject<Ipv4L3Protocol> ();
		Ipv4InterfaceAddress iface = l3->GetAddress(1, 0);

		if (iface.GetLocal() == Ipv4Address("127.0.0.1"))
			return;
		lfn.originator = unDestination[0];
		lfn.dest.clear();

		Ipv4Address destination;
			 		if (iface.GetMask() == Ipv4Mask::GetOnes()) {
			 			destination = Ipv4Address("255.255.255.255");
			 		} else {
			 			destination = iface.GetBroadcast();
			 		}
		for (uint32_t j = 0; j < pre.size(); j++) {
			  //cout << unDestination[j]<<" ";
		  	  //if (destination != unDestination[j]){
			   lfn.dest.push_back(pre[j]);
		  	  //}
		}//for
		//cout <<"\n------\n";


		lfn.SetLdest(pre.size());

		uint32_t bytes = m_stat.getBytesLFNSent() + msg.GetSerializedSize();
		uint32_t num = m_stat.getLFNSent() + 1;
		m_stat.setBytesLFNSent(bytes);
		m_stat.setLFNSent(num);

	    Ptr<Packet> packet = Create<Packet> ();
	    packet->AddHeader(msg);
	     Ptr<Socket> socket;
				socket = FindSocketWithInterfaceAddress(iface);
				int numBytes = socket->SendTo(packet, 0, InetSocketAddress(destination,
						ANTOR_DLR_PORT));
				if (numBytes != -1) {
	//cout << "[ " << Simulator::Now().GetSeconds()
	//	 << " sg. ]  LFN message from interface "<<iface.GetLocal()<<" to "<<pre[pre.size() - 1 ]<<"\n";
				}



}

void AntORDLRCore::SendHello() {

	NS_LOG_FUNCTION(this);
	ns3::antordlr::MessageHeader msg;
	msg.SetMessageType(ns3::antordlr::MessageHeader::HELLO_MESSAGE);
	msg.SetTimeToLive(1);
	msg.SetHopCount(0);
	msg.SetMessageSequenceNumber(GetMessageSequenceNumber());

	ns3::antordlr::MessageHeader::Hello &hello = msg.GetHello();
	hello.SetDelay((float) Simulator::Now().GetSeconds());

	float max_ValuePhero = 0.0;
	uint32_t index = 0;
	//Añade destinos al mensaje.
	std::vector<Ipv4Address> destinations;
	destinations.clear();
	destinations = m_table.GetAllDestination(num_report_destination);
	NS_LOG_INFO(" Calculate destinations correctly, number of destination : "<< destinations.size());

	std::vector < uint32_t > entry;
	for (uint32_t i = 0; i < destinations.size(); i++) {
		entry.clear();
		bool result = m_table.LookupEntriesByDest(destinations[i], entry);
		if (result) {
			index = 0;
			max_ValuePhero = 0.0;
			double reg;
			double vir;
			double max_value;
			for (uint32_t n = 0; n < entry.size(); n++) {

				uint32_t idx = entry[n];
				Ipv4Address dst = m_table.GetRoutingEntry(idx).GetDestination();
				//if ( dst == Ipv4Address::GetLoopback() )
				// continue;
				NS_LOG_INFO("Destination "<< m_table.GetRoutingEntry(idx).GetDestination() << " NextHop "
						<< m_table.GetRoutingEntry(idx).GetNextHop());
				reg = m_table.GetRoutingEntry(idx).GetPheroTable().pheroReg;
				vir = m_table.GetRoutingEntry(idx).GetPheroTable().pheroVir;
				max_value = std::max(reg, vir);
				if (max_value > max_ValuePhero) {
					max_ValuePhero = (float) max_value;
					index = entry[n];
				}

			}//for

			hello.destValue.push_back(
					(ns3::antordlr::MessageHeader::Hello::DestinationValue ) {
						destinations[i],(float) max_ValuePhero
					});
		}
	} //for

	m_stat.setHelloSent(m_stat.getHelloSent() + 1);
	m_stat.setByteHelloSent(m_stat.getByteHelloSent() + msg.GetSerializedSize());

	for (std::map<Ptr<Socket> , Ipv4InterfaceAddress>::const_iterator j =
			m_socketAddresses.begin(); j != m_socketAddresses.end(); ++j) {
		Ptr<Packet> packet = Create<Packet> ();
		packet->AddHeader(msg);
		Ptr<Socket> socket = j->first;
		Ipv4InterfaceAddress iface = j->second;
		// Send to all-hosts broadcast if on /32 addr, subnet-directed otherwise
		Ipv4Address destination;
		if (iface.GetMask() == Ipv4Mask::GetOnes()) {
			destination = Ipv4Address("255.255.255.255");
		} else {
			destination = iface.GetBroadcast();
		}
		socket->SendTo(packet, 0,
				InetSocketAddress(destination, ANTOR_DLR_PORT));

	}
}

void AntORDLRCore::SendPFA(Ipv4Address dst) {
	NS_LOG_FUNCTION(this);

	 ns3::antordlr::MessageHeader msg;
	 msg.SetMessageType(ns3::antordlr::MessageHeader::PFA_MESSAGE);
	 uint32_t minhop = 255;
	 RoutingTableEntry rt;
	 m_table.LookupMinHop(dst,rt);
	 minhop = rt.GetHop();
	 //std::cout << "MIN HOPS : " << (uint32_t) minhop << endl;
	 msg.SetTimeToLive( minhop );
     msg.SetHopCount(0);
	 msg.SetMessageSequenceNumber(GetMessageSequenceNumber());
	 ns3::antordlr::MessageHeader::Pfa &pfa = msg.GetPfa();
	 Ptr<Ipv4L3Protocol> l3 = m_ipv4->GetObject<Ipv4L3Protocol> ();
	 	  	Ipv4InterfaceAddress iface = l3->GetAddress(1, 0);

	 pfa.addrSession.source = iface.GetLocal();
	 pfa.addrSession.destination = dst;

	 pfa.listP.push_back(iface.GetLocal());
	 // Send it
	  vector<RoutingTableEntry> toDst;
	  m_table.LookupRoutesWithReg(dst,toDst);

	 if (toDst.size() == 0) {
	 // Las PFAs nunca se envian en modo BROADCAST

	 cout << "[ " << Simulator::Now().GetSeconds() << " sg. ] Nodo "
	 << iface.GetLocal() << " no tiene ruta alternativa al destino "
	 << dst << std::endl;

	 } else {
	 //uint32_t entry;
	 uint32_t hop = m_heuristic.CalculateProactiveNextHop(toDst,
	 parameter_beta2);
	 Ptr<Ipv4Route> route = m_table.GetRoutingEntry(hop).GetRoute();
	 bool f = toDst[hop].GetVisitedLink();
	 if ( f ) {

	 cout << " [ " << Simulator::Now().GetSeconds()
	 << " sg. ] Nodo " <<toDst[hop].GetNextHop()
	 << " ya pertenece a la session [ "
	 << pfa.addrSession.source << " - "
	 << pfa.addrSession.destination << " ]" << std::endl;
	 	 return;
	 }

	 // if best virtual pheromone es at least 10% better than the best regular pheromone, send PFA
	 // if (routes[hop].GetPheroTable().pheroVir*0.1 >= routes[hop].GetPheroTable().pheroReg) {
	 m_stat.setBytesPFASent(m_stat.getBytesPFASent()
	 + msg.GetSerializedSize());
	 m_stat.setPFASent(m_stat.getPFASent() + 1);
	  Ipv4Address dir = route->GetGateway();
	 /*
	  cout << "[ " << Simulator::Now().GetSeconds() << " sg. ] Node "
	 << iface.GetLocal() << " sending PFA ( "
	 << m_stat.getPFASent() << " ) to " << dst << std::endl;
	 */
	 //Send it
	 Ptr<Packet> packet = Create<Packet> ();
	 packet->AddHeader(msg);
	 Ptr<Socket> socket = FindSocketWithInterfaceAddress(iface);

	 socket->SendTo(packet, 0, InetSocketAddress(toDst[hop].GetNextHop(),ANTOR_DLR_PORT));


	 //}



	 }
}
/*
void AntORDLRCore::LFNTimerExpire() {

	NS_LOG_FUNCTION (this);
	m_LFNCount = 0;
	m_LFNTimer.Schedule(Seconds(1));

}
*/
void AntORDLRCore::RFATimerExpire(Ipv4Address dst) {
	std::vector<RoutingTableEntry> rts;

 m_table.LookupRoutesWithReg(dst, rts);
		if ( rts.size() == 0) {

		Time t = Time(0.01 * MilliSeconds(UniformVariable().GetInteger(0, 100)));
     if (!m_RFATimer.IsRunning() && dst != Ipv4Address()){
        m_RFATimer.SetArguments(dst);
		m_RFATimer.Schedule(Seconds(1.0) - t);
		SendRFA(dst);
     }
		//m_queue.DropPacketWithDst(dst);
		} else {
			uint32_t hop = m_heuristic.CalculateReactiveNextHop(rts,
					parameter_beta3);
			Ptr<Ipv4Route> rt = rts[hop].GetRoute();
			SendPacketFromQueue(dst, rt);
			Time t = Time(0.01 * MilliSeconds(UniformVariable().GetInteger(0, 100)));

			if (!m_RFATimer.IsRunning() && dst != Ipv4Address()){
				m_RFATimer.SetArguments(dst);
			m_RFATimer.Schedule(Seconds(1.0) - t);
			}
		}


}

void AntORDLRCore::RRFATimerExpire(Ipv4Address dst) {
	NS_LOG_FUNCTION (this);

	m_RRFATimer.Cancel();

	std::cout << "[ " << Simulator::Now().GetSeconds() << " sg. ] expire RRFA to destination " << dst << "\n";

	std::vector<RoutingTableEntry> toDest;
	if (m_table.LookupRoutesWithReg(dst, toDest)){
		uint32_t hop = m_heuristic.CalculateReactiveNextHop(toDest,parameter_beta3);
		SendPacketFromQueue(toDest[hop].GetDestination(),toDest[hop].GetRoute());
		 cout << "SE\n";
	 }
}
void AntORDLRCore::SendRRFA(Ipv4Address dir, uint8_t hops) {

	NS_LOG_FUNCTION(this);
	ns3::antordlr::MessageHeader msg;
	msg.SetMessageType(ns3::antordlr::MessageHeader::RRFA_MESSAGE);
	msg.SetTimeToLive(255);
	msg.SetHopCount(0);
	msg.SetMessageSequenceNumber(GetMessageSequenceNumber());
	m_local_hops = hops;
	ns3::antordlr::MessageHeader::Rrfa &rrfa = msg.GetRrfa();
	for (std::map<Ptr<Socket> , Ipv4InterfaceAddress>::const_iterator j =
			m_socketAddresses.begin(); j != m_socketAddresses.end(); ++j) {
		Ptr<Socket> socket = j->first;
		Ipv4InterfaceAddress iface = j->second;
		rrfa.addrSession.source = iface.GetLocal();
        rrfa.listP.push_back(iface.GetLocal());
		rrfa.addrSession.destination = dir;
		// Send it
		std::vector<RoutingTableEntry> toDest;
		m_table.LookupRoutesWithReg(dir, toDest);
		if (toDest.size() == 0) {
			//cout <<"Pass BROADCAST\n";
			msg.SetNumBroadcast(1);
			m_stat.setBytesRRFASent(m_stat.getBytesRRFASent()
					+ msg.GetSerializedSize());
			m_stat.setRRFASent(m_stat.getRRFASent() + 1);

			cout << "[ " << Simulator::Now().GetSeconds() << " sg. ] Node "
					<< iface.GetLocal() << " sending RRFA ( "
					<< m_stat.getRRFASent() << " ) in mode BROADCAST to "
					<< dir << "\n";

			//Send it
			Ptr<Packet> packet = Create<Packet> ();
			packet->AddHeader(msg);
			socket->SendTo(packet, 0, InetSocketAddress(Ipv4Address(
					"255.255.255.255"), ANTOR_DLR_PORT));
		} else {
			msg.SetNumBroadcast(0);
			//cout <<"Pass NO BROADCAST\n";
			m_stat.setBytesRRFASent(m_stat.getBytesRRFASent()
					+ msg.GetSerializedSize());
			m_stat.setRRFASent(m_stat.getRRFASent() + 1);

			cout << "[ " << Simulator::Now().GetSeconds() << " sg. ] Node "
					<< iface.GetLocal() << " sending RRFA ( "
					<< m_stat.getRRFASent() << " ) to "
					<< rrfa.addrSession.destination << "\n";

			uint32_t hop = m_heuristic.CalculateProactiveNextHop(toDest,
					parameter_beta1);
			//Sent it
			Ptr<Packet> packet = Create<Packet> ();
			packet->AddHeader(msg);
			Ipv4Address destination = toDest[hop].GetNextHop();
			socket->SendTo(packet, 0, InetSocketAddress(destination,
					ANTOR_DLR_PORT));

		}
	}
}

void AntORDLRCore::SendRFA(Ipv4Address dst) {

	NS_LOG_FUNCTION(this);

	ns3::antordlr::MessageHeader msg;
	msg.SetMessageType(ns3::antordlr::MessageHeader::RFA_MESSAGE);
	msg.SetTimeToLive(255);
	msg.SetHopCount(0);
	msg.SetMessageSequenceNumber(GetMessageSequenceNumber());

	ns3::antordlr::MessageHeader::Rfa &rfa = msg.GetRfa();
	//std::vector<RoutingTableEntry> rts;
	//std::vector<RoutingTableEntry> routes = m_table.GetEntriesWithReg(
	//		rfa.addrSession.destination);
	vector < uint32_t > idxs;


    /*
	bool b = m_table.LookupEntriesByDest(dst, idxs);
	if (b) {
		RoutingTableEntry old = m_table.GetRoutingEntry(idxs[0]);
    	m_table.DeleteRoute(old.GetDestination(),idxs[0]);
			}
     */
	//if (routes.size() == 0) {
	// Send it
	// Send RFA as subnet directed broadcast from each interface used by aodv
	for (std::map<Ptr<Socket> , Ipv4InterfaceAddress>::const_iterator j =
			m_socketAddresses.begin(); j != m_socketAddresses.end(); ++j) {
		Ptr<Socket> socket = j->first;
		Ipv4InterfaceAddress iface = j->second;
		rfa.addrSession.source = iface.GetLocal();
		rfa.addrSession.destination = dst;

		rfa.listP.push_back(iface.GetLocal());
		m_stat.setBytesRFASent(m_stat.getBytesRFASent()
				+ msg.GetSerializedSize());
		m_stat.setRFASent(m_stat.getRFASent() + 1);
		// La primera RFA se envia en modo BROADCAST
		/*
		cout << "[ " << Simulator::Now().GetSeconds() << " sg. ] Node "
				<< iface.GetLocal() << " sending RFA ( " << m_stat.getRFASent()
				<< " ) in mode BROADCAST to " << dst << "\n";
        */
		Ptr<Packet> packet = Create<Packet> ();
		packet->AddHeader(msg);
		//DataT::PrepareDataTag(packet, ANTOR_DLR_CODE);

		m_stat.setPacketSent(m_stat.getPacketSent() + 1);
		m_stat.setBytePacketSent(m_stat.getBytePacketSent() + packet->GetSize());

		// Send to all-hosts broadcast if on /32 addr, subnet-directed otherwise
		Ipv4Address destination;
		if (iface.GetMask() == Ipv4Mask::GetOnes()) {
			destination = Ipv4Address("255.255.255.255");
		} else {
			destination = iface.GetBroadcast();
		}
		//destination = Ipv4Address("255.255.255.255");
		if (socket->SendTo(packet, 0, InetSocketAddress(destination,
				ANTOR_DLR_PORT))) {
			//	cout << "sending from "<< iface.GetLocal()<<"\n";
		}
	}

	 /*
			cout << "[ " << Simulator::Now().GetSeconds() << " sg. ] Node "
					<< iface.GetLocal() << " sending RFA ( " << m_stat.getRFASent()
					<< " ) in mode BROADCAST to " << dst << "\n";
     */

			//ScheduleRreqRetry(dst);
	/*
	 if (!m_helloTimer.IsRunning()) {
	 m_helloTimer.Cancel();
	 m_helloTimer.Schedule(m_helloInterval - Time(0.01 * MilliSeconds(
	 UniformVariable().GetInteger(0, 10))));
	 }
	 */
}

void AntORDLRCore::SendRRBA(Ipv4InterfaceAddress srcIface, Ipv4Address dst,
		std::vector<Ipv4Address> dirs, uint32_t hops) {
	NS_LOG_FUNCTION(this);
	Ipv4Address src = srcIface.GetLocal();
	ns3::antordlr::MessageHeader msg;
	msg.SetMessageType(ns3::antordlr::MessageHeader::RRBA_MESSAGE);
	msg.SetTimeToLive(255);
	msg.SetHopCount(0);
	msg.SetMessageSequenceNumber(GetMessageSequenceNumber());

	ns3::antordlr::MessageHeader::Rrba &rrba = msg.GetRrba();
	rrba.addrSession.source = src;
	rrba.addrSession.destination = dst;
	rrba.delay = (float) Simulator::Now().GetSeconds();

	Ipv4Address dir = rrba.addrSession.destination;
	if (dirs.size() > 0) {
		//FIXME for (uint32_t k = 0; k < dirs.size() -1; k++) {
		for (uint32_t k = 0; k < dirs.size() - 1; k++) {
			rrba.listP.push_back(dirs[k]);

		}
		dir = dirs[dirs.size() - 1];
	}
	m_stat.setBytesRRBASent(m_stat.getBytesRRBASent() + msg.GetSerializedSize());
	m_stat.setRRBASent(m_stat.getRRBASent() + 1);

	/*
	 cout << "Next Hop "<< dir<<endl;
	 cout << src << " has next precusor list : \n";

	 for (uint32_t m = 0; m < dirs.size();m++)
	 cout << dirs[m]<<" ";

	 cout << endl;
	 */
	cout << "[ " << Simulator::Now().GetSeconds() << " sg. ] Nodo " << src
			<< " enviando RRBA a " << dst << std::endl;

	Ptr<Socket> socket = FindSocketWithInterfaceAddress(srcIface);
	Ptr<Packet> packet = Create<Packet> ();
	packet->AddHeader(msg);
	std::vector<RoutingTableEntry> rts;
		bool result = m_table.LookupAllRoutes(dir, rts);
		if (result) {
			int numBytes = socket->SendTo(packet, 0, InetSocketAddress(dir,
					ANTOR_DLR_PORT));
			if (numBytes != -1) {

				NS_LOG_INFO(src<<" sending packet to "<<dir<<" of bytes "<< numBytes);
				//cout <<src <<" sending packet to "<<dir<<" of bytes "<< numBytes << endl;
			}
		} else {

			cout << "Not allowing unicast message to " << dir << endl;
		}

}

void AntORDLRCore::SendRBA(Ipv4InterfaceAddress srcIface, Ipv4Address dst,
		std::vector<Ipv4Address> dirs) {

	NS_LOG_FUNCTION(this);
	Ipv4Address src = srcIface.GetLocal();
	ns3::antordlr::MessageHeader msg;
	msg.SetMessageType(ns3::antordlr::MessageHeader::RBA_MESSAGE);
	msg.SetTimeToLive(50);
	msg.SetHopCount(0);
	msg.SetMessageSequenceNumber(GetMessageSequenceNumber());

	ns3::antordlr::MessageHeader::Rba &rba = msg.GetRba();
	rba.addrSession.source = src;
	rba.addrSession.destination = dst;
	rba.delay = (float) Simulator::Now().GetSeconds();


	Ipv4Address dir = rba.addrSession.destination;
	if (dirs.size() > 0) {
		//FIXME for (uint32_t k = 0; k < dirs.size() -1; k++) {
		for (uint32_t k = 0; k < dirs.size()-1; k++) {
			rba.listP.push_back(dirs[k]);

		}
		dir = dirs[dirs.size() - 1];
	}

	m_stat.setBytesRBASent(m_stat.getBytesRBASent() + msg.GetSerializedSize());
		m_stat.setRBASent(m_stat.getRBASent() + 1);
	/*
	 uint32_t idx;
	 if (m_table.GetIndexRoute(dir,dir,idx)) {
	 m_table.DeleteRoute(dir,idx);
	 cout << "Delete temporal route to "<< dir <<" correctly"<<endl;
	 }
	 */
     /*
	 cout << "Next Hop "<< dir<<endl;
	 cout << src << " has next precusor list : \n";

	 for (uint32_t m = 0; m < rba.listP.size();m++)
	 cout << rba.listP[m]<<" ";

	 cout << endl;
	 */
/*
	cout << "[ " << Simulator::Now().GetSeconds() << " sg. ] Nodo " << src
			<< " enviando RBA a " << dst << std::endl;
    */
	Ptr<Socket> socket = FindSocketWithInterfaceAddress(srcIface);
	Ptr<Packet> packet = Create<Packet> ();
	packet->AddHeader(msg);
	/*
	dir = Ipv4Address("255.255.255.255");
	socket->SendTo(packet, 0, InetSocketAddress(dir,
					ANTOR_DLR_PORT));
	//cout <<src <<" sending packet "<<endl;
	*/
	std::vector<RoutingTableEntry> rts;

	bool result = m_table.LookupAllRoutes(dir, rts);
	if (result) {
		int numBytes = socket->SendTo(packet, 0, InetSocketAddress(dir,
				ANTOR_DLR_PORT));
		if (numBytes != -1) {

			NS_LOG_INFO(src<<" sending packet to "<<dir<<" of bytes "<< numBytes);
			//cout <<src <<" sending packet to "<<dir<<" of bytes "<< numBytes << endl;
		}
	} else {

		cout << "Not allowing RBA unicast message to " << dir << endl;
	}

}

void AntORDLRCore::SendPBA(Ipv4Address addr, std::vector<Ipv4Address> dirs) {

	 NS_LOG_FUNCTION(this);
	 ns3::antordlr::MessageHeader msg;
	 msg.SetMessageType(ns3::antordlr::MessageHeader::PBA_MESSAGE);
	 msg.SetTimeToLive(255);
	 msg.SetHopCount(0);
	 msg.SetMessageSequenceNumber(GetMessageSequenceNumber());

	 ns3::antordlr::MessageHeader::Pba &pba = msg.GetPba();
	 Ptr<Ipv4L3Protocol> l3 = m_ipv4->GetObject<Ipv4L3Protocol> ();
	 	 		 	  	Ipv4InterfaceAddress iface = l3->GetAddress(1, 0);
	 pba.addrSession.source = iface.GetLocal();
	 pba.addrSession.destination = addr;
	 pba.delay = (float) Simulator::Now().GetSeconds();
	 m_stat.setBytesPBASent(m_stat.getBytesPBASent() + msg.GetSerializedSize());
	 m_stat.setPBASent(m_stat.getPBASent() + 1);
	 Ipv4Address dir = pba.addrSession.destination;
	 if (dirs.size() > 0) {
	 for (uint32_t k = 0; k < dirs.size() - 1; k++) {
	 pba.listP.push_back(dirs[k]);

	 }
	 dir = dirs[dirs.size() - 1];
	 }
	 cout << "[ " << Simulator::Now().GetSeconds() << " sg. ] Nodo "
	 << iface.GetLocal() << " enviando PBA a " << addr << std::endl;
	    std::vector<RoutingTableEntry> rts;
	 	bool result = m_table.LookupRoutesWithReg(dir, rts);
	 	if (result) {
	 		Ptr<Packet> packet = Create<Packet> ();
	 			 packet->AddHeader(msg);
	 			 Ptr<Socket> socket = FindSocketWithInterfaceAddress(iface);
	 		int numBytes = socket->SendTo(packet, 0, InetSocketAddress(dir,
	 				ANTOR_DLR_PORT));
	 		if (numBytes != -1) {

	 			NS_LOG_INFO(iface.GetLocal()<<" sending packet to "<<dir<<" of bytes "<< numBytes);
	 			//cout <<src <<" sending packet to "<<dir<<" of bytes "<< numBytes << endl;
	 		}
	 	} else {

	 		cout << "Not allowing unicast message to " << dir << endl;
	 	}
}

void AntORDLRCore::ProcessPFA(ns3::antordlr::MessageHeader &msg,
		const Ipv4Address &receiverAddr, const Ipv4Address &senderAddr) {

	 ns3::antordlr::MessageHeader::Pfa &pfa = msg.GetPfa();
	 Time nowActual = Simulator::Now();
	 msg.SetTimeToLive(msg.GetTimeToLive() - 1);
	 msg.SetHopCount(msg.GetHopCount() + 1);

		 	Ipv4Address origin = pfa.addrSession.source;

			//std::cout <<"[ " << Simulator::Now().GetSeconds() << " sg. ] "<<receiverAddr <<" receives PFA from "<< senderAddr<<endl;
			uint32_t entry;
				bool b = m_table.LookupEntryByDestAndNextHop(senderAddr,senderAddr, entry);
				if (!b){
					PheroTable f_t;
							f_t.pheroReg = 100.0;
							f_t.numAvgHop = m_heuristic.CalculateNewNumAvgHop(0.0, parameter_gamma,
									msg.GetHopCount());
							f_t.pheroVir = 0.0;
							Ptr<NetDevice> dev = m_ipv4-> GetNetDevice(
									m_ipv4->GetInterfaceForAddress(receiverAddr));
							Ipv4InterfaceAddress iface = m_ipv4 -> GetAddress(
									m_ipv4->GetInterfaceForAddress(receiverAddr), 0);
							RoutingTableEntry rt(dev, senderAddr, iface, 1,
									senderAddr, f_t);

							m_table.AddRoute(rt);
			           //std::cout <<receiverAddr << " "<<senderAddr<<endl;
				}
	 	Ptr<Ipv4L3Protocol> l3 = m_ipv4->GetObject<Ipv4L3Protocol> ();
	 		 	  	Ipv4InterfaceAddress iface = l3->GetAddress(1, 0);
	 if (iface.GetLocal() == pfa.addrSession.destination) {
	 m_stat.setPFAReceived(m_stat.getPFAReceived() + 1);
	 m_stat.setBytesPFAReceived(m_stat.getBytesPFAReceived()
	 + msg.GetSerializedSize());
	 SendPBA(pfa.addrSession.source, pfa.listP);

	 } else {
      Ipv4Address dst = pfa.addrSession.destination;
	 // Añadiendo nuevo nodo a la lista P
	 pfa.listP.push_back(iface.GetLocal());
	 // Send it
		 vector<RoutingTableEntry> toDst;
		  m_table.LookupRoutesWithReg( dst,toDst );

		 if (toDst.size() == 0) {
		 // Las PFAs nunca se envian en modo BROADCAST
		 cout << "[ " << Simulator::Now().GetSeconds() << " sg. ] Nodo "
		 << iface.GetLocal() << " no tiene ruta alternativa al destino "
		 << dst << std::endl;
			  return;
		 } else {
		  //uint32_t entry;
		 uint32_t hop = m_heuristic.CalculateProactiveNextHop(toDst,
		 parameter_beta2);
		 //bool b2 = m_table.Lookup(routes[hop].GetDestination(),
		 //				routes[hop].GetNextHop(), entry);
		 //////           PARA RUTAS DE NODOS DISJUNTOS    ///
	     Ptr<Ipv4Route> route = m_table.GetRoutingEntry(hop).GetRoute();
	     bool f = toDst[hop].GetVisitedLink();

	     if ( f ) { //&& toDst.size() > 1) {

		 cout << " [ " << Simulator::Now().GetSeconds()
		 << " sg. ] Nodo " << toDst[hop].GetNextHop()
		 << " ya pertenece a la session [ "
		 << pfa.addrSession.source << " - "
		 << pfa.addrSession.destination << " ]" << std::endl;

	     return;
	     }
		  // if best virtual pheromone es at least 10% better than the best regular pheromone, send PFA
		 // if (routes[hop].GetPheroTable().pheroVir*0.1 >= routes[hop].GetPheroTable().pheroReg) {
		 m_stat.setBytesPFASent(m_stat.getBytesPFASent()
		 + msg.GetSerializedSize());
		 m_stat.setPFASent(m_stat.getPFASent() + 1);
		 /*
		 cout << "[ " << Simulator::Now().GetSeconds() << " sg. ] Node "
		 << iface.GetLocal() << " re-sending PFA ( "
		 << m_stat.getPFASent() << " ) to " << dst << std::endl;
		 */
		 //Send it
		 Ptr<Packet> packet = Create<Packet> ();
		 packet->AddHeader(msg);
		 Ptr<Socket> socket = FindSocketWithInterfaceAddress(iface);
		 Ipv4Address dir = toDst[hop].GetNextHop();
		 socket->SendTo(packet, 0, InetSocketAddress(dir,ANTOR_DLR_PORT));


		 }
		 }
}

void AntORDLRCore::ProcessRRFA(ns3::antordlr::MessageHeader &msg,
		const Ipv4InterfaceAddress &receiverIface,
		const Ipv4Address &senderAddr) {

	Ipv4Address receiverAddr = receiverIface.GetLocal();
	ns3::antordlr::MessageHeader::Rrfa &rrfa = msg.GetRrfa();
	Time nowActual = Simulator::Now();
	msg.SetTimeToLive(msg.GetTimeToLive() - 1);
	msg.SetHopCount(msg.GetHopCount() + 1);
	uint32_t id = msg.GetMessageSequenceNumber();

	Ipv4Address origin = rrfa.addrSession.source;

	Ipv4Address destination = rrfa.addrSession.destination;
	/*
	 *  Node checks to determine whether it has received a RREQ with the same Originator IP Address and RREQ ID.
	 *  If such a RREQ has been received, the node silently discards the newly received RREQ.
	 */
	if (m_rreqIdCache.IsDuplicate(origin, id)) {
		//std::cout << "Ignoring RREQ due to duplicate to "<< rfa.addrSession.destination <<std::endl;
		return;
	}


	UpdateNeighbor(receiverAddr,senderAddr ,Seconds(1.0));



	if (receiverAddr == rrfa.addrSession.destination) {
		//if (idRFA < msg.GetMessageSequenceNumber()) {
		//	idRFA = msg.GetMessageSequenceNumber();
		m_stat.setRRFAReceived(m_stat.getRRFAReceived() + 1);
		m_stat.setBytesRRFAReceived(m_stat.getBytesRRFAReceived()
				+ msg.GetSerializedSize());
		SendRRBA(receiverIface, origin, rrfa.listP,msg.GetHopCount());

		//}

	} else {
		// Añadiendo nuevo nodo a la lista P
		rrfa.listP.push_back(receiverAddr);
		Ptr<Socket> socket = FindSocketWithInterfaceAddress(receiverIface);
		// Send it
		std::vector<RoutingTableEntry> toDest;
		m_table.LookupRoutesWithReg(destination, toDest);
		if (toDest.size() == 0) {
			msg.SetNumBroadcast(msg.GetNumBroadcast() + 1);
			m_stat.setBytesRRFASent(m_stat.getBytesRRFASent()
					+ msg.GetSerializedSize());
			m_stat.setRRFASent(m_stat.getRRFASent() + 1);
			//Send it
			Ptr<Packet> packet = Create<Packet> ();
			packet->AddHeader(msg);
			socket->SendTo(packet, 0, InetSocketAddress(Ipv4Address(
					"255.255.255.255"), ANTOR_DLR_PORT));
		} else {
			m_stat.setBytesRRFASent(m_stat.getBytesRRFASent()
					+ msg.GetSerializedSize());
			m_stat.setRRFASent(m_stat.getRRFASent() + 1);


			uint32_t hop = m_heuristic.CalculateProactiveNextHop(toDest,
					parameter_beta1);
			//Sent it
			Ptr<Packet> packet = Create<Packet> ();
			packet->AddHeader(msg);
			Ipv4Address dir = toDest[hop].GetNextHop();
			socket->SendTo(packet, 0, InetSocketAddress(dir, ANTOR_DLR_PORT));

		}
	}

}

void AntORDLRCore::ProcessRFA(ns3::antordlr::MessageHeader &msg,
		const Ipv4InterfaceAddress &receiverIface,
		const Ipv4Address &senderAddr) {

	Ipv4Address receiverAddr = receiverIface.GetLocal();
	ns3::antordlr::MessageHeader::Rfa &rfa = msg.GetRfa();
	Time nowActual = Simulator::Now();
	msg.SetTimeToLive(msg.GetTimeToLive() - 1);
	msg.SetHopCount(msg.GetHopCount() + 1);
	uint32_t id = msg.GetMessageSequenceNumber();
    //std::cout << receiverAddr << " receive RFA from "<<senderAddr<<endl;
	Ipv4Address origin = rfa.addrSession.source;
   if (IsMyOwnAddress(origin)) {
		return;
	}

	if (m_rreqIdCache.IsDuplicate(origin, id)) {
		//std::cout << "Ignoring RREQ due to duplicate to "<< rfa.addrSession.destination <<std::endl;
		return;
	}
	//UpdateRoutingTable(receiverAddr,senderAddr, msg.GetHopCount(), id);

	UpdateNeighbor(receiverAddr,senderAddr ,Seconds(1.0));

	Time t = Time::From(AllowedHelloLoss * m_helloInterval);
	m_nbs.Update(senderAddr, t);

	if (receiverAddr == rfa.addrSession.destination) {
		// if (IsMyOwnAddress (rfa.addrSession.destination)){
			m_stat.setRFAReceived(m_stat.getRFAReceived() + 1);
			m_stat.setBytesRFAReceived(m_stat.getBytesRFAReceived()
					+ msg.GetSerializedSize());
			if (IsMyOwnAddress (receiverAddr ))
			    {
				  NS_LOG_DEBUG ("Send reply since I am the destination");
			      SendRBA(receiverIface, origin, rfa.listP);
			      return;
			    }


	} else {
		// Añadiendo nuevo nodo a la lista P
		rfa.listP.push_back(receiverAddr);

		for (std::map<Ptr<Socket> , Ipv4InterfaceAddress>::const_iterator j =
				m_socketAddresses.begin(); j != m_socketAddresses.end(); ++j) {
			Ptr<Socket> socket = j->first;
			Ipv4InterfaceAddress iface = j->second;

			// Send it
			Ptr<Packet> packet = Create<Packet> ();
			packet->AddHeader(msg);

			// Send to all-hosts broadcast if on /32 addr, subnet-directed otherwise
			Ipv4Address destination;
			if (receiverIface.GetMask() == Ipv4Mask::GetOnes()) {
				destination = Ipv4Address("255.255.255.255");
			} else {
				destination = iface.GetBroadcast();
			}

			//destination = Ipv4Address("255.255.255.255");
			socket->SendTo(packet, 0, InetSocketAddress(destination,
					ANTOR_DLR_PORT));
		}
	}

}

void AntORDLRCore::ProcessPBA(ns3::antordlr::MessageHeader &msg,
		const Ipv4Address &receiverAddr, const Ipv4Address &senderAddr) {

	 ns3::antordlr::MessageHeader::Pba &pba = msg.GetPba();
	 Time nowActual = Simulator::Now();


	 Time old = Seconds((double) pba.GetDelay());
	 Time t = DELAY_ALTERNATIVE(nowActual, old);
	 msg.SetTimeToLive(msg.GetTimeToLive() - 1);
	 msg.SetHopCount(msg.GetHopCount() + 1);
	 pba.SetDelay((float) (pba.GetDelay() + t.GetSeconds()));


	 Ptr<Ipv4L3Protocol> l3 = m_ipv4->GetObject<Ipv4L3Protocol> ();
	 Ipv4InterfaceAddress iface = l3->GetAddress(1,0);
     Ipv4Address receiver = iface.GetLocal();
     Ipv4Address dst = pba.addrSession.source;
	 // UPDATE ROUTING PHEROMONE TABLE
	 		uint32_t entry;
	        bool b = m_table.LookupEntryByDestAndNextHop(dst, senderAddr,entry);
	 		Ptr<NetDevice> dev = m_ipv4->GetNetDevice (m_ipv4->GetInterfaceForAddress (receiver));
	 		  RoutingTableEntry newEntry (/*device=*/ dev, /*dst=*/ dst, /*iface=*/ m_ipv4->GetAddress (m_ipv4->GetInterfaceForAddress (receiver), 0),/*hop=*/ msg.GetHopCount(),
	 		                                          /*nextHop=*/ senderAddr);

	 		if (!b) {
	 			// Nueva informacion de encaminamiento
	 			//std::cout << m_mainAddress << " NEW Phero table \n";
	 			PheroTable f_t;
	 			f_t.pheroReg = m_heuristic.CalculateNewPheroReg(0.0, parameter_alpha,
	 					t);
	 			f_t.numAvgHop = m_heuristic.CalculateNewNumAvgHop(0.0, parameter_gamma,
	 					msg.GetHopCount());
	 			f_t.pheroVir = 0.0;
	 			newEntry.SetVisitedLink(true);
	 			newEntry.SetPheroTable(f_t);
	 			m_table.AddRoute(newEntry);

	 		} else {

	 			//Actualizacion de tabla de rutas


	 			double reg = m_table.GetRoutingEntry(entry).GetPheroTable().pheroReg;
	 			PheroTable f_t;
	 					f_t.pheroReg = m_heuristic.CalculateNewPheroReg(reg,
	 							parameter_alpha, t);

	 					f_t.numAvgHop = m_heuristic.CalculateNewNumAvgHop(m_table.GetRoutingEntry(entry).GetPheroTable().numAvgHop,
	 							parameter_gamma, msg.GetHopCount());

	 					f_t.pheroVir = 0.0;
	 					newEntry.SetVisitedLink(true);
	 					newEntry.SetPheroTable(f_t);
	 					if (m_table.UpdateNew(newEntry)) {
	 						//cout << " [ " << Simulator::Now().GetSeconds()
	 						//				<< " sg. ] Updating RBA Route from "<<receiverAddr << " to "<<rt.GetDestination()<<" by next hop "<<rt.GetNextHop() <<endl;

	 					}

	 		}



       Time t_n = Time::From(AllowedHelloLoss * m_helloInterval);
		m_nbs.Update(senderAddr, t_n);
	 if (iface.GetLocal() == pba.addrSession.destination) {

	 cout << " [ " << Simulator::Now().GetSeconds()
	 << " sg. ] Session ALTERNATIVA establecida entre "
	 << iface.GetLocal() << " y " << pba.addrSession.source
	 << std::endl;
	 std::cout << (uint32_t) msg.GetHopCount() << " <--- HOPs  to "
	 << pba.addrSession.source << std::endl;
	 vector<RoutingTableEntry> rts;
	 bool result2 = m_table.LookupRoutesWithReg(pba.addrSession.source, rts);
	 		if (result2) {
	 uint32_t hop = m_heuristic.CalculateReactiveNextHop(rts,
	 					parameter_beta3);
	 			Ptr<Ipv4Route> rt = rts[hop].GetRoute();
	 			SendPacketFromQueue(pba.addrSession.source, rt);
	 }
	 //Active Data Session
	 //m_session.dst = pba.addrSession.source;
	 //m_session.src = pba.addrSession.destination;


	 m_stat.setPBAReceived(m_stat.getPBAReceived() + 1);
	 m_stat.setBytesPBAReceived(m_stat.getBytesPBAReceived()
	 + msg.GetSerializedSize());

	 } else {

		 uint32_t entry;
		 		bool b = m_table.LookupEntryByDestAndNextHop(pba.addrSession.source,
		 					senderAddr, entry);
		 			if (b) {
		 				//std::cout <<receiver <<" to destination "<<rba.addrSession.source<<" next hop "<<sender <<" has :\n";
		 				RoutingTableEntry rt = m_table.GetRoutingEntry(entry);
		 				rt.DeletePrecursor(pba.addrSession.source);
		 				for (uint32_t m=0; m < pba.listP.size();m++){
		 				  // std::cout << rba.listP[m]<<" ";
		 					rt.InsertPrecursor(pba.listP[m]);
		 				}
		 				rt.InsertPrecursor(pba.addrSession.destination);
		 				// cout << endl;
		                m_table.Update(rt,entry);
		 			}

	 uint32_t tam = pba.listP.size();
	 Ipv4Address dir = pba.addrSession.destination;
	 if (tam > 0) {
	 dir = pba.listP[tam - 1];
	 pba.listP.erase(pba.listP.begin() + tam - 1);
	 }

	 // Send it
	 Ptr<Packet> packet = Create<Packet> ();
	 packet->AddHeader(msg);
	 Ptr<Socket> socket = FindSocketWithInterfaceAddress(iface);
	 std::vector<RoutingTableEntry> rts;
	 	bool result = m_table.LookupRoutesWithReg(dir, rts);
	 	if (result) {
	 		int numBytes = socket->SendTo(packet, 0, InetSocketAddress(dir,
	 				ANTOR_DLR_PORT));
	 		if (numBytes != -1) {

	 			NS_LOG_INFO(iface.GetLocal()<<" sending packet to "<<dir<<" of bytes "<< numBytes);
	 			//cout <<src <<" sending packet to "<<dir<<" of bytes "<< numBytes << endl;
	 		}
	 	} else {

	 		cout << "Not allowing unicast message to " << dir << endl;
	 	}
	 }

}

void AntORDLRCore::ProcessRRBA(ns3::antordlr::MessageHeader &msg,
		const Ipv4InterfaceAddress &receiverIface,
		const Ipv4Address &senderAddr) {
	Ipv4Address receiverAddr = receiverIface.GetLocal();

	cout << "received from "<< senderAddr <<endl;
	ns3::antordlr::MessageHeader::Rrba &rrba = msg.GetRrba();
	Time nowActual = Simulator::Now();
	Time old = Seconds((double) rrba.GetDelay());
	Time t = DELAY_ALTERNATIVE(nowActual, old);
	msg.SetTimeToLive(msg.GetTimeToLive() - 1);
	msg.SetHopCount(msg.GetHopCount() + 1);
	rrba.SetDelay((float) (rrba.GetDelay() + t.GetSeconds()));

	// UPDATE ROUTING PHEROMONE TABLE
		uint32_t entry;
        Ipv4Address dst = rrba.addrSession.source;
		bool b = m_table.LookupEntryByDestAndNextHop(dst, senderAddr,entry);
		Ptr<NetDevice> dev = m_ipv4->GetNetDevice (m_ipv4->GetInterfaceForAddress (receiverAddr));
		  RoutingTableEntry newEntry (/*device=*/ dev, /*dst=*/ dst, /*iface=*/ m_ipv4->GetAddress (m_ipv4->GetInterfaceForAddress (receiverAddr), 0),/*hop=*/ msg.GetHopCount(),
		                                          /*nextHop=*/ senderAddr );

		if (!b) {
			// Nueva informacion de encaminamiento
			//std::cout << m_mainAddress << " NEW Phero table \n";
			PheroTable f_t;
			f_t.pheroReg = m_heuristic.CalculateNewPheroReg(0.0, parameter_alpha,
					t);
			f_t.numAvgHop = m_heuristic.CalculateNewNumAvgHop(0.0, parameter_gamma,
					msg.GetHopCount());
			f_t.pheroVir = 0.0;
			newEntry.SetVisitedLink(true);
			newEntry.SetPheroTable(f_t);
			m_table.AddRoute(newEntry);

		} else {

			//Actualizacion de tabla de rutas


			double reg = m_table.GetRoutingEntry(entry).GetPheroTable().pheroReg;
			PheroTable f_t;
					f_t.pheroReg = m_heuristic.CalculateNewPheroReg(reg,
							parameter_alpha, t);

					f_t.numAvgHop = m_heuristic.CalculateNewNumAvgHop(m_table.GetRoutingEntry(entry).GetPheroTable().numAvgHop,
							parameter_gamma, msg.GetHopCount());

					f_t.pheroVir = 0.0;

					newEntry.SetPheroTable(f_t);
					newEntry.SetVisitedLink(true);
					m_table.UpdateNew(newEntry);

		}

		Time t_n = Time::From(AllowedHelloLoss * m_helloInterval);
		m_nbs.Update(senderAddr, t_n);

	//NS_LOG_INFO("RBA Route from "<<receiverAddr << " to "<< senderAddr);


	if (receiverAddr == rrba.addrSession.destination) {
		cout << " [ " << Simulator::Now().GetSeconds()
					<< " sg. ] Session REPAIR between " << receiverAddr
					<< " and " << rrba.addrSession.source << std::endl;
		//m_RFATimer.Cancel();
		m_stat.setRRBAReceived(m_stat.getRRBAReceived() + 1);
		m_stat.setBytesRRBAReceived(m_stat.getBytesRRBAReceived()
				+ msg.GetSerializedSize());
		std::cout << (uint32_t) msg.GetHopCount() << " <--- HOPs  to "
				<< rrba.addrSession.source << std::endl;
		m_RRFATimer.Cancel();
		std::vector<RoutingTableEntry> rts;
		bool result = m_table.LookupRoutesWithReg(rrba.addrSession.source, rts);
		if (result) {
			uint32_t hop = 0;//m_heuristic.CalculateReactiveNextHop(rts,
			//parameter_beta3);
			Ptr<Ipv4Route> rt = rts[hop].GetRoute();
			SendPacketFromQueue(rrba.addrSession.source,rt);


			return;
		}
		//if (IsMyOwnAddress (rba.addrSession.source))
		//  {


		//  }
	} else {
		uint32_t entry;
				bool b = m_table.LookupEntryByDestAndNextHop(rrba.addrSession.source,
							senderAddr, entry);
					if (b) {
						//std::cout <<receiver <<" to destination "<<rba.addrSession.source<<" next hop "<<sender <<" has :\n";
						RoutingTableEntry rt = m_table.GetRoutingEntry(entry);
						for (uint32_t m=0; m < rrba.listP.size();m++){
						  // std::cout << rba.listP[m]<<" ";
							rt.InsertPrecursor(rrba.listP[m]);
						}
						rt.InsertPrecursor(rrba.addrSession.destination);
						// cout << endl;
		               m_table.Update(rt,entry);
					}

		int32_t tam = rrba.listP.size();
		Ipv4Address dir = rrba.addrSession.destination;
		if (tam > 0) {
			dir = rrba.listP[tam - 1];
			rrba.listP.erase(rrba.listP.begin() + tam - 1);
		}

		Ptr<Socket> socket;
		socket = FindSocketWithInterfaceAddress(receiverIface);
		//    Ipv4InterfaceAddress iface = j->second;
		// Send it
		Ptr<Packet> packet = Create<Packet> ();
		packet->AddHeader(msg);
		std::vector<RoutingTableEntry> rts;
				bool result = m_table.LookupRoutesWithReg(dir, rts);
				if (result) {
					int numBytes = socket->SendTo(packet, 0, InetSocketAddress(dir,
							ANTOR_DLR_PORT));
					if (numBytes != -1) {

						//cout <<(uint32_t) msg.GetHopCount()<<" "<< receiver << " sending packet to " << dir
						//		<< " of bytes " << numBytes << endl;
					}
				} else {

					cout << "Not allowing unicast message to " << dir << endl;
				}


	}
}

void AntORDLRCore::ProcessRBA(ns3::antordlr::MessageHeader &msg,
		const Ipv4InterfaceAddress &receiverIface, const Ipv4Address &sender) {

	NS_LOG_FUNCTION (this << " src " << sender);
	Ipv4Address receiver = receiverIface.GetLocal();

	ns3::antordlr::MessageHeader::Rba &rba = msg.GetRba();
    Ipv4Address dst = rba.addrSession.source;

	//cout << receiver <<" received from "<< sender <<endl;

	 Ipv4Address dir = rba.addrSession.destination;

	//FIXME
	Time nowActual = Simulator::Now();
	Time old = Seconds((double) rba.GetDelay());
	Time t = DELAY_ALTERNATIVE(nowActual, old);
	msg.SetTimeToLive(msg.GetTimeToLive() - 1);
	msg.SetHopCount(msg.GetHopCount() + 1);
	rba.SetDelay((float) (rba.GetDelay() + t.GetSeconds()));
    uint32_t hop = msg.GetHopCount();

/*
    if (dir == dst){

    	   UpdateNeighbor(receiver,dst, t);
    	std::cout << receiver << " "<<dst<<std::endl;
    	   return;

    	}
*/
    // UPDATE ROUTING PHEROMONE TABLE
	uint32_t entry;

	bool b = m_table.LookupEntryByDestAndNextHop(dst, sender,entry);
	Ptr<NetDevice> dev = m_ipv4->GetNetDevice (m_ipv4->GetInterfaceForAddress (receiver));
	  RoutingTableEntry newEntry (/*device=*/ dev, /*dst=*/ dst, /*iface=*/ m_ipv4->GetAddress (m_ipv4->GetInterfaceForAddress (receiver), 0),/*hop=*/ hop,
	                                          /*nextHop=*/ sender );

	if (!b) {
		// Nueva informacion de encaminamiento
		//std::cout << m_mainAddress << " NEW Phero table \n";
		PheroTable f_t;
		f_t.pheroReg = m_heuristic.CalculateNewPheroReg(0.0, parameter_alpha,
				t);
		f_t.numAvgHop = m_heuristic.CalculateNewNumAvgHop(0.0, parameter_gamma,
				msg.GetHopCount());
		f_t.pheroVir = 0.0;
		newEntry.SetPheroTable(f_t);
		newEntry.SetVisitedLink(true);
		m_table.AddRoute(newEntry);

	} else {

		//Actualizacion de tabla de rutas


		double reg = m_table.GetRoutingEntry(entry).GetPheroTable().pheroReg;
		PheroTable f_t;
				f_t.pheroReg = m_heuristic.CalculateNewPheroReg(reg,
						parameter_alpha, t);

				f_t.numAvgHop = m_heuristic.CalculateNewNumAvgHop(m_table.GetRoutingEntry(entry).GetPheroTable().numAvgHop,
						parameter_gamma, msg.GetHopCount());

				f_t.pheroVir = 0.0;
                newEntry.SetVisitedLink(true);
				newEntry.SetPheroTable(f_t);
				if (m_table.UpdateNew(newEntry)) {
					//cout << " [ " << Simulator::Now().GetSeconds()
					//				<< " sg. ] Updating RBA Route from "<<receiverAddr << " to "<<rt.GetDestination()<<" by next hop "<<rt.GetNextHop() <<endl;

				}

	}

	Time t_n = Time::From(AllowedHelloLoss * m_helloInterval);
	m_nbs.Update(sender, t_n);

	//NS_LOG_INFO("RBA Route from "<<receiverAddr << " to "<< senderAddr);


	if (receiver == rba.addrSession.destination) {

		//if (m_stat.getRFASent() == 1) {
			cout << " [ " << Simulator::Now().GetSeconds()
					<< " sg. ] Session PRINCIPAL entre "
					<< receiver << " y " << rba.addrSession.source << std::endl;
		//}
		/*
		else {
		 cout << " [ " << Simulator::Now().GetSeconds()
		 << " sg. ] Session PRINCIPAL RESTABLECIDA entre "
		 << receiver << " y " << rba.addrSession.source << std::endl;

		 }
*/

		vector<RoutingTableEntry> toDst;
		if (m_table.LookupRoutesWithReg(rba.addrSession.source, toDst)) {

			uint32_t hop = m_heuristic.CalculateReactiveNextHop(toDst,
								parameter_beta3);
			SendPacketFromQueue(rba.addrSession.source, toDst[hop].GetRoute());
		}
		m_stat.setRBAReceived(m_stat.getRBAReceived() + 1);
		m_stat.setBytesRBAReceived(m_stat.getBytesRBAReceived()
				+ msg.GetSerializedSize());
		//std::cout << (uint32_t) msg.GetHopCount() << " <--- HOPs  to "
		//		<< rba.addrSession.source << std::endl;
		 PFATimerExpire(rba.addrSession.source);
	} else {
        uint32_t entry;
		bool b = m_table.LookupEntryByDestAndNextHop(rba.addrSession.source,
					sender, entry);
			if (b) {

				//std::cout <<receiver <<" to destination "<<rba.addrSession.source<<" next hop "<<sender <<" has :\n";
				RoutingTableEntry rt = m_table.GetRoutingEntry(entry);
				for (uint32_t m=0; m < rba.listP.size();m++){
				  // std::cout << rba.listP[m]<<" ";
					rt.InsertPrecursor(rba.listP[m]);
				}
				rt.InsertPrecursor(rba.addrSession.destination);
				// cout << endl;
               m_table.Update(rt,entry);
			}

		int32_t tam = rba.listP.size();
					Ipv4Address dir = rba.addrSession.destination;
					if (tam > 0) {
						dir = rba.listP[tam - 1];
						rba.listP.erase(rba.listP.begin() + tam - 1);
					}


		m_stat.setRBAReceived(m_stat.getRBAReceived() + 1);
				m_stat.setBytesRBAReceived(m_stat.getBytesRBAReceived()
						+ msg.GetSerializedSize());
		//cout << dir << endl;
		Ptr<Socket> socket;
		socket = FindSocketWithInterfaceAddress(receiverIface);
		//    Ipv4InterfaceAddress iface = j->second;
		// Send it
		Ptr<Packet> packet = Create<Packet> ();
		packet->AddHeader(msg);
		/*
		dir = Ipv4Address("255.255.255.255");
		socket->SendTo(packet, 0, InetSocketAddress(dir,
						ANTOR_DLR_PORT));
		*/
		//cout << receiver << "resending packet\n";


		std::vector<RoutingTableEntry> rts;
		bool result = m_table.LookupAllRoutes(dir, rts);
		if (result) {
			int numBytes = socket->SendTo(packet, 0, InetSocketAddress(dir,
					ANTOR_DLR_PORT));
			if (numBytes != -1) {

				//cout <<(uint32_t) msg.GetHopCount()<<" "<< receiver << " sending packet to " << dir
				//		<< " of bytes " << numBytes << endl;
			}
		} else {

			cout << "Not allowing seconds RBA unicast message to " << dir << endl;
		}

	}
}

void AntORDLRCore::ProcessLFN(ns3::antordlr::MessageHeader &msg,
		const Ipv4Address &receiverAddr, const Ipv4Address &senderAddr) {

	/*
	NS_LOG_FUNCTION (this << " from " << senderAddr);
	ns3::antordlr::MessageHeader::Lfn &lfn = msg.GetLfn();
	msg.SetTimeToLive(msg.GetTimeToLive() - 1);
	msg.SetHopCount(msg.GetHopCount() + 1);
	uint32_t id = msg.GetMessageSequenceNumber();

	Ipv4Address org = lfn.originator;

if (IsMyOwnAddress(org)) {
	 return;
	}
   if  (m_rreqIdCache.IsDuplicate(org, id)) {
	 //std::cout << "Ignoring RREQ due to duplicate to "<< rfa.addrSession.destination <<std::endl;
	return;
    }
vector<vector<Ipv4Address> > ss;
   std::vector<Ipv4Address> un2 = m_table.GetDestinationLFN(senderAddr);
   	cout << " [ " << Simulator::Now().GetSeconds() << " sg. ] " << receiverAddr
		<< endl;
    vector<Ipv4Address> neigh = m_nbs.GetNeighbors();
    vector<Ipv4Address> sendAddr;
    vector<Ipv4Address> sources;
	Ptr<Ipv4L3Protocol> l3 = m_ipv4->GetObject<Ipv4L3Protocol> ();
				Ipv4InterfaceAddress iface = l3->GetAddress(1, 0);
	for (uint32_t m = 0; m < lfn.dest.size(); m++) {
			Ipv4Address dest = lfn.dest[m];
			uint32_t idx;
			if (m_table.LookupEntryByDestAndNextHop(dest, senderAddr, idx)) {
				RoutingTableEntry rt = m_table.GetRoutingEntry(idx);
				sources.clear();
				sendAddr.clear();
				rt.GetPrecursors(sources);
				m_table.DeleteRoute(rt.GetDestination(),idx);
				uint32_t ids;
				for (uint32_t n = 0; n < sources.size();n++){
				if (FindCopies(sources[n],neigh,ids) && sources[n] != iface.GetLocal()){
					sendAddr.push_back(sources[n]);
				}//if


		}//for
				}
			ss.push_back(sendAddr);
	}
for (uint32_t j=0; j< ss.size();j++){
		std::vector<Ipv4Address> unSend;
		unSend = ss[j];
	  if (unSend.size() > 0){
				Ipv4Address dir;// = pre[pre.size() - 1];
				uint32_t bytes = m_stat.getBytesLFNSent()
									+ msg.GetSerializedSize();
				uint32_t num = m_stat.getLFNSent() + 1;
				m_stat.setBytesLFNSent(bytes);
				m_stat.setLFNSent(num);

				Ptr<Socket> socket; // = FindSocketWithInterfaceAddress(iface);
				Ptr<Packet> packet = Create<Packet> ();
							packet->AddHeader(msg);
							socket = FindSocketWithInterfaceAddress(iface);
                       bool cond = false;
							for (uint32_t c=0; c < unSend.size();c++) {
							if (!cond){
								dir = unSend[c];

								std::vector<RoutingTableEntry> rts;
								dir = Ipv4Address("255.255.255.255");
							bool result = m_table.LookupRoutesWithReg(dir, rts);
							if (result ) {
								int numBytes = socket->SendTo(packet, 0, InetSocketAddress(
										dir, ANTOR_DLR_PORT));
								if (numBytes != -1) {

									//cout << iface.GetLocal() << " sending packet to "
									//		<< dir << " of bytes " << numBytes << endl;
								}
								cond = true;
								break;
							}


							}

						}
							if (!cond){
							dir = Ipv4Address("255.255.255.255");
							//socket->SendTo(packet, 0, InetSocketAddress(dir, ANTOR_DLR_PORT));
							}


	}
	}
*/
	NS_LOG_FUNCTION (this << " from " << senderAddr);
	ns3::antordlr::MessageHeader::Lfn &lfn = msg.GetLfn();
	msg.SetTimeToLive(msg.GetTimeToLive() - 1);
	msg.SetHopCount(msg.GetHopCount() + 1);
	//uint32_t id = msg.GetMessageSequenceNumber();

	Ipv4Address dest = lfn.originator;

/*
	if (IsMyOwnAddress(dest)) {
		return;
	}
*/
	 Ptr<Ipv4L3Protocol> l3 = m_ipv4->GetObject<Ipv4L3Protocol> ();
	 Ipv4InterfaceAddress iface = l3->GetAddress(1, 0);



	 int32_t tam = lfn.ldest - 1;




	 if (iface.GetLocal() != lfn.dest[tam]){

		 return;
	 }

//	 cout <<" [ " << Simulator::Now().GetSeconds()
//		<< " sg. ] "<<receiverAddr << endl;

   	/*
	cout << iface.GetLocal()<< " has next precursors :\n";
	for (uint32_t m=0; m < sources.size(); m++){
	       Ipv4Address d = sources[m];
	       cout << d << " ";

		}
     std::cout << endl;
     */
		uint32_t idx;
		if (m_table.LookupEntryByDestAndNextHop(dest,senderAddr,idx)){
					//std::cout << lfn.dest[k]<<endl;
					RoutingTableEntry rt = m_table.GetRoutingEntry(idx);
					PheroTable ft = rt.GetPheroTable();
					ft.pheroReg = 0.0;
					ft.pheroVir = 0.0;
					//rt.SetPheroTable(ft);
	    			//m_table.Update(rt,idx);

					m_table.DeleteRoute(rt.GetDestination(),idx);
					//vector<Ipv4Address> un;
					//un.push_back(unreachable[k]);
    				}

if (tam > 0){
  lfn.dest.erase(lfn.dest.begin() + lfn.dest.size() - 1);
	lfn.SetLdest(lfn.dest.size());
	uint32_t bytes = m_stat.getBytesLFNSent() + msg.GetSerializedSize();
		uint32_t num = m_stat.getLFNSent() + 1;
		m_stat.setBytesLFNSent(bytes);
		m_stat.setLFNSent(num);

		Ipv4Address destination;
				 		if (iface.GetMask() == Ipv4Mask::GetOnes()) {
				 			destination = Ipv4Address("255.255.255.255");
				 		} else {
				 			destination = iface.GetBroadcast();
				 		}


       	Ptr<Socket> socket; // = FindSocketWithInterfaceAddress(iface);
       	    Ptr<Packet> packet = Create<Packet> ();
       	    packet->AddHeader(msg);
       		socket = FindSocketWithInterfaceAddress(iface);
			int numBytes = socket->SendTo(packet, 0, InetSocketAddress(destination,
							ANTOR_DLR_PORT));
			if (numBytes != -1) {
		//cout << "[ " << Simulator::Now().GetSeconds()
		//	 << " sg. ] RE-SEND LFN message from interface "<<iface.GetLocal()<<" to "<<lfn.dest[lfn.GetLdest() - 1]<<"\n";
					}

}else {
  RFATimerExpire(dest);
  m_PFATimer.Cancel();
}


}
void AntORDLRCore::ProcessULN(ns3::antordlr::MessageHeader &msg,
		const Ipv4Address &receiverAddr, const Ipv4Address &senderAddr) {


	NS_LOG_FUNCTION (this << " from " << senderAddr);
	ns3::antordlr::MessageHeader::Uln uln = msg.GetUln();
	msg.SetTimeToLive(msg.GetTimeToLive() - 1);
	msg.SetHopCount(msg.GetHopCount() + 1);
	cout << " [ " << Simulator::Now().GetSeconds() << " sg. ] " << receiverAddr
			<<" ULN " <<endl;
	uint32_t idx;
	m_table.GetIndexRoute(uln.dst, senderAddr, idx);
	RoutingTableEntry myrt = m_table.GetRoutingEntry(idx);
	PheroTable f_t;
	f_t.pheroReg = 0;
	f_t.numAvgHop = myrt.GetPheroTable().numAvgHop;
	f_t.pheroVir = myrt.GetPheroTable().pheroVir;
	//m_table.UpdateNew(myrt);
	m_table.DeleteRoute(myrt.GetDestination(),idx);
    //
	 Ptr<Ipv4L3Protocol> l3 = m_ipv4->GetObject<Ipv4L3Protocol> ();
	 Ipv4InterfaceAddress iface = l3->GetAddress(1, 0);
	if (uln.src != receiverAddr){

   //FIXME

	Ptr<Packet> packet = Create<Packet> ();
	packet->AddHeader(msg);
	Ptr<Socket> socket;
	socket = FindSocketWithInterfaceAddress(iface);
	std::vector<RoutingTableEntry> toOrigin;
	if (m_table.LookupAllRoutes(uln.src, toOrigin)){
	int numBytes = socket->SendTo(packet, 0, InetSocketAddress(toOrigin[0].GetNextHop(),
					ANTOR_DLR_PORT));
	if (numBytes != -1) {
		cout << "[ " << Simulator::Now().GetSeconds()
		 << " sg. ]  ULN message from interface "<<iface.GetLocal()<<" to "<< toOrigin[0].GetNextHop()<<"\n";
	}

}
}else {
	m_table.DeleteRoute(myrt.GetDestination(),idx);
	std::cout << "Source "<< iface.GetLocal() << "receives ULN\n";
}
}

void AntORDLRCore::ProcessHello(ns3::antordlr::MessageHeader &msg,
		const Ipv4Address &receiverAddr, const Ipv4Address &senderAddr) {

	ns3::antordlr::MessageHeader::Hello &hello = msg.GetHello();
	Time nowActual = Simulator::Now();
	float _timeOld = (float) hello.GetDelay();
	Time timeOld2 = Seconds((double) _timeOld);
	//std::cout << "Node : " << m_mainAddress << " -- " << nowActual.GetSeconds() << " -- " << timeOld2.GetSeconds() << "\n";
	Time res2 = DELAY_ALTERNATIVE(nowActual, timeOld2);
	m_stat.setHelloReceived(m_stat.getHelloReceived() + 1);
	m_stat.setByteHelloReceived(m_stat.getByteHelloReceived()
			+ msg.GetSerializedSize());
	Time t = Time::From(AllowedHelloLoss * m_helloInterval);
	m_nbs.Update(senderAddr, t);
	//UpdateRoutingNeighbor(receiverAddr, senderAddr, res2, msg.GetHopCount(),
	//hello, msg.GetMessageSequenceNumber());
	UpdateNeighbor(receiverAddr,senderAddr ,res2);

	UpdateRoutingNeighbor(receiverAddr, senderAddr, res2,1, hello);
}

// calcula informacion de encaminamiento de los vecinos a un salto
void AntORDLRCore::UpdateNeighbor(Ipv4Address receiverAddr,
		Ipv4Address dst, Time t) {

	bool b;
	uint32_t idx;
	b = m_table.LookupEntryByDestAndNextHop(dst, dst, idx);
	PheroTable f_t;

	if (!b) {
		// Nueva informacion de encaminamiento al destino (VECINO)

		f_t.pheroReg = 100;
		f_t.numAvgHop = 1.0;

		f_t.pheroVir = 0.0;
		Ptr<NetDevice> dev = m_ipv4-> GetNetDevice(
				m_ipv4->GetInterfaceForAddress(receiverAddr));
		Ipv4InterfaceAddress iface = m_ipv4 -> GetAddress(
				m_ipv4->GetInterfaceForAddress(receiverAddr), 0);
		RoutingTableEntry rt(dev, dst,iface, 1, dst, f_t);
		//Add Routing Entry
		//FIXME
		m_table.AddRoute(rt);
		}











}

// calcula informacion de encaminamiento de los vecinos a un salto
void AntORDLRCore::UpdateRoutingNeighbor(Ipv4Address receiverAddr,
		Ipv4Address addr, Time t, uint32_t hops,
		ns3::antordlr::MessageHeader::Hello hello) {


	 // PROCESS PHEROMONE DIFUSSION

	 for (uint32_t i = 0; i < hello.destValue.size(); i++) {
	 ns3::antordlr::MessageHeader::Hello::DestinationValue data =
	 hello.destValue[i];
	 // Send to all-hosts broadcast if on /32 addr, subnet-directed otherwise
	 Ptr<Ipv4L3Protocol> l3 = m_ipv4->GetObject<Ipv4L3Protocol> ();
	 	Ipv4InterfaceAddress iface = l3->GetAddress(1, 0);
	 Ipv4Address destination;
	 		if (iface.GetMask() == Ipv4Mask::GetOnes()) {
	 			destination = Ipv4Address("255.255.255.255");
	 		} else {
	 			destination = iface.GetBroadcast();
	 		}
	 if (receiverAddr != data.address && data.address != destination) {
	 bool b;
	 uint32_t _entry;
	 b = m_table.LookupEntryByDestAndNextHop(data.address, addr, _entry);
	 PheroTable pherot;
	 double pheroVir;
	 // Set new Information
	 if (!b) {

	 pheroVir = m_heuristic.CalculateDifussionPhero(data.best, t);
	 ////  FIXME  ////////////
	 pherot.pheroVir = pheroVir;
	 pherot.pheroReg = 0.0;
	 pherot.numAvgHop = m_heuristic.CalculateNewNumAvgHop(0.0, parameter_gamma,
	 hops);
	 Ptr<NetDevice> dev = m_ipv4-> GetNetDevice(
	 m_ipv4->GetInterfaceForAddress(receiverAddr));
	 Ipv4InterfaceAddress iface = m_ipv4 -> GetAddress(
	 m_ipv4->GetInterfaceForAddress(receiverAddr), 0);
	 RoutingTableEntry rt(dev, data.address, iface, hops, addr, pherot);
	 rt.SetVisitedLink(false);
	 m_table.AddRoute(rt);
	 //std::cout << receiverAddr <<" set "<<pherot.pheroVir<<" to destination "<<data.address <<endl;
	 } //if
	 else {// Update Information


	 //if (m_table.GetTableEntry(_entry).GetHop() > 1) {
		double reg = m_table.GetRoutingEntry(_entry).GetPheroTable().pheroReg;
		double vir = m_table.GetRoutingEntry(_entry).GetPheroTable().pheroVir;

	 if (vir > reg) {

	 pheroVir
	 = m_heuristic.CalculateDifussionPhero(data.best, t);
	 // FIXME//////////
	 pherot.pheroVir = pheroVir;
	 pherot.pheroReg = 0.0;
	 pherot.numAvgHop = m_heuristic.CalculateNewNumAvgHop(
	 m_table.GetRoutingEntry(_entry).GetPheroTable().numAvgHop,
	 parameter_gamma, m_table.GetRoutingEntry(_entry).GetHop() + hops);
	 RoutingTableEntry newRt = m_table.GetRoutingEntry(_entry);
	 newRt.SetPheroTable(pherot);
	 newRt.SetHop(newRt.GetHop() + hops);
	 newRt.SetVisitedLink(false);
	 //std::cout << receiverAddr <<" update "<<pherot.pheroVir<<" "<< newRt.GetHop()<<endl;
	 m_table.Update(newRt,_entry);
	 }


	 //} //if

	 }//if

	 } //for

	 }



}
//FIXME
void AntORDLRCore::SendMessageWhenBreaksLinkToNextHop(Ipv4Address nextHop) {

	/*
	NS_LOG_FUNCTION (this << nextHop);
		vector<Ipv4Address> unreachable;
		unreachable.clear();

		unreachable = m_table.GetDestinationLFN(nextHop);
		vector<Ipv4Address> sources;

        vector<Ipv4Address> sendAddr;
		Ptr<Ipv4L3Protocol> l3 = m_ipv4->GetObject<Ipv4L3Protocol> ();
			Ipv4InterfaceAddress iface = l3->GetAddress(1, 0);
		std::vector<Ipv4Address> neigh = m_nbs.GetNeighbors();
		vector<vector<Ipv4Address> > ss;
		for (uint32_t k= 0; k < unreachable.size();k++){
			uint32_t idx;
			if (m_table.LookupEntryByDestAndNextHop(unreachable[k],nextHop,idx)){
				RoutingTableEntry rt = m_table.GetRoutingEntry(idx);
			    sources.clear();
				sources = m_table.GetSources(unreachable[k],nextHop);
			    //m_table.DeleteRoute(rt.GetDestination(),idx);
				for (uint32_t n=0;n < sources.size();n++){
					uint32_t id;
					if (FindCopies(sources[n],neigh,id) && sources[n] != iface.GetLocal()){
						sendAddr.push_back(sources[n]);
					}
				}


         	ss.push_back(sendAddr);
			sendAddr.clear();

			}

			}
for (uint32_t j=0; j < ss.size();j++)
   if (ss[j].size() > 0){
	SendLFN(ss[j],unreachable);
   }
*/
	NS_LOG_FUNCTION (this << nextHop);
	vector<Ipv4Address> unreachable;
	unreachable.clear();

	unreachable = m_table.GetDestinationLFN(nextHop);
	vector<Ipv4Address> sources;


	Ptr<Ipv4L3Protocol> l3 = m_ipv4->GetObject<Ipv4L3Protocol> ();
		Ipv4InterfaceAddress iface = l3->GetAddress(1, 0);
	for (uint32_t k= 0; k < unreachable.size();k++){
		uint32_t idx;
		if (m_table.LookupEntryByDestAndNextHop(unreachable[k],nextHop,idx)){
			RoutingTableEntry rt = m_table.GetRoutingEntry(idx);
		    sources = m_table.GetSources(unreachable[k],nextHop);
			PheroTable ft = rt.GetPheroTable();
			ft.pheroReg = 0.0;
	   	    ft.pheroVir = 0.0;
		    //rt.SetPheroTable(ft);
		    //m_table.Update(rt,idx);
		    m_table.DeleteRoute(rt.GetDestination(),idx);
            vector<Ipv4Address> un; // = iface.GetLocal();
            un.push_back(unreachable[k]);
            SendLFN(sources,un);
		}
	}

}


void AntORDLRCore::PFATimerExpire(Ipv4Address dir) {
	std::vector<RoutingTableEntry> rts;

	 m_table.LookupRoutesWithReg(dir, rts);
			if ( rts.size() > 0) {

	 SendPFA(dir);

	 if (!m_PFATimer.IsRunning()){
	  m_PFATimer.SetArguments(dir);
	  m_PFATimer.Schedule(m_PFAInterval);
	 }

}
}

/// Increments message sequence number and returns the new value.
uint16_t AntORDLRCore::GetMessageSequenceNumber() {
	m_messageSequenceNumber = (m_messageSequenceNumber + 1)
			% (ANTOR_DLR_MAX_SEQ_NUM + 1);
	return m_messageSequenceNumber;
}

void AntORDLRCore::SendPacketFromQueue(Ipv4Address dst, Ptr<Ipv4Route> route) {

	NS_LOG_FUNCTION (this);
	QueueEntry queueEntry;
	while (m_queue.Dequeue(dst, queueEntry)) {
		DeferredRouteOutputTag tag;
		Ptr<Packet> p = ConstCast<Packet> (queueEntry.GetPacket());
		if (p->RemovePacketTag(tag) && tag.oif != -1 && tag.oif
				!= m_ipv4->GetInterfaceForDevice(route->GetOutputDevice())) {
			NS_LOG_DEBUG ("Output device doesn't match. Dropped.");
			return;
		}

		UnicastForwardCallback ucb = queueEntry.GetUnicastForwardCallback();
		Ipv4Header header = queueEntry.GetIpv4Header();
		header.SetSource(route->GetSource());
		header.SetTtl(header.GetTtl() + 1); // compensate extra TTL decrement by fake loopback routing
		/*
		 cout << " [ " << Simulator::Now().GetSeconds() << " sg. ] "
		 << header.GetSource() << " sending to "
		 << header.GetDestination() << " by NextHop "
		 << route->GetGateway() << " Packet ID " << p->GetUid() << endl;
		 */
		ucb(route, p, header);
	}

	/*
	 QueueEntry queueEntry;
	 if (m_queue.Dequeue (dst,queueEntry))
	 {
	 DeferredRouteOutputTag tag;
	 Ptr<Packet> p = ConstCast<Packet> (queueEntry.GetPacket ());
	 if (p->RemovePacketTag (tag))
	 {
	 if (tag.oif != -1 && tag.oif != m_ipv4->GetInterfaceForDevice (route->GetOutputDevice ()))
	 {
	 NS_LOG_DEBUG ("Output device doesn't match. Dropped.");
	 return;
	 }
	 }
	 UnicastForwardCallback ucb = queueEntry.GetUnicastForwardCallback ();
	 Ipv4Header header = queueEntry.GetIpv4Header ();
	 header.SetSource (route->GetSource ());
	 header.SetTtl (header.GetTtl () + 1); // compensate extra TTL decrement by fake loopback routing
	 // cout << " [ " << Simulator::Now().GetSeconds()
	 //     									<< " sg. ] "<< header.GetSource() << " sending to "<<header.GetDestination()<<" by NextHop "<< route->GetGateway()
	 //     									<<" Packet ID "<<p->GetUid()<<endl;
	 ucb (route,p,header);
	 if (m_queue.GetSize () != 0 && m_queue.Find (dst))
	 {
	 Simulator::Schedule (MilliSeconds (UniformVariable ().GetInteger (0,100)),
	 &AntORDLRCore::SendPacketFromQueue,this,dst,route);
	 }
	 }

	 */
}
////////////////////////////////////   IP ROUTING   ///////////////////////////

Ptr<Ipv4Route> AntORDLRCore::RouteOutput(Ptr<Packet> p,
		const Ipv4Header &header, Ptr<NetDevice> oif,
		Socket::SocketErrno &sockerr) {
	NS_LOG_FUNCTION (this << header << (oif ? oif->GetIfIndex () : 0));
	if (!p) {
		return LoopbackRoute(header, oif); // later
	}
	if (m_table.GetSize() == 0) {
		sockerr = Socket::ERROR_NOROUTETOHOST;
		NS_LOG_LOGIC ("No antor-dlr interfaces");
		Ptr<Ipv4Route> route;
		return route;
	}
	sockerr = Socket::ERROR_NOTERROR;
	Ptr<Ipv4Route> route;
	Ipv4Address dst = header.GetDestination();
	std::vector<RoutingTableEntry> rts;
	bool result = m_table.LookupRoutesWithReg(dst, rts);
	if (result) {
		uint32_t hop = m_heuristic.CalculateReactiveNextHop(rts,
				parameter_beta3);
		route = rts[hop].GetRoute();
		NS_ASSERT (route != 0);
		NS_LOG_INFO("Exist route to " << route->GetDestination () << " from interface " << route->GetSource ());
		if (oif != 0 && route->GetOutputDevice() != oif) {
			NS_LOG_INFO("Output device doesn't match. Dropped.");
			sockerr = Socket::ERROR_NOROUTETOHOST;
			return Ptr<Ipv4Route> ();
		}

		Time t = Time::From(AllowedHelloLoss * m_helloInterval);
		m_nbs.Update(route->GetGateway(), t);
		//SendPacketFromQueue(dst,route);
		return route;
	}

	// Valid route not found, in this case we return loopback.
	// Actual route request will be deferred until packet will be fully formed,
	// routed to loopback, received from loopback and passed to RouteInput (see below)
	int32_t iif = (oif ? m_ipv4->GetInterfaceForDevice(oif) : -1);
	DeferredRouteOutputTag tag(iif);
	if (!p->PeekPacketTag(tag)) {
		p->AddPacketTag(tag);
	}
	return LoopbackRoute(header, oif);
}



bool AntORDLRCore::RouteInput(Ptr<const Packet> p, const Ipv4Header &header,
		Ptr<const NetDevice> idev, UnicastForwardCallback ucb,
		MulticastForwardCallback mcb, LocalDeliverCallback lcb,
		ErrorCallback ecb) {

	NS_LOG_FUNCTION (this << p->GetUid () << header.GetDestination () << idev->GetAddress ());
	if (m_socketAddresses.empty()) {
		NS_LOG_LOGIC ("No antOR-DLR interfaces");
		return false;
	}NS_ASSERT (m_ipv4 != 0);NS_ASSERT (p != 0);
	// Check if input device supports IP
	NS_ASSERT (m_ipv4->GetInterfaceForDevice (idev) >= 0);

	int32_t iif = m_ipv4->GetInterfaceForDevice(idev);
	Ipv4Address dst = header.GetDestination();
	Ipv4Address origin = header.GetSource();
	//uint32_t d = DataT::GetDataTag(p);
	// Deferred route request
	if (idev == m_lo) {

		DeferredRouteOutputTag tag;
		if (p->PeekPacketTag(tag)) {
			if (tag.oif == -1){
			 DeferredRouteOutput(p, header, ucb, ecb);
			return true;
			}
		}
	}

	// Duplicate of own packet
	if (IsMyOwnAddress(origin))
	  return true;

	// AntOR-DLR doesn't implement a multicast routing protocol yet
	if (dst.IsMulticast()) {
		return false;
	}

	// Broadcast local delivery/forwarding
	Ipv4InterfaceAddress iface = m_ipv4->GetAddress(iif, 0);
	Ipv4Address addr = iface.GetLocal();
	// Broadcast local delivery/forwarding
	for (std::map<Ptr<Socket> , Ipv4InterfaceAddress>::const_iterator j =
			m_socketAddresses.begin(); j != m_socketAddresses.end(); ++j) {
		Ipv4InterfaceAddress iface = j->second;
		if (m_ipv4->GetInterfaceForAddress(iface.GetLocal()) == iif)
			if (dst == iface.GetBroadcast() || dst.IsBroadcast()) {

				if (m_dpd.IsDuplicate(p, header)) {
					NS_LOG_DEBUG ("Duplicated packet " << p->GetUid () << " from " << origin << ". Drop.");
					return true;
				}
				//UpdateRouteLifeTime (origin, ActiveRouteTimeout);
				Ptr<Packet> packet = p->Copy();
				if (lcb.IsNull() == false) {
					NS_LOG_LOGIC ("Broadcast local delivery to " << iface.GetLocal ());
					lcb(p, header, iif);
					// Fall through to additional processing
				} else {
					NS_LOG_ERROR ("Unable to deliver packet locally due to null callback " << p->GetUid () << " from " << origin);
					ecb(p, header, Socket::ERROR_NOROUTETOHOST);
				}

				if (!EnableBroadcast) {
					return true;
				}

				if (header.GetTtl() > 1) {
					NS_LOG_LOGIC ("Forward broadcast. TTL " << (uint16_t) header.GetTtl ());
					std::vector<RoutingTableEntry> toBroadcast;
					if (m_table.LookupRoutesWithReg(dst, toBroadcast)) {
						uint32_t hop = m_heuristic.CalculateReactiveNextHop(toBroadcast,parameter_beta3);

						Ptr<Ipv4Route> route = toBroadcast[hop].GetRoute();//toBroadcast.GetRoute ();
						//Time t = Time::From(AllowedHelloLoss * m_helloInterval);
						//m_nbs.Update(route->GetGateway(), t);
						ucb(route, packet, header);
					} else {
						NS_LOG_DEBUG ("No route to forward broadcast. Drop packet " << p->GetUid ());
					}
				} else {
					NS_LOG_DEBUG ("TTL exceeded. Drop packet " << p->GetUid ());
				}
				return true;
			}
	}
	/*
	 // Unicast local delivery
	 if (m_ipv4->IsDestinationAddress (dst, iif))
	 {
	 if (lcb.IsNull () == false)
	 {
	 NS_LOG_LOGIC ("Unicast local delivery to " << dst);
	 lcb (p, header, iif);
	 }
	 else
	 {
	 NS_LOG_ERROR ("Unable to deliver packet locally due to null callback " << p->GetUid () << " from " << origin);
	 ecb (p, header, Socket::ERROR_NOROUTETOHOST);
	 }
	 return true;
	 }
	 */

	// Unicast local delivery
	if (m_ipv4->IsDestinationAddress(dst, iif)) {
		std::vector<RoutingTableEntry> toOrigin;
		if (m_table.LookupRoutesWithReg(dst, toOrigin)) {
			uint32_t hop = m_heuristic.CalculateReactiveNextHop(toOrigin,parameter_beta3);
			Time t = Time::From(AllowedHelloLoss * m_helloInterval);
			m_nbs.Update(toOrigin[hop].GetNextHop(), t);
		}
		if (lcb.IsNull() == false) {
			NS_LOG_LOGIC ("Unicast local delivery to " << dst);
			lcb(p, header, iif);
		} else {
			NS_LOG_ERROR ("Unable to deliver packet locally due to null callback " << p->GetUid () << " from " << origin);
			ecb(p, header, Socket::ERROR_NOROUTETOHOST);
		}
		return true;
	}

	// Forwarding
	return Forwarding(p, header, ucb, ecb, addr);

}

bool AntORDLRCore::Forwarding(Ptr<const Packet> packet,
		const Ipv4Header & header, UnicastForwardCallback ucb,
		ErrorCallback ecb, Ipv4Address src) {
	NS_LOG_FUNCTION (this);
	m_table.Purge();
	Ipv4Address dst = header.GetDestination();
	Ipv4Address origin = header.GetSource();
	uint32_t newHops1 = DataT::GetDataTag(packet);
	uint32_t newHops = newHops1 + 1;
	Ptr<Packet> p = DataT::UpdateDataTag(packet, newHops1);
	Ptr<Ipv4L3Protocol> l3 = m_ipv4->GetObject<Ipv4L3Protocol> ();
	Ipv4InterfaceAddress iface = l3->GetAddress(1, 0);
	Ipv4Address addr = iface.GetLocal();
    vector<Ipv4Address> neigh = m_nbs.GetNeighbors();
    /*
	 SessionData data;
	 bool res = GetSession(header.GetDestination(),data);
	 if (!res){
	 SessionData newData;
	 newData.src = header.GetSource();
	 newData.dst = header.GetDestination();
	 m_session.insert (std::make_pair (header.GetDestination (), newData));
	 }
	 */
	//m_session.dst = header.GetDestination();
	//m_queue.UpdatePacketHops(dst, newHops);
	//m_table.Purge ();

	std::vector<RoutingTableEntry> toDest;
	RoutingTableEntry toDstRoute;


	if (m_table.LookupAllRoutes(dst, toDest) ) {


		uint32_t hop = m_heuristic.CalculateReactiveNextHop(toDest,parameter_beta3);
		toDstRoute = toDest[hop];
		uint32_t mx;
		bool res = FindCopies(toDstRoute.GetNextHop(),neigh,mx);
		if (res){
		Ptr<Ipv4Route> route = toDstRoute.GetRoute();
		Time t = Time::From(AllowedHelloLoss * m_helloInterval);
		m_nbs.Update (route->GetGateway (), t);
		//UpdateRouteLifeTime (origin, ActiveRouteTimeout);
		SendPacketFromQueue(dst,route);
		ucb(route, p, header);
		 return true;
		} else


		{

          	cout << "Processing RRFA \n";
          	 QueueEntry newEntry(p, header, ucb, ecb, newHops);
						 	bool r = m_queue.Enqueue(newEntry);
            if (r){
			double vv = toDstRoute.GetPheroTable().numAvgHop;
			uint32_t ii;
			m_table.GetIndexRoute(toDstRoute.GetDestination(),toDstRoute.GetNextHop(),ii);
            m_table.DeleteRoute(toDstRoute.GetDestination(),ii);
			Time val = Time(2*MilliSeconds(500)*vv);
          	 std::cout << val.GetSeconds()<<endl;
          	  if (!m_RRFATimer.IsRunning()){
          	    SendRRFA(dst,newHops);
          			          	       // SendULN(origin,toSrcRoute.GetNextHop());
          			 m_RRFATimer.SetArguments(dst);
          			m_RRFATimer.Schedule(val);
          		}


          		}
            return false;
		}
          	 }
/*
	std::vector<RoutingTableEntry> toOrigin;

QueueEntry newEntry(p, header, ucb, ecb, newHops);
	m_queue.Enqueue(newEntry);
		if (m_table.LookupAllRoutes(origin, toOrigin) ) {
		cout << "Processing ULN \n";
		SendULN(origin,toOrigin[0].GetNextHop());
		}
*/

		return false;
         }


void AntORDLRCore::NotifyInterfaceUp(uint32_t i) {

	NS_LOG_FUNCTION (this << m_ipv4->GetAddress (i, 0).GetLocal ());
	Ptr<Ipv4L3Protocol> l3 = m_ipv4->GetObject<Ipv4L3Protocol> ();
	if (l3->GetNAddresses(i) > 1) {
		NS_LOG_INFO ("AntOR-DLR does not work with more then one address per each interface.");
	}
	Ipv4InterfaceAddress iface = l3->GetAddress(i, 0);
	if (iface.GetLocal() == Ipv4Address("127.0.0.1"))
		return;
	//cout << iface.GetLocal() << "\n";
	// Create a socket to listen only on this interface
	Ptr<Socket> socket = Socket::CreateSocket(GetObject<Node> (),
			UdpSocketFactory::GetTypeId());
	NS_ASSERT (socket != 0);
	socket->SetRecvCallback(MakeCallback(&AntORDLRCore::RecvAntORDLR, this));
	socket->BindToNetDevice(l3->GetNetDevice(i));
	socket->Bind(InetSocketAddress(Ipv4Address::GetAny(), ANTOR_DLR_PORT));
		//NS_FATAL_ERROR ("Failed to bind() ANTOR-DLR socket");


		socket->SetAllowBroadcast(true);
		socket->SetAttribute("IpTtl", UintegerValue(1));
		//socket->Connect(InetSocketAddress(Ipv4Address::GetBroadcast(),
		//			ANTOR_DLR_PORT));
		m_socketAddresses.insert(std::make_pair(socket, iface));


	 // Add local broadcast record to the routing table
	 PheroTable f_t;
	 f_t.numAvgHop = 1.0;
	 f_t.pheroReg = 1.0;
	 f_t.pheroVir = 0.0;

	Ptr<NetDevice> dev = m_ipv4->GetNetDevice(m_ipv4->GetInterfaceForAddress(
			iface.GetLocal()));

	RoutingTableEntry rt(/*device=*/dev, /*dst=*/iface.GetBroadcast(), /*iface=*/iface,
			/*hops=*/1, /*next hop=*/iface.GetBroadcast(),f_t);
	 m_table.AddRoute(rt);

	// Allow neighbor manager use this interface for layer 2 feedback if possible
	Ptr<WifiNetDevice> wifi = dev->GetObject<WifiNetDevice> ();
	if (wifi == 0)
		return;
	Ptr<WifiMac> mac = wifi->GetMac();
	if (mac == 0)
		return;

	mac->TraceConnectWithoutContext("TxErrHeader", m_nbs.GetTxErrorCallback());
	m_nbs.AddArpCache(l3->GetInterface(i)->GetArpCache());

}
void AntORDLRCore::NotifyInterfaceDown(uint32_t i) {

	NS_LOG_FUNCTION (this << m_ipv4->GetAddress (i, 0).GetLocal ());

	// Disable layer 2 link state monitoring (if possible)
	Ptr<Ipv4L3Protocol> l3 = m_ipv4->GetObject<Ipv4L3Protocol> ();
	Ptr<NetDevice> dev = l3->GetNetDevice(i);
	Ptr<WifiNetDevice> wifi = dev->GetObject<WifiNetDevice> ();
	if (wifi != 0) {
		Ptr<WifiMac> mac = wifi->GetMac()->GetObject<AdhocWifiMac> ();
		if (mac != 0) {
			mac->TraceDisconnectWithoutContext("TxErrHeader",
					m_nbs.GetTxErrorCallback());
			m_nbs.DelArpCache(l3->GetInterface(i)->GetArpCache());
		}
	}

	// Close socket
	Ptr<Socket> socket = FindSocketWithInterfaceAddress(
			m_ipv4->GetAddress(i, 0));
	NS_ASSERT (socket);
	socket->Close();
	m_socketAddresses.erase(socket);
	if (m_socketAddresses.empty()) {
		NS_LOG_LOGIC ("No antOR-DLR interfaces");
		m_helloTimer.Cancel();
		m_nbs.Clear();
		m_table.Clear();
		return;
	}
	m_table.DeleteAllRoutesFromInterface(m_ipv4->GetAddress(i, 0));
}
void AntORDLRCore::NotifyAddAddress(uint32_t interface,
		Ipv4InterfaceAddress address) {

	NS_LOG_FUNCTION (this << " interface " << interface << " address " << address);
	Ptr<Ipv4L3Protocol> l3 = m_ipv4->GetObject<Ipv4L3Protocol> ();
	if (!l3->IsUp(interface))
		return;
	if (l3->GetNAddresses(interface) == 1) {
		Ipv4InterfaceAddress iface = l3->GetAddress(interface, 0);
		Ptr<Socket> socket = FindSocketWithInterfaceAddress(iface);
		if (!socket) {
			if (iface.GetLocal() == Ipv4Address("127.0.0.1"))
				return;
			// Create a socket to listen only on this interface
			Ptr<Socket> socket = Socket::CreateSocket(GetObject<Node> (),
					UdpSocketFactory::GetTypeId());
			NS_ASSERT (socket != 0);
			socket->SetRecvCallback(MakeCallback(&AntORDLRCore::RecvAntORDLR,
					this));
			socket->BindToNetDevice(l3->GetNetDevice(interface));
			// Bind to any IP address so that broadcasts can be received
			socket->Bind(InetSocketAddress(Ipv4Address::GetAny(),
					ANTOR_DLR_PORT));
			socket->SetAllowBroadcast(true);
			//socket->Connect(InetSocketAddress(Ipv4Address::GetBroadcast(),
			//			ANTOR_DLR_PORT));

			m_socketAddresses.insert(std::make_pair(socket, iface));

			// Add local broadcast record to the routing table

			Ptr<NetDevice> dev = m_ipv4->GetNetDevice(
					m_ipv4->GetInterfaceForAddress(iface.GetLocal()));
			 PheroTable f_t;
				 f_t.numAvgHop = 1.0;
				 f_t.pheroReg = 1.0;
				 f_t.pheroVir = 0.0;
			RoutingTableEntry rt(/*device=*/dev, /*dst=*/iface.GetBroadcast(), /*iface=*/iface, /*hops=*/1,
			/*next hop=*/iface.GetBroadcast(),f_t);
			m_table.AddRoute(rt);
		}
	} else {
		NS_LOG_LOGIC ("AntOR-DLR does not work with more then one address per each interface. Ignore added address");
	}

}

void AntORDLRCore::NotifyRemoveAddress(uint32_t interface,
		Ipv4InterfaceAddress address) {

	NS_LOG_FUNCTION (this);
	Ptr<Socket> socket = FindSocketWithInterfaceAddress(address);
	if (socket) {
		m_table.DeleteAllRoutesFromInterface(address);
		m_socketAddresses.erase(socket);

		Ptr<Ipv4L3Protocol> l3 = m_ipv4->GetObject<Ipv4L3Protocol> ();
		if (l3->GetNAddresses(interface)) {
			Ipv4InterfaceAddress iface = l3->GetAddress(interface, 0);
			// Create a socket to listen only on this interface
			Ptr<Socket> socket = Socket::CreateSocket(GetObject<Node> (),
					UdpSocketFactory::GetTypeId());
			NS_ASSERT (socket != 0);
			socket->SetRecvCallback(MakeCallback(&AntORDLRCore::RecvAntORDLR,
					this));
			// Bind to any IP address so that broadcasts can be received
			socket->Bind(InetSocketAddress(Ipv4Address::GetAny(),
					ANTOR_DLR_PORT));
			socket->SetAllowBroadcast(true);
			//socket->Connect(InetSocketAddress(Ipv4Address::GetBroadcast(),
			//			ANTOR_DLR_PORT));

			m_socketAddresses.insert(std::make_pair(socket, iface));
			// Add local broadcast record to the routing table
			Ptr<NetDevice> dev = m_ipv4->GetNetDevice(
					m_ipv4->GetInterfaceForAddress(iface.GetLocal()));
			RoutingTableEntry rt(/*device=*/dev, /*dst=*/iface.GetBroadcast(), /*iface=*/iface,
			/*hops=*/1, /*next hop=*/iface.GetBroadcast());
			m_table.AddRoute(rt);
		}

		if (m_table.GetSize() == 0) {
			NS_LOG_LOGIC ("No AntOR-DLR interfaces");
			m_helloTimer.Cancel();
			m_nbs.Clear();
			m_table.Clear();
			return;
		}
	} else {
		NS_LOG_LOGIC ("Remove address not participating in AntOR-DLR operation");
	}
}

void AntORDLRCore::PrintRoutingTable(Ptr<OutputStreamWrapper> stream) const {
	// *stream->GetStream () << "Node: " << m_ipv4->GetObject<Node> ()->GetId () << " Time: " << Simulator::Now ().GetSeconds () << "s ";
	// m_table.Print (stream);
}

Ptr<Ipv4Route> AntORDLRCore::LoopbackRoute(const Ipv4Header & hdr, Ptr<
		NetDevice> oif)  {

	NS_LOG_FUNCTION (this << hdr);NS_ASSERT (m_lo != 0);
	Ptr<Ipv4Route> rt = Create<Ipv4Route> ();
	rt->SetDestination(hdr.GetDestination());
	//
	// Source address selection here is tricky.  The loopback route is
	// returned when AntOR does not have a route; this causes the packet
	// to be looped back and handled (cached) in RouteInput() method
	// while a route is found. However, connection-oriented protocols
	// like TCP need to create an endpoint four-tuple (src, src port,
	// dst, dst port) and create a pseudo-header for checksumming.  So,
	// AntOR needs to guess correctly what the eventual source address
	// will be.
	//
	// For single interface, single address nodes, this is not a problem.
	// When there are possibly multiple outgoing interfaces, the policy
	// implemented here is to pick the first available AODV interface.
	// If RouteOutput() caller specified an outgoing interface, that
	// further constrains the selection of source address
	//
	std::map<Ptr<Socket> , Ipv4InterfaceAddress>::const_iterator j =
			m_socketAddresses.begin();
	if (oif) {
		// Iterate to find an address on the oif device
		for (j = m_socketAddresses.begin(); j != m_socketAddresses.end(); ++j) {
			Ipv4Address addr = j->second.GetLocal();
			int32_t interface = m_ipv4->GetInterfaceForAddress(addr);
			if (oif == m_ipv4->GetNetDevice(static_cast<uint32_t> (interface))) {
				rt->SetSource(addr);
				break;
			}
		}
	} else {
		rt->SetSource(j->second.GetLocal());
	}
	NS_ASSERT_MSG (rt->GetSource () != Ipv4Address (), "Valid AntOR-DLR source address not found");
	rt->SetGateway(Ipv4Address("127.0.0.1"));
	rt->SetOutputDevice(m_lo);


	 //m_session.dst = hdr.GetDestination();
     //m_session.src = hdr.GetSource();
	return rt;

}

void AntORDLRCore::DeferredRouteOutput(Ptr<const Packet> p,
		const Ipv4Header & header, UnicastForwardCallback ucb,
		ErrorCallback ecb) {

	NS_LOG_FUNCTION (this << p << header);
	NS_ASSERT (p != 0 && p != Ptr<Packet> ());
	vector<QueueEntry> entries;
	uint32_t max_hops = 0;
	if (m_queue.GetPackets(header.GetDestination(), entries)) {
		for (uint32_t k = 0; k < entries.size(); k++) {
			if (entries[k].GetHops() > max_hops) {
				max_hops = entries[k].GetHops();
			}
		}

	}

	QueueEntry newEntry(p, header, ucb, ecb, max_hops + 1);
	bool result1 = m_queue.Enqueue(newEntry);
	if (result1) {
		// NS_LOG_LOGIC ("Add packet " << p->GetUid () << " to queue. Protocol " << (uint16_t) header.GetProtocol ());
		std::vector<RoutingTableEntry> rts;
		bool result2 = m_table.LookupAllRoutes(header.GetDestination(), rts);
		if (!result2) { //((rts[0].GetFlag () != IN_SEARCH) && result2) ) {
			//Time t2 = Time(MilliSeconds(
			//			UniformVariable().GetInteger(1000, 2000)));
			//Simulator::Schedule(t2, &AntORDLRCore::SendRFA,this,header.GetDestination());

			 Ptr<Ipv4L3Protocol> l3 = m_ipv4->GetObject<Ipv4L3Protocol> ();
		Ipv4InterfaceAddress iface = l3->GetAddress(1,0);


			if (m_stat.getRFASent() == 0){
			Time t = MilliSeconds(UniformVariable().GetInteger(0, 2000));
			Simulator::Schedule(t,&AntORDLRCore::RFATimerExpire,this,header.GetDestination());
			//RFATimerExpire(header.GetDestination());
		  }
		}
	}
}

Ptr<Socket> AntORDLRCore::FindSocketWithInterfaceAddress(
		Ipv4InterfaceAddress addr) const {
	NS_LOG_FUNCTION (this << addr);
	for (std::map<Ptr<Socket> , Ipv4InterfaceAddress>::const_iterator j =
			m_socketAddresses.begin(); j != m_socketAddresses.end(); ++j) {
		Ptr<Socket> socket = j->first;
		Ipv4InterfaceAddress iface = j->second;
		if (iface == addr)
			return socket;
	}
	Ptr<Socket> socket;
	return socket;
}

bool AntORDLRCore::IsMyOwnAddress(Ipv4Address src) {
	NS_LOG_FUNCTION (this << src);
	for (std::map<Ptr<Socket> , Ipv4InterfaceAddress>::const_iterator j =
			m_socketAddresses.begin(); j != m_socketAddresses.end(); ++j) {
		Ipv4InterfaceAddress iface = j->second;
		if (src == iface.GetLocal()) {
			return true;
		}
	}
	return false;
}

bool AntORDLRCore::FindCopies(Ipv4Address dir1, std::vector<Ipv4Address> dirs,
		uint32_t& idx) {
	for (uint32_t t = 0; t < dirs.size(); t++) {
		if (dirs[t] == dir1) {
			idx = t;
			return true;
		}
	}
	return false;
}
/*
void AntORDLRCore::ScheduleRreqRetry(Ipv4Address dst) {
	NS_LOG_FUNCTION (this << dst);

	if (m_addressReqTimer.find(dst) == m_addressReqTimer.end()) {
		Timer timer(Timer::CANCEL_ON_DESTROY);
		m_addressReqTimer[dst] = timer;

	} else {

		m_addressReqTimer[dst].SetFunction(
				&AntORDLRCore::RouteRequestTimerExpire, this);
		m_addressReqTimer[dst].Remove();
		m_addressReqTimer[dst].SetArguments(dst);

		//vector<uint32_t> idxs;

		//bool cond = m_table.LookupEntriesByDest(dst, idxs);
		//NS_ASSERT_MSG( cond, " It's necessary putting the route to "<< dst << " in \"IN_SEARCH\" mode. Please check SendRFA function");

		//RoutingTableEntry rt = m_table.GetRoutingEntry(idxs[0]);
		//rt.IncrementRreqCnt();
		//m_table.DeleteRoute(rt.GetDestination(), idxs[0]);


		idRFA++;
		Time t;
		Time t2 =
				Time(0.01 * MilliSeconds(UniformVariable().GetInteger(0, 10)));
		t = Time(idRFA * NetTraversalTime) - t2;

		m_addressReqTimer[dst].Schedule(t);
		NS_LOG_LOGIC ("Scheduled RFA retry in " << t.GetSeconds () << " seconds");
		cout << "Scheduled RFA retry in " << t.GetSeconds() << " seconds"
				<< endl;
	}
}
*/
/*
void AntORDLRCore::RouteRequestTimerExpire(Ipv4Address dst) {
	NS_LOG_LOGIC (this);
	std::vector<RoutingTableEntry> toDst;
	if (m_table.LookupRoutes(dst, toDst)) {
		//SendPacketFromQueue(dst, toDst[0].GetRoute());
		NS_LOG_LOGIC ("route to " << dst << " found");
		m_addressReqTimer.erase(dst);
		return;
	}
*/
	//m_addressReqTimer.erase(dst);
	/*
	 if (toDst[0].GetRreqCnt() == RreqRetries) {
	 NS_LOG_LOGIC ("route discovery to " << dst << " has been attempted RreqRetries (" << RreqRetries << ") times");
	 cout <<"route discovery to " << dst << " has been attempted RreqRetries (" << RreqRetries << ") times\n";
	 m_addressReqTimer.erase(dst);
	 vector < uint32_t > entries;
	 m_table.LookupEntriesByDest(dst, entries);
	 m_table.DeleteRoute(dst, entries[0]);
	 NS_LOG_DEBUG ("Route not found. Drop all packets with dst " << dst);
	 m_queue.DropPacketWithDst(dst);
	 return;
	 }
	 */
	/*
	 if (toDst[0].GetFlag() == IN_SEARCH) {
	 NS_LOG_LOGIC ("Resend RFA to " << dst << " ttl " << NetDiameter);
	 cout << "Resend RFA to " << dst<<endl;
	 SendRFA(dst);
	 } else {
	 NS_LOG_DEBUG ("Route down. Stop search. Drop packet with destination " << dst);
	 m_addressReqTimer.erase(dst);
	 vector < uint32_t > entries;
	 m_table.LookupEntriesByDest(dst, entries);
	 m_table.DeleteRoute(dst, entries[0]);
	 m_queue.DropPacketWithDst(dst);
	 }

}
*/

} // namespace antordlr


} // namespace ns3


