/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2010 UCM
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Author:  Delfín Rupérez Cañas <delfinrc@fdi.ucm.es>
 *
 */
#ifndef ANTOR_DLR_CORE_H
#define ANTOR_DLR_CORE_H


#include "ns3/node.h"
#include "ns3/ipv4-routing-protocol.h"
#include "ns3/ipv4-interface.h"
#include "ns3/ipv4-l3-protocol.h"
#include "ns3/udp-l4-protocol.h"
#include "ns3/timer.h"
#include "ns3/type-id.h"
#include "ns3/udp-socket-factory.h"
#include "ns3/packet.h"
#include "ns3/ipv4.h"
#include "ns3/antor-dlr-header.h"
#include "ns3/log.h"
#include "ns3/log-file.h"
#include "ns3/ipv4-route.h"
#include "ns3/random-variable.h"
#include "ns3/inet-socket-address.h"
#include "ns3/antor-dlr-stat.h"
#include "ns3/antor-dlr-routing-table.h"
#include "ns3/antor-dlr-heuristic.h"
#include "ns3/antor-dlr-neighbor.h"
#include "ns3/antor-dlr-id-cache.h"
#include "ns3/antor-dlr-dpd.h"
#include "ns3/data-tag.h"
#include "ns3/antor-dlr-queue.h"
#include "ns3/ipv4-static-routing.h"
#include "ns3/wifi-mac-header.h"
#include "ns3/arp-cache.h"
#include "ns3/output-stream-wrapper.h"
#include "ns3/boolean.h"
#include "ns3/wifi-net-device.h"
#include "ns3/adhoc-wifi-mac.h"
//#include <iostream>
//#include <algorithm>
//#include <limits>
#include "ns3/log.h"
//#include "ns3/application.h"

namespace ns3
{
namespace antordlr
{


/**
 * \ingroup antor-dlr
 *
 * \brief ANTOR-DLR routing protocol
 */
class AntORDLRCore : public Ipv4RoutingProtocol
{
public:
  static TypeId GetTypeId (void);
  static const uint32_t ANTOR_DLR_PORT;

  /// c-tor
  AntORDLRCore ();
  virtual ~AntORDLRCore();
  virtual  void DoDispose ();
  ///\name From Ipv4RoutingProtocol
  //\{
  Ptr<Ipv4Route> RouteOutput (Ptr<Packet> p, const Ipv4Header &header, Ptr<NetDevice> oif, Socket::SocketErrno &sockerr);
  bool RouteInput (Ptr<const Packet> p, const Ipv4Header &header, Ptr<const NetDevice> idev,
                           UnicastForwardCallback ucb, MulticastForwardCallback mcb,
                           LocalDeliverCallback lcb, ErrorCallback ecb);
  virtual void NotifyInterfaceUp (uint32_t interface);
  virtual void NotifyInterfaceDown (uint32_t interface);
  virtual void NotifyAddAddress (uint32_t interface, Ipv4InterfaceAddress address);
  virtual void NotifyRemoveAddress (uint32_t interface, Ipv4InterfaceAddress address);
  virtual void SetIpv4 (Ptr<Ipv4> ipv4);

  virtual void PrintRoutingTable (Ptr<OutputStreamWrapper> stream) const;

  //\}


private:

  //\}

    ///\name Handle protocol parameters
    //\{
    Time GetMaxQueueTime () const { return MaxQueueTime; }
    void SetMaxQueueTime (Time t);
    uint32_t GetMaxQueueLen () const { return MaxQueueLen; }
    void SetMaxQueueLen (uint32_t len);
  //  bool GetGratuitousReplyFlag () const { return GratuitousReply; }
  //  void SetGratuitousReplyFlag (bool f) { GratuitousReply = f; }
    //\}
  private:
    ///\name Protocol parameters.
    //\{
     uint32_t NetDiameter;             ///< Net diameter measures the maximum possible number of hops between two nodes in the network
    /**
     *  NodeTraversalTime is a conservative estimate of the average one hop traversal time for packets
     *  and should include queuing delays, interrupt processing times and transfer times.
     */
    Time NodeTraversalTime;
    Time NetTraversalTime;             ///< Estimate of the average net traversal time.
    Time PathDiscoveryTime;            ///< Estimate of maximum time needed to find route in network.
    /**
     * Every HelloInterval the node checks whether it has sent a broadcast  within the last HelloInterval.
     * If it has not, it MAY broadcast a  Hello message
     */
    Time m_helloInterval;
    uint32_t AllowedHelloLoss;         ///< Number of hello messages which may be loss for valid link
    uint32_t MaxQueueLen;              ///< The maximum number of packets that we allow a routing protocol to buffer.
    Time MaxQueueTime;                 ///< The maximum period of time that a routing protocol is allowed to buffer a packet for.
    bool EnableBroadcast;              ///< Indicates whether a a broadcast data packets forwarding enable

    // Proactive Forward Ant messages' emission interval
    Time m_PFAInterval;

    // Maximum number of Report destination
    uint32_t num_report_destination;

    uint32_t max_broadcast_allowed;
    //\}

    /// IP protocol
    Ptr<Ipv4> m_ipv4;
    /// Raw socket per each IP interface, map socket -> iface address (IP + mask)
    std::map< Ptr<Socket>, Ipv4InterfaceAddress > m_socketAddresses;
    /// Loopback device used to defer RREQ until packet will be fully formed
    Ptr<NetDevice> m_lo;

    // Calculate Heuristic
    AntORHeuristic m_heuristic;
    /// Routing table
    RoutingTable m_table;
    /// A "drop-front" queue used by the routing layer to buffer packets to which it does not have a route.
    RequestQueue m_queue;
    /// Local Hop Number in Repair Process
    uint32_t m_local_hops;
    /// Handle duplicated RREQ
    IdCache m_rreqIdCache;
    /// Handle duplicated broadcast/multicast packets
    DuplicatePacketDetection m_dpd;
    /// Handle neighbors
    Neighbors m_nbs;

    uint32_t m_Application_Port;

    /// Messages sequence number counter.
    uint16_t m_messageSequenceNumber;


    AntORStat m_stat;


private:

    void Start ();


    Ptr<Ipv4Route>
      LoopbackRoute (const Ipv4Header & hdr, Ptr<NetDevice> oif);

    void
      DeferredRouteOutput (Ptr<const Packet> p, const Ipv4Header & header,
      UnicastForwardCallback ucb, ErrorCallback ecb);

    Ptr<Socket>
       FindSocketWithInterfaceAddress (Ipv4InterfaceAddress addr ) const;

    /// Increments message sequence number and returns the new value.
    uint16_t GetMessageSequenceNumber ();

    bool IsMyOwnAddress (Ipv4Address src);

    bool FindCopies(Ipv4Address dir1, std::vector<Ipv4Address> dirs, uint32_t& idx);


    void UpdateRoutingNeighbor(Ipv4Address addrRecv,Ipv4Address addr, Time time,
    		                   uint32_t hops, ns3::antordlr::MessageHeader::Hello h);

    void UpdateNeighbor(Ipv4Address receiverAddr,Ipv4Address dst, Time t);


    void UpdateRoutingTable(Ipv4Address receiverAddr,
    		                   Ipv4Address senderAddr, uint32_t hops, uint16_t seqno);

    bool Forwarding (Ptr<const Packet> p, const Ipv4Header & header,
                                 UnicastForwardCallback ucb, ErrorCallback ecb, Ipv4Address src);


    ///\name Receive control packets
     //\{
     /// Receive and process control packet
    void RecvAntORDLR (Ptr<Socket> socket);

    void ProcessHello (ns3::antordlr::MessageHeader &msg,
                               const Ipv4Address &receiverAddr,
                               const Ipv4Address &senderAddr);



            void ProcessRFA ( ns3::antordlr::MessageHeader &msg,
            		          const Ipv4InterfaceAddress &receiverIface,
                              const Ipv4Address &senderAddr);


            void ProcessRBA ( ns3::antordlr::MessageHeader &msg,
                              const Ipv4InterfaceAddress &receiverIface,
                              const Ipv4Address &senderAddr);

            void ProcessPFA ( ns3::antordlr::MessageHeader &msg,
                              const Ipv4Address &receiverAddr,
                              const Ipv4Address &senderAddr);

            void ProcessPBA ( ns3::antordlr::MessageHeader &msg,
                              const Ipv4Address &receiverAddr,
                              const Ipv4Address &senderAddr);

            void ProcessLFN ( ns3::antordlr::MessageHeader &msg,
                              const Ipv4Address &receiverAddr,
                              const Ipv4Address &senderAddr);


            void ProcessRRFA ( ns3::antordlr::MessageHeader &msg,
            		           const Ipv4InterfaceAddress &receiverIface,
                               const Ipv4Address &senderAddr);

            void ProcessRRBA ( ns3::antordlr::MessageHeader &msg,
            		     const Ipv4InterfaceAddress &receiverIface,
                               const Ipv4Address &senderAddr);

            void ProcessULN ( ns3::antordlr::MessageHeader &msg,
                                const Ipv4Address &receiverAddr,
                                         const Ipv4Address &senderAddr);

      //\}





    ///\name Send
    //\{
    /// Forward packet from route request queue
    void SendMessageWhenBreaksLinkToNextHop(Ipv4Address addr);

    void SendLFNWhenNoRouteToForward (Ipv4Address dst, Ipv4Address origin);

    void SendPacketFromQueue (Ipv4Address dst, Ptr<Ipv4Route> route);

    void SendPacketFromQueueInIntervals(Ipv4Address dst, Ptr<Ipv4Route> route);

    void SendHello ();

    void SendLFN (std::vector<Ipv4Address> pre, std::vector<Ipv4Address> unDestination );

    void ReSendLFN (Ipv4Address dir,ns3::antordlr::MessageHeader &msg);

    void SendRFA (Ipv4Address dst);

        void SendRBA (Ipv4InterfaceAddress srcIface,Ipv4Address dst, std::vector<Ipv4Address> dirs);

        void SendPFA (Ipv4Address dst);

        void SendPBA (Ipv4Address addr,std::vector<Ipv4Address> dirs);

        void SendRBAByIntermediateNode (Ipv4InterfaceAddress srcIface,Ipv4Address dir, Ipv4Address dst, std::vector<Ipv4Address> dirs);

        void SendRRFA ( Ipv4Address dst, uint8_t hops);

        void SendRRBA ( Ipv4InterfaceAddress srcIface,Ipv4Address dst, std::vector<Ipv4Address> dirs,uint32_t hops);

        void SendULN ( Ipv4Address unDestination, Ipv4Address nextHop );
        //\}

    // Timer handlers

    /// Hello timer
    Timer m_helloTimer;

    /// Schedule next send of hello message
    void HelloTimerExpire ();

    Timer m_LFNTimer;

    void LFNTimerExpire();

    Timer m_PFATimer;

    void PFATimerExpire(Ipv4Address dir);

    Timer m_RRFATimer;

    void RRFATimerExpire(Ipv4Address dir);

    void RFATimerExpire(Ipv4Address dst);

    Timer m_RFATimer;

    Timer m_rreqRateLimitTimer;

    void RreqRateLimitTimerExpire ();

    /// RLR rate limit timer
    Timer m_RRFARateLimitTimer;
    //Reset RERR count and schedule RERR rate limit timer with delay 1 sec.
    void RRFARateLimitTimerExpire ();
/*
    /// Map IP address + RREQ timer.
     std::map<Ipv4Address, Timer> m_addressReqTimer;
    /// Handle route discovery process
     void RouteRequestTimerExpire (Ipv4Address dst);
*/
  };

}
}
#endif /* ANTOR_DLR_CORE_H */
