/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2010 UCM
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Author:  Delfín Rupérez Cañas <delfinrc@fdi.ucm.es>
 *          
 */


#ifndef ANTOR_DLR_HEADER_H
#define ANTOR_DLR_HEADER_H

#include "ns3/header.h"
#include "ns3/ipv4-address.h"
#include "ns3/nstime.h"


namespace ns3 {

namespace antordlr {


//        0                   1                   2                   3
//        0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//       |  Message Type | Num Broadcast |         Message Size          |
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//       |  Time To Live |   Hop Count   |    Message Sequence Number    |
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

class MessageHeader : public Header {
public:

    enum MessageType {
        HELLO_MESSAGE = 1,
        RFA_MESSAGE = 2,
        RBA_MESSAGE = 3,
        PFA_MESSAGE = 4,
        PBA_MESSAGE = 5,
        LFN_MESSAGE = 6,
        RRFA_MESSAGE = 7,
        RRBA_MESSAGE = 8,
        ULN_MESSAGE = 9,
    };

    MessageHeader ();
    virtual ~MessageHeader ();

    void SetMessageType (MessageType messageType) {
        m_messageType = messageType;
    }
    MessageType GetMessageType () const {
        return m_messageType;
    }

    void SetTimeToLive (uint8_t timeToLive) {
        m_timeToLive = timeToLive;
    }
    uint8_t GetTimeToLive () const {
        return m_timeToLive;
    }

    void SetHopCount (uint8_t hopCount) {
        m_hopCount = hopCount;
    }
    uint8_t GetHopCount () const {
        return m_hopCount;
    }

    void SetNumBroadcast (uint8_t num) {
        m_numBroadcast = num;
    }
    uint8_t GetNumBroadcast () const {
        return m_numBroadcast;
    }

    void SetMessageSequenceNumber (uint16_t messageSequenceNumber) {
        m_messageSequenceNumber = messageSequenceNumber;
    }
    uint16_t GetMessageSequenceNumber () const {
        return m_messageSequenceNumber;
    }


private:
    MessageType m_messageType;
    uint8_t m_timeToLive;
    uint8_t m_hopCount;
    uint8_t m_numBroadcast;
    uint16_t m_messageSequenceNumber;
    uint16_t m_messageSize;

public:
    static TypeId GetTypeId (void);
    virtual TypeId GetInstanceTypeId (void) const;
    virtual void Print (std::ostream &os) const;
    virtual uint32_t GetSerializedSize (void) const;
    virtual void Serialize (Buffer::Iterator start) const;
    virtual uint32_t Deserialize (Buffer::Iterator start);


    // 1.  HELLO Message Format
    //
    //        0                   1                   2                   3
    //        0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
    //       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    //       |                             DELAY                             |
    //       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    //       |   		  DESTINATION ADDRESS 1                              |
    //       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    //       |                    BEST PHEROMONE VALUE 1                     |
    //       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    //       |                             ...                               |
    //       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    //       |                    DESTINATION ADDRESS N                      |
    //       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    //       |                    BEST PHEROMONE VALUE N                     |
    //       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

    struct Hello {

        struct DestinationValue {
            Ipv4Address address;
            float best;
        };


        float delay;
        void SetDelay (float time) {
            this->delay = time;
        }
        float GetDelay () const {
            return this-> delay;
        }
        std::vector<DestinationValue> destValue;

        void Print (std::ostream &os) const;
        uint32_t GetSerializedSize (void) const;
        void Serialize (Buffer::Iterator start) const;
        uint32_t Deserialize (Buffer::Iterator start, uint32_t messageSize);
    };

    // 2.  REACTIVE FORWARD ANT (RFA) Message Format
    //
    //        0                   1                   2                   3
    //        0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
    //       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    //       |                    SOURCE ADDRESS SESSION                     |
    //       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    //       |   	        DESTINATION ADDRESS SESSION                  |
    //       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    //       |                          NODE LIST 1                          |
    //       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    //       |                             ...                               |
    //       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    //       |                          NODE LIST N                          |
    //       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

    struct Rfa {

        struct AddressSession {
            Ipv4Address source;
            Ipv4Address destination;
        };

        AddressSession addrSession;
        std::vector<Ipv4Address> listP;

        void Print (std::ostream &os) const;
        uint32_t GetSerializedSize (void) const;
        void Serialize (Buffer::Iterator start) const;
        uint32_t Deserialize (Buffer::Iterator start, uint32_t messageSize);
    };

    // 3.  REACTIVE BACKWARD ANT (RBA) Message Format
    //
    //        0                   1                   2                   3
    //        0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
    //
    //       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    //       |                    SOURCE ADDRESS SESSION                     |
    //       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    //       |   	            DESTINATION ADDRESS SESSION                  |
    //       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    //       |                       END-TO-END DELAY                        |
    //       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    //       |                          NODE LIST 1                          |
    //       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    //       |                             ...                               |
    //       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    //       |                          NODE LIST N                          |
    //       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

    struct Rba {

        struct AddressSession {
            Ipv4Address source;
            Ipv4Address destination;
        };
        float delay;

        void SetDelay (float time) {
            this->delay = time;
        }
        float GetDelay () const {
            return this-> delay;
        }
        AddressSession addrSession;
        std::vector<Ipv4Address> listP;

        void Print (std::ostream &os) const;
        uint32_t GetSerializedSize (void) const;
        void Serialize (Buffer::Iterator start) const;
        uint32_t Deserialize (Buffer::Iterator start, uint32_t messageSize);
    };

    // 4.  PROACTIVE FORWARD ANT (PFA) Message Format
    //
    //        0                   1                   2                   3
    //        0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
    //
    //       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    //       |                    SOURCE ADDRESS SESSION                     |
    //       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    //       |   	            DESTINATION ADDRESS SESSION                  |
    //       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    //       |                          NODE LIST 1                          |
    //       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    //       |                             ...                               |
    //       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    //       |                          NODE LIST N                          |
    //       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

    struct Pfa {

        struct AddressSession {
            Ipv4Address source;
            Ipv4Address destination;
        };

        AddressSession addrSession;
        std::vector<Ipv4Address> listP;

        void Print (std::ostream &os) const;
        uint32_t GetSerializedSize (void) const;
        void Serialize (Buffer::Iterator start) const;
        uint32_t Deserialize (Buffer::Iterator start, uint32_t messageSize);
    };

    // 5.  PROACTIVE BACKWARD ANT (PBA) Message Format
    //
    //        0                   1                   2                   3
    //        0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
    //
    //       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    //       |                    SOURCE ADDRESS SESSION                     |
    //       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    //       |   		        DESTINATION ADDRESS SESSION                  |
    //       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    //       |                       END-TO-END DELAY                        |
    //       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    //       |                          NODE LIST 1                          |
    //       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    //       |                             ...                               |
    //       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    //       |                          NODE LIST N                          |
    //       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

    struct Pba {

        struct AddressSession {
            Ipv4Address source;
            Ipv4Address destination;
        };
        float delay;

        void SetDelay (float time) {
            this->delay = time;
        }
        float GetDelay () const {
            return this-> delay;
        }
        AddressSession addrSession;
        std::vector<Ipv4Address> listP;

        void Print (std::ostream &os) const;
        uint32_t GetSerializedSize (void) const;
        void Serialize (Buffer::Iterator start) const;
        uint32_t Deserialize (Buffer::Iterator start, uint32_t messageSize);

    };

    // 6.  LINK FAILURE NOTIFICATION (LFN) Message Format
    //
    //        0                   1                   2                   3
    //        0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
    //
    //       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    //       |               RESERVED	    |  DESTINATION ADDRESS LENGTH    |
    //       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    //       |                        ORIGINATOR ADDRESS                     |
    //       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    //       |                     DESTINATION ADDRESS 1                     |
    //       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    //       |                             ...                               |
    //       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    //       |                     DESTINATION ADDRESS N                     |
    //       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

    struct Lfn {

        uint16_t reserved;
    	uint16_t ldest;

    	Ipv4Address originator;

        void SetLdest (uint16_t ld) {
            this->ldest = ld;
        }
        uint16_t GetLdest () const {
            return this-> ldest;
        }

        std::vector<Ipv4Address> dest;

        void Print (std::ostream &os) const;
        uint32_t GetSerializedSize (void) const;
        void Serialize (Buffer::Iterator start) const;
        uint32_t Deserialize (Buffer::Iterator start, uint32_t messageSize);
    };

    // 7.  ROUTE REPAIR FORWARD ANT (RRFA) Message Format
    //
    //        0                   1                   2                   3
    //        0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
    //
    //       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    //       |                    SOURCE ADDRESS SESSION                     |
    //       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    //       |   	        DESTINATION ADDRESS SESSION                  |
    //       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    //       |                          NODE LIST 1                          |
    //       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    //       |                             ...                               |
    //       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    //       |                          NODE LIST N                          |
    //       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

    struct Rrfa {

        struct AddressSession {
            Ipv4Address source;
            Ipv4Address destination;
            Ipv4Address local;
        };

        AddressSession addrSession;
        std::vector<Ipv4Address> listP;

        void Print (std::ostream &os) const;
        uint32_t GetSerializedSize (void) const;
        void Serialize (Buffer::Iterator start) const;
        uint32_t Deserialize (Buffer::Iterator start, uint32_t messageSize);
    };

    // 8.  ROUTE REPAIR BACKWARD ANT (RRBA) Message Format
    //
    //        0                   1                   2                   3
    //        0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
    //
    //       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    //       |                    SOURCE ADDRESS SESSION                     |
    //       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    //       |   	        DESTINATION ADDRESS SESSION                  |
    //       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    //       |                       END-TO-END DELAY                        |
    //       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    //       |                          NODE LIST 1                          |
    //       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    //       |                             ...                               |
    //       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    //       |                          NODE LIST N                          |
    //       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

    struct Rrba {

        struct AddressSession {
            Ipv4Address source;
            Ipv4Address destination;

        };
        float delay;

        void SetDelay (float time) {
            this->delay = time;
        }
        float GetDelay () const {
            return this-> delay;
        }
        AddressSession addrSession;
        std::vector<Ipv4Address> listP;

        void Print (std::ostream &os) const;
        uint32_t GetSerializedSize (void) const;
        void Serialize (Buffer::Iterator start) const;
        uint32_t Deserialize (Buffer::Iterator start, uint32_t messageSize);
    };

       // 9.  UNICAST LINK NOTIFICATION MESSAGE (ULN) Message Format
       //
       //        0                   1                   2                   3
       //        0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
       //
       //       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
       //       |   	            DESTINATION ADDRESS SESSION                 |
       //       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
       //       |   	               SOURCE   ADDRESS SESSION                 |
       //       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+


       struct Uln {


           Ipv4Address dst;

           Ipv4Address src;

           void Print (std::ostream &os) const;
           uint32_t GetSerializedSize (void) const;
           void Serialize (Buffer::Iterator start) const;
           uint32_t Deserialize (Buffer::Iterator start, uint32_t messageSize);
       };

private:
    struct {
        ///////////////
        Rfa rfa;
        Rba rba;
        ///////////////
        Hello hello;
        Pfa pfa;
        Pba pba;
        ///////////////
        Lfn lfn;
        Rrfa rrfa;
        Rrba rrba;

        Uln uln;

    }
    m_message;

public:

    Hello& GetHello () {
        if (m_messageType == 0) {
            m_messageType = HELLO_MESSAGE;
        } else {
            NS_ASSERT (m_messageType == HELLO_MESSAGE);
        }
        return m_message.hello;
    }


    const Hello& GetHello () const {
        NS_ASSERT (m_messageType == HELLO_MESSAGE);
        return m_message.hello;
    }

    Rfa& GetRfa () {
        if (m_messageType == 0) {
            m_messageType = RFA_MESSAGE;
        } else {
            NS_ASSERT (m_messageType == RFA_MESSAGE);
        }
        return m_message.rfa;
    }


    const Rfa& GetRfa () const {
        NS_ASSERT (m_messageType == RFA_MESSAGE);
        return m_message.rfa;
    }

    Rba& GetRba () {
        if (m_messageType == 0) {
            m_messageType = RBA_MESSAGE;
        } else {
            NS_ASSERT (m_messageType == RBA_MESSAGE);
        }
        return m_message.rba;
    }


    const Rba& GetRba () const {
        NS_ASSERT (m_messageType == RBA_MESSAGE);
        return m_message.rba;
    }


    Pfa& GetPfa () {
        if (m_messageType == 0) {
            m_messageType = PFA_MESSAGE;
        } else {
            NS_ASSERT (m_messageType == PFA_MESSAGE);
        }
        return m_message.pfa;
    }


    const Pfa& GetPfa () const {
        NS_ASSERT (m_messageType == PFA_MESSAGE);
        return m_message.pfa;
    }



    Pba& GetPba () {
        if (m_messageType == 0) {
            m_messageType = PBA_MESSAGE;
        } else {
            NS_ASSERT (m_messageType == PBA_MESSAGE);
        }
        return m_message.pba;
    }


    const Pba& GetPba () const {
        NS_ASSERT (m_messageType == PBA_MESSAGE);
        return m_message.pba;
    }

    Lfn& GetLfn () {
        if (m_messageType == 0) {
            m_messageType = LFN_MESSAGE;
        } else {
            NS_ASSERT (m_messageType == LFN_MESSAGE);
        }
        return m_message.lfn;
    }


    const Lfn& GetLfn () const {
        NS_ASSERT (m_messageType == LFN_MESSAGE);
        return m_message.lfn;
    }

    Rrfa& GetRrfa () {
        if (m_messageType == 0) {
            m_messageType = RRFA_MESSAGE;
        } else {
            NS_ASSERT (m_messageType == RRFA_MESSAGE);
        }
        return m_message.rrfa;
    }


    const Rrfa& GetRrfa () const {
        NS_ASSERT (m_messageType == RRFA_MESSAGE);
        return m_message.rrfa;
    }

    Rrba& GetRrba () {
        if (m_messageType == 1) {
            m_messageType = RRBA_MESSAGE;
        } else {
            NS_ASSERT (m_messageType == RRBA_MESSAGE);
        }
        return m_message.rrba;
    }


    const Rrba& GetRrba () const {
        NS_ASSERT (m_messageType == RRBA_MESSAGE);
        return m_message.rrba;
    }


    const Uln& GetUln () const {
        NS_ASSERT (m_messageType == ULN_MESSAGE);
        return m_message.uln;
    }


};

static inline std::ostream& operator<< (std::ostream& os, const MessageHeader & message) {
    message.Print (os);
    return os;
}

}  // namespace antordlr

} // namespace ns3

#endif /* ANTOR_DNR_HEADER_H */

