/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2010 UCM
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Author:  Delfín Rupérez Cañas <delfinrc@fdi.ucm.es>
 *          
 */
#include "antor-dlr-routing-table.h"

NS_LOG_COMPONENT_DEFINE ("AntORDNRRoutingTable");

namespace ns3
{
namespace antordlr
{

/*
 The Routing Table
 */

RoutingTableEntry::RoutingTableEntry (Ptr<NetDevice> dev, Ipv4Address dst,Ipv4InterfaceAddress iface,
		                              uint16_t hops, Ipv4Address nextHop,PheroTable pherot,bool b) :
                                      m_hops (hops),
                                      m_iface (iface),
                                      m_pheroTable(pherot),
                                      m_visitedlink(b)
{
  m_ipv4Route = Create<Ipv4Route> ();
  m_ipv4Route->SetDestination (dst);
  m_ipv4Route->SetGateway (nextHop);
  m_ipv4Route->SetSource (m_iface.GetLocal ());
  m_ipv4Route->SetOutputDevice (dev);
}

RoutingTableEntry::~RoutingTableEntry ()
{
}

bool
RoutingTableEntry::InsertPrecursor (Ipv4Address id)
{
  NS_LOG_FUNCTION (this << id);
  if (!LookupPrecursor (id))
    {
      m_precursorList.push_back (id);
      return true;
    }
  else
    return false;
}

bool
RoutingTableEntry::LookupPrecursor (Ipv4Address id)
{
  NS_LOG_FUNCTION (this << id);
  for (std::vector<Ipv4Address>::const_iterator i = m_precursorList.begin (); i
       != m_precursorList.end (); ++i)
    {
      if (*i == id)
        {
          NS_LOG_LOGIC ("Precursor " << id << " found");
          return true;
        }
    }
  NS_LOG_LOGIC ("Precursor " << id << " not found");
  return false;
}

bool
RoutingTableEntry::DeletePrecursor (Ipv4Address id)
{
  NS_LOG_FUNCTION (this << id);
  std::vector<Ipv4Address>::iterator i = std::remove (m_precursorList.begin (),
                                                      m_precursorList.end (), id);
  if (i == m_precursorList.end ())
    {
      NS_LOG_LOGIC ("Precursor " << id << " not found");
      return false;
    }
  else
    {
      NS_LOG_LOGIC ("Precursor " << id << " found");
      m_precursorList.erase (i, m_precursorList.end ());
    }
  return true;
}

void
RoutingTableEntry::DeleteAllPrecursors ()
{
  NS_LOG_FUNCTION (this);
  m_precursorList.clear ();
}

bool
RoutingTableEntry::IsPrecursorListEmpty () const
{
  return m_precursorList.empty ();
}

void
RoutingTableEntry::GetPrecursors (std::vector<Ipv4Address> & prec) const
{
  NS_LOG_FUNCTION (this);
  if (IsPrecursorListEmpty ())
    return;
  for (std::vector<Ipv4Address>::const_iterator i = m_precursorList.begin (); i
       != m_precursorList.end (); ++i)
    {
      bool result = true;
      for (std::vector<Ipv4Address>::const_iterator j = prec.begin (); j
           != prec.end (); ++j)
        {
          if (*j == *i)
            result = false;
        }
      if (result)
        prec.push_back (*i);
    }
}



//
// The Routing Table
//


void
RoutingTable::DeleteAllRoutesFromInterface (Ipv4InterfaceAddress iface)
{
  NS_LOG_FUNCTION (this);
  if (m_table.empty ())
    return;
  for (std::multimap<Ipv4Address, RoutingTableEntry>::iterator i =
         m_table.begin (); i != m_table.end ();)
    {
      if (i->second.GetInterface () == iface)
        {
          std::multimap<Ipv4Address, RoutingTableEntry>::iterator tmp = i;
          ++i;
          m_table.erase (tmp);
        }
      else
        ++i;
    }
}

bool
RoutingTable::AddRoute (RoutingTableEntry & rt)
{
  NS_LOG_FUNCTION (this);
  //FIXME
  Purge ();
  std::multimap<Ipv4Address, RoutingTableEntry>::iterator it;
  it = m_table.insert (std::make_pair (rt.GetDestination (), rt));
  if (it == m_table.end())
   return false;
  else
   return true;

}

bool
RoutingTable::DeleteRoute (Ipv4Address dest, uint32_t idx)
{
  NS_LOG_FUNCTION (this << idx);
  //FIXME
  Purge ();
  if (m_table.empty())
   return false;
  uint32_t cont_idx = 0;
  for (std::multimap<Ipv4Address, RoutingTableEntry>::iterator i =
           m_table.begin(); i != m_table.end ();++i)
      {
        if (cont_idx == idx)
	      {
            m_table.erase (i);
           return true;
          }
        cont_idx++;
       }

  return false;
}
RoutingTableEntry RoutingTable::GetRoutingEntry(uint32_t idx){
 NS_LOG_FUNCTION (this << idx);
 Purge();
 uint32_t temp_idx = 0;
 RoutingTableEntry routingTableEntry;
 for (std::multimap<Ipv4Address, RoutingTableEntry>::iterator i =
           m_table.begin(); i != m_table.end ();++i)
      {
	    if (idx == temp_idx){
	     //return i->second;
	      routingTableEntry = i->second;
	      return routingTableEntry;
	    }
	    temp_idx++;

      }
 return routingTableEntry;
}

bool
RoutingTable::GetIndexRoute (Ipv4Address dst, Ipv4Address next, uint32_t& idx)
{
  NS_LOG_FUNCTION (this << dst);
  //FIXME
  Purge ();

  if (m_table.empty())
     return false;
   uint32_t cont_idx = 0;
   for (std::multimap<Ipv4Address, RoutingTableEntry>::iterator i =
           m_table.find(dst); i != m_table.end ();++i)
      {
       if (i->second.GetNextHop() == next)
        {
          idx = cont_idx;
          return true;
        }
       cont_idx++;
      }

  return false;
}


bool
RoutingTable::LookupAllRoutes (Ipv4Address id, std::vector<RoutingTableEntry>&  rts)
{
  NS_LOG_FUNCTION (this << id);

  //FIXME
  Purge ();

  if (m_table.empty ())
    {
     return false;
    }

  rts.clear();
  for (std::multimap<Ipv4Address, RoutingTableEntry>::iterator it =
             m_table.begin(); it != m_table.end ();++it)
        {
	     if (it->first == id){
	    	rts.push_back(it->second);

	     }
        }

  if (rts.size () > 0)
		  return true;
	  else
		  return false;


  }


bool
RoutingTable::LookupRoutesWithReg (Ipv4Address id, std::vector<RoutingTableEntry>&  results)
{
      NS_LOG_FUNCTION (this << id);

	  //FIXME
	  Purge ();
	   if (m_table.empty ())
	    {
	     return false;
	    }

	  std::vector<RoutingTableEntry> rts;
	  bool r = LookupAllRoutes(id,rts);
	  if (!r)
	   return false;
	  bool res = false;
      for (uint32_t k=0; k < rts.size(); k++)
	        {
		     if (rts[k].GetPheroTable().pheroReg > 0.0){
                results.push_back(rts[k]);
                res = true;
		     }
		     }
	  return res;

}


bool
RoutingTable::Update (RoutingTableEntry & rt, uint32_t idx)
{
  //NS_LOG_FUNCTION (this << rt << idx);

   //FIXME
    Purge ();
    for (std::multimap<Ipv4Address, RoutingTableEntry>::iterator it =
            m_table.begin(); it != m_table.end ();++it)
       {
         if (rt.GetDestination() == it->first && rt.GetNextHop() == it->second.GetNextHop()){
        	  it->second = rt;
             return true;
             }
           }



   return false;

}

bool
RoutingTable::UpdateNew (RoutingTableEntry & rt)
{
  //NS_LOG_FUNCTION (this << rt << idx);

   //FIXME
    Purge ();
    for (std::multimap<Ipv4Address, RoutingTableEntry>::iterator it =
            m_table.begin(); it != m_table.end ();++it)
       {
         if (rt.GetDestination() == it->first && rt.GetNextHop() == it->second.GetNextHop()){
        	  it->second = rt;
             return true;
             }
           }



   return false;

}

void
RoutingTable::Purge ()
{
  NS_LOG_FUNCTION (this);

 if (m_table.empty ())
   return;

  for (std::multimap<Ipv4Address, RoutingTableEntry>::iterator i =
		  m_table.begin (); i != m_table.end ();)
    {
     double vals = std::max(i->second.GetPheroTable().pheroReg,i->second.GetPheroTable().pheroVir);
		  if ( vals == 0.0 ) //|| i->second.GetNextHop() == Ipv4Address())
            {
        	  //std::cout <<"[ " << Simulator::Now().GetSeconds() << " sg. ] Erasing route with destination address " << i->first<<std::endl;
        	  std::multimap<Ipv4Address, RoutingTableEntry>::iterator tmp = i;
              ++i;
             // m_table.erase (tmp);
            }

      else
        {
          ++i;
        }
    }

}


std::vector<Ipv4Address>
RoutingTable::GetAllDestination(uint32_t max){

	std::vector<Ipv4Address> listDestination;
	listDestination.clear();
	if (m_table.empty ())
	    return listDestination;

	for (std::multimap<Ipv4Address, RoutingTableEntry>::iterator i =
			  m_table.begin (); i != m_table.end ();++i)
	    {
		//double vals = std::max(i->second.GetPheroTable().pheroReg,i->second.GetPheroTable().pheroVir);

		 if (!FindDestinationAux(i->first,listDestination) && i->second.GetHop() > 1 && i->first != Ipv4Address::GetLoopback()
		                                                   && i->first != Ipv4Address::GetBroadcast()) {
		  listDestination.push_back(i->first);

	     }
	    }
	if (listDestination.size() <= max){
	  return listDestination;

	}
	NS_ASSERT(listDestination.size() > max);


	RandomVariable rv;
	while (listDestination.size() > max) {
		rv = RandomVariable(UniformVariable(0, listDestination.size() - 1));
        listDestination.erase(listDestination.begin() + (uint32_t) floor(rv.GetValue() ));

	}
	NS_ASSERT_MSG (listDestination.size() == max, "invalid maximum of list destination");
	return listDestination;

}

bool RoutingTable::FindDestinationAux(Ipv4Address dest, std::vector<Ipv4Address> temp)
{

   for (uint32_t i = 0; i < temp.size(); i++)
   {
			if (temp[i] == dest)
				return true;
   }
		return false;
}

bool
RoutingTable::LookupEntriesByDest (Ipv4Address dest,std::vector<uint32_t> &idxs) const {

	if (m_table.size() == 0)
		return false;
	uint32_t cont_idx = 0;
	for (std::multimap<Ipv4Address, RoutingTableEntry>::const_iterator it =
		         m_table.begin(); it != m_table.end (); ++it)
	  {
	      if (it->first == dest){
		   idxs.push_back(cont_idx);
	      }
		   cont_idx++;

	  }
 if (idxs.size() > 0)
	 return true;
 else
	 return false;
}
bool RoutingTable::LookupEntriesByNextHop (Ipv4Address next, std::vector<uint32_t> &idx) const {
	if (m_table.size() == 0)
			return false;
		uint32_t cont_idx = 0;
		for (std::multimap<Ipv4Address, RoutingTableEntry>::const_iterator it =
			         m_table.begin(); it != m_table.end (); ++it)
		  {
		    if (it->second.GetNextHop() == next){
			  idx.push_back(cont_idx);

		    }
		   cont_idx++;
		  }
	if (idx.size() == 0)
	 return false;
	else
 	 return true;
}

bool RoutingTable::LookupEntryByDestAndNextHop (Ipv4Address dst, Ipv4Address next, uint32_t &idx) const {
	if (m_table.size() == 0)
			return false;
		uint32_t cont_idx = 0;
		for (std::multimap<Ipv4Address, RoutingTableEntry>::const_iterator it =
			         m_table.begin(); it != m_table.end (); ++it)
		  {
		    if (it->first == dst && it->second.GetNextHop() == next){
			  idx = cont_idx;
              return true;
		    }
		   cont_idx++;
		  }
	 return false;

}


////////////////////////////////////////////////////////////////////
std::vector<Ipv4Address> RoutingTable::GetDestinationLFN(Ipv4Address nexthop){
std::vector<Ipv4Address> res;

   for (std::multimap<Ipv4Address, RoutingTableEntry>::iterator it =
	         m_table.begin(); it != m_table.end (); ++it)
	    {
	  if ( it->second.GetNextHop() == nexthop && it->second.GetVisitedLink() == true) {
         res.push_back(it->first);
   }
  }
return res;
}
////////////////////////////////////////////////////////////////////
std::vector<Ipv4Address> RoutingTable::GetSources(Ipv4Address dst, Ipv4Address nexthop){
std::vector<Ipv4Address> res;

   for (std::multimap<Ipv4Address, RoutingTableEntry>::iterator it =
	         m_table.begin(); it != m_table.end (); ++it)
	    {
	  if ( it->second.GetNextHop() == nexthop && it->first == dst ) {
         //res.push_back(it->second.GetRoute()->GetSource());
		  it->second.GetPrecursors(res);
          return res;
	  }
  }
return res;
}

//////////////////
bool RoutingTable::LookupMinHop(Ipv4Address dest, RoutingTableEntry& rt) {
//	 NS_LOG_FUNCTION (this << dest << rt);
	  if (m_table.empty ())
	    return false;
	  uint32_t maxHop = 255;
	  for (std::multimap<Ipv4Address, RoutingTableEntry>::iterator i =
	         m_table.find(dest); i != m_table.end (); ++i)
	    {
          if (i->first == dest){
        	if (i->second.GetHop() < maxHop){
        		maxHop = i->second.GetHop();
        		rt = i->second;
        	}
          }
	    }
	return true;
}





} // namespace antordlr
} // namespace ns3
