/*
 * Copyright 2012 IDSIA
 *
 *
 * Author:  Michal Kudelski
 *
 * (based on packet-sink.h and .cc)
 *
 */

#include "ns3/address.h"
#include "ns3/address-utils.h"
#include "ns3/log.h"
#include "ns3/uinteger.h"
#include "ns3/inet-socket-address.h"
#include "ns3/node.h"
#include "ns3/socket.h"
#include "ns3/udp-socket.h"
#include "ns3/simulator.h"
#include "ns3/socket-factory.h"
#include "ns3/packet.h"
#include "ns3/trace-source-accessor.h"
#include "ns3/udp-socket-factory.h"
#include "ns3/wifi-net-device.h"
#include "ns3/regular-wifi-mac.h"
#include "ns3/PhySignalTag.h"
#include "ns3/MacDelaysTag.h"
#include "ns3/MacQueueTag.h"
#include "ns3/vector.h"
#include "ns3/ipv4-l3-protocol.h"
#include "ns3/llc-snap-header.h"
#include "ns3/string.h"
#include "MAPPLE-app.h"

using namespace std;

namespace ns3 {

// initialize static members
FILE* MappleApplication::fp = 0;				// file for writing samples
FILE* MappleApplication::fp_samples_geo = 0;	// file for writing geographical coordinates related to samples
FILE* MappleApplication::fp_mobility = 0;		// file for the whole mobility trace
FILE* MappleApplication::fp_processed = 0;		// file for writing processed samples
uint32_t MappleApplication::sampleID = 0;		// unique sample's ID
double MappleApplication::errSum = 0.0;
uint32_t MappleApplication::errCounter = 0;

uint32_t* MappleApplication::classCounter = NULL;

std::string MappleApplication::modelName = "";
double* MappleApplication::featuremax = NULL;
double* MappleApplication::featuremin = NULL;
bool MappleApplication::useDistance = true;
bool MappleApplication::useRSSI = true;
bool MappleApplication::useSNR = true;

NS_LOG_COMPONENT_DEFINE ("MAPPLE");
NS_OBJECT_ENSURE_REGISTERED (MappleApplication);

TypeId
MappleApplication::GetTypeId (void)
{
  static TypeId tid = TypeId ("ns3::MappleApplication")
    .SetParent<Application> ()
    .AddConstructor<MappleApplication> ()
    .AddAttribute("port", "Port number on which the ns3 <-> ns3 socket comunicates",
	    UintegerValue(9),
	    MakeUintegerAccessor(&MappleApplication::portBase),
	    MakeUintegerChecker<uint16_t>(1))
    .AddAttribute ("Protocol", "The type id of the protocol to use for the rx socket.",
                   TypeIdValue (UdpSocketFactory::GetTypeId ()),
                   MakeTypeIdAccessor (&MappleApplication::m_tid),
                   MakeTypeIdChecker ())
    .AddTraceSource ("Rx", "A packet has been received",
                    MakeTraceSourceAccessor (&MappleApplication::m_rxTrace))
    .AddTraceSource ("Tx", "A new packet is created and is sent",
					MakeTraceSourceAccessor (&MappleApplication::m_txTrace))
  ;
  return tid;
}

MappleApplication::MappleApplication ()
{
  NS_LOG_FUNCTION (this);
  m_socket = 0;
  sendSocket = 0;
  m_totalRx = 0;

  outFilePath = "samples";

  ////////////////////////////////////////////////////////////////////////////////////////////////////
  // initialize DATA RELATED TO MAPPLE
  ////////////////////////////////////////////////////////////////////////////////////////////////////
	my_counter = 0;
	currentState = STATE_NULL;
	oldState = STATE_NULL;
	currentRate = RATE_NULL;
	probe = false;

	// random variables initialization


	// State transition timer settings
	// Transition time is uniform random (min, max)
	min_state_change_time = 2.5; // 10.0 secs	// new training procedure: 50.0 secs
	max_state_change_time = 7.5; // 15.0 secs	// new training procedure: 75.0 secs

  ////////////////////////////////////////////////////////////////////////////////////////////////////
  // end of: initialize DATA RELATED TO MAPPLE
  ////////////////////////////////////////////////////////////////////////////////////////////////////
}

MappleApplication::~MappleApplication()
{
  NS_LOG_FUNCTION (this);
}

uint32_t MappleApplication::GetTotalRx () const
{
  return m_totalRx;
}

Ptr<Socket>
MappleApplication::GetListeningSocket (void) const
{
  NS_LOG_FUNCTION (this);
  return m_socket;
}

std::list<Ptr<Socket> >
MappleApplication::GetAcceptedSockets (void) const
{
  NS_LOG_FUNCTION (this);
  return m_socketList;
}

void MappleApplication::DoDispose (void)
{
  NS_LOG_FUNCTION (this);
  m_socket = 0;
  m_socketList.clear ();

  // Only the first node: close the file for writing samples
  if (GetNode()->GetId() == 0) {
	  fclose(fp);
	  fclose(fp_samples_geo);
	  fclose(fp_mobility);
	  fclose(fp_processed);
  }

  // chain up
  Application::DoDispose ();
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// ns3::Application Methods
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////

/*
 * Called at time specified by Start in a simulation script
 */
void MappleApplication::StartApplication ()
{
	NS_LOG_FUNCTION (this);

	// Create the listening socket if not already
	if (!m_socket)
	{
		m_socket = Socket::CreateSocket (GetNode (), m_tid);
		m_socket->Bind (InetSocketAddress(Ipv4Address::GetAny (), portBase));
		m_socket->Listen ();
		m_socket->ShutdownSend ();
	}
	m_socket->SetRecvCallback (MakeCallback (&MappleApplication::HandleRead, this));
	m_socket->SetAcceptCallback (
			MakeNullCallback<bool, Ptr<Socket>, const Address &> (),
			MakeCallback (&MappleApplication::HandleAccept, this));
	m_socket->SetCloseCallbacks (
			MakeCallback (&MappleApplication::HandlePeerClose, this),
			MakeCallback (&MappleApplication::HandlePeerError, this));

	// Also create a socket for sending data
	if (!sendSocket) {
		sendSocket = ns3::Socket::CreateSocket(GetNode(),m_tid);
		sendSocket->Bind();
		sendSocket->ShutdownRecv ();
		//	  sendSocket->SetRecvCallback(MakeCallback(&IDSIA::NetworkApplication::Receive,this));
		//	  sendSocket->SetSendCallback(MakeCallback(&IDSIA::NetworkApplication::Sent,this));
		sendSocket->SetAllowBroadcast (true);
		sendSocket->SetAttribute ("IpTtl", UintegerValue (1));	// Only on hop messages !!! TODO: verify if this really works...
	}


	// Save a pointer to own mobility model
	m_mobility =  GetNode ()->GetObject<WaypointMobilityModel>();
	//m_mobility =  GetNode ()->GetObject<MobilityModel>();

	// Sava a pointer to MAC queue
	std::pair< Ptr< Ipv4 >, uint32_t > pair = interfaces.Get(GetNode()->GetId());
	Ptr< Ipv4 > ipv4 = pair.first;
	Ptr< WifiNetDevice > wifiNetDevice = DynamicCast<WifiNetDevice>(ipv4->GetNetDevice(pair.second));	// get wifi net device
	Ptr< RegularWifiMac > regularWifiMAC = DynamicCast<RegularWifiMac>(wifiNetDevice->GetMac());		// get the MAC object
	macQueue = regularWifiMAC->GetDcaTxop()->GetQueue();								// get the MAC queue object (tricky, required changing the RegularWifiMac class in order to make 'GetDcaTxop()' accessible...)

	// Also, send a promiscuous callback (when in unicast mode...)
	if (SEND_UNICAST)
		wifiNetDevice->SetPromiscReceiveCallback(MakeCallback (&MappleApplication::HandlePromiscuous, this));

	////////////////////////////////////////////////////////////////////////////////////////////////////
	// initialize DATA RELATED TO SVM PREDICTION MODEL
	////////////////////////////////////////////////////////////////////////////////////////////////////

	classCounter = new uint32_t[10];
	for (int i=0; i < 10; i++) {
		classCounter[i] = 0;
	}

#ifdef VALIDATION
	featuremax = new double[19];
	featuremin = new double[19];
	if (modelName == "model1") {
		modelfilename = "/home/mkudelsk/IDSIA/Link quality prediction/Models/trainedmodels/model1.txt";
		for (int i=0; i < 19; i++) {
			featuremax[i] = featuremaxM1[i];
			featuremin[i] = featureminM1[i];
		}
	} else if (modelName == "model_new") {
		modelfilename = "/home/mkudelsk/IDSIA/Link quality prediction/Models/trainedmodels/model_new.txt";
		for (int i=0; i < 19; i++) {
			featuremax[i] = featuremaxMnew[i];
			featuremin[i] = featureminMnew[i];
		}
		useDistance = true;
		useRSSI = true;
		useSNR = true;
	} else if (modelName == "model2") {
		modelfilename = "/home/mkudelsk/IDSIA/Link quality prediction/Models/trainedmodels/model2.txt";
		for (int i=0; i < 18; i++) {
			featuremax[i] = featuremaxM2[i];
			featuremin[i] = featureminM2[i];
		}
		useDistance = true;
		useRSSI = true;
		useSNR = false;
	} else if (modelName == "model3") {
		modelfilename = "/home/mkudelsk/IDSIA/Link quality prediction/Models/trainedmodels/model3.txt";
		for (int i=0; i < 17; i++) {
			featuremax[i] = featuremaxM3[i];
			featuremin[i] = featureminM3[i];
		}
		useDistance = true;
		useRSSI = false;
		useSNR = false;
	} else if (modelName == "model4") {
		modelfilename = "/home/mkudelsk/IDSIA/Link quality prediction/Models/trainedmodels/model4.txt";
		for (int i=0; i < 17; i++) {
			featuremax[i] = featuremaxM4[i];
			featuremin[i] = featureminM4[i];
		}
		useDistance = false;
		useRSSI = true;
		useSNR = false;
	} else if (modelName == "model5") {
		modelfilename = "/home/mkudelsk/IDSIA/Link quality prediction/Models/trainedmodels/model5.txt";
		for (int i=0; i < 16; i++) {
			featuremax[i] = featuremaxM5[i];
			featuremin[i] = featureminM5[i];
		}
		useDistance = false;
		useRSSI = false;
		useSNR = false;
	} else {
		// DEFAULT
		modelfilename = "/home/mkudelsk/IDSIA/Link quality prediction/Models/trainedmodels/model1.txt";
		for (int i=0; i < 19; i++) {
			featuremax[i] = featuremaxM1[i];
			featuremin[i] = featureminM1[i];
		}
	}

	// load trained model
	if((model=svm_load_model(modelfilename.c_str()))==0)
	{
		printf("[ERROR] can't open SVM model file: %s\n",modelfilename.c_str());
		exit(1);
	}
	else
	{
		nooffeatures = model->vecDim;
		//nooffeatures = 19;
		svm_type=svm_get_svm_type(model);
		nr_class=svm_get_nr_class(model);
	}

	//Copy the link quality features into 1D array for testing with the SVM
	testfeatures = (struct svm_node*) malloc((nooffeatures+1)*sizeof(struct svm_node));
#endif
	////////////////////////////////////////////////////////////////////////////////////////////////////
	// END OF: initialize DATA RELATED TO SVM PREDICTION MODEL
	////////////////////////////////////////////////////////////////////////////////////////////////////

	////////////////////////////////////////////////////////////////////////////////////////////////////
	// initialize DATA RELATED TO MAPPLE
	////////////////////////////////////////////////////////////////////////////////////////////////////
	// Initialize neighbor table
	for(unsigned int i=0;i<MAX_NODE_ID;i++)
	{
		memset(&Neighbors[i],0x0,sizeof(neighbor_entry_t));
		Neighbors[i].sum_rssi = 0;
		Neighbors[i].sum_snr = 0;
		Neighbors[i].rssi_samples = 0;
		Neighbors[i].state = STATE_NULL;
		memset(&samples[i],0x0,sizeof(sample_t));

		// estimators


		// estimators here
		Neighbors[i].last_rssi = 0.0;
		Neighbors[i].last_snir = 0.0;
		Neighbors[i].mw_rssi.window_size = 20;
		Neighbors[i].mw_snir.window_size = 20;
		Neighbors[i].ewma_rssi.alpha = 0.925;	// 0.925 corresponds to the history of c.a. 25 packets
		Neighbors[i].ewma_snir.alpha = 0.925;
		Neighbors[i].ewma_prr.alpha = 0.925;

		Neighbors[i].wmemwa_prr.alpha = 0.6;
		Neighbors[i].wmemwa_prr.time_interval = 0.5;	// in seconds

	}

	Simulator::Schedule (Seconds(Neighbors[0].wmemwa_prr.time_interval), &MappleApplication::updateEstimatorsPeriodic, this);

	// Initialize status
	for(unsigned int i=0;i < 5; i++)
	{
		my_status[i] = 0;
	}

	// Only the first node: open the file for writing samples
	if (GetNode()->GetId() == 0) {
	  std::string tmp = outFilePath;
	  tmp.append(".txt");
	  fp = fopen(tmp.c_str(),"w");					// file for writing samples
	  tmp = outFilePath;
	  tmp.append("_geo.txt");
	  fp_samples_geo = fopen(tmp.c_str(),"w");		// file for writing geographical coordinates related to samples
	  tmp = outFilePath;
	  tmp.append("_mobility.txt");
	  fp_mobility = fopen(tmp.c_str(),"w");			// file for the whole mobility trace
	  tmp = outFilePath;
	  tmp.append("_processed.txt");
	  fp_processed = fopen(tmp.c_str(),"w");		// file for writing processed samples
	}

	////////////////////////////////////////////////////////////////////////////////////////////////////
	// end of: initialize DATA RELATED TO MAPPLE
	////////////////////////////////////////////////////////////////////////////////////////////////////

	// Schedule a cleaning timer
	m_cleanEvent = Simulator::Schedule (Seconds(CLEANING_TIMER_PERIOD), &MappleApplication::CleaningEvent, this);

	// Intially all idle
	currentState = STATE_DISABLED;
//	#ifdef VALIDATION
//		if(currentState == STATE_MOVING || currentState == STATE_DISABLED)
//		{
//			//printf("Initial state MOVING, changing\n");
//			currentState = STATE_PROBING;
//		}
//	#endif
	changeState(currentState);

	// announce your state to neighbours (do it with a slight delay for the first time: we do not want to collide with our neighbours)
	Simulator::Schedule (Seconds(0.001*GetNode()->GetId()), &MappleApplication::announceState, this,ANNOUNCE_TIMES);

	// Schedule a state-change timer
	if( currentState != STATE_MOVING) {
		// In order to avoid a state change while moving
		// NOTE: When node is moving, simulation environment signals that node stops moving
		double ttt = stateChangeInterval();
		Time nextStateChangeTime = Seconds(ttt);
		m_stateChangeEvent = Simulator::Schedule (nextStateChangeTime, &MappleApplication::ChangeStateEvent, this);
		// Also, should add a waypoint for the mobility model: it is important that the waypoint event is executed later than the next state event...
		Time waypointTime = nextStateChangeTime+Simulator::Now()+NanoSeconds(1);
		Vector waypointPosition = m_mobility->GetPosition();
		m_mobility->AddWaypoint (Waypoint(waypointTime, waypointPosition));
#ifdef SAVE_WAYPOINTS
		fprintf(fp_mobility,"[%f] %u (%f,%f)\n",waypointTime.GetSeconds(),GetNode()->GetId(),waypointPosition.x,waypointPosition.y);
#endif
	}
}

/*
 * Called at time specified by Stop in a simulation script
 */
void MappleApplication::StopApplication ()
{
  NS_LOG_FUNCTION (this);
  while(!m_socketList.empty ()) //these are accepted sockets, close them
    {
      Ptr<Socket> acceptedSocket = m_socketList.front ();
      m_socketList.pop_front ();
      acceptedSocket->Close ();
    }
  if (m_socket)
    {
      m_socket->Close ();
      m_socket->SetRecvCallback (MakeNullCallback<void, Ptr<Socket> > ());
    }
  if (sendSocket) {
	  sendSocket->Close ();
  }

}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Events handlers for various timer-triggered events
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////

/*
 *
 */
void MappleApplication::SendPacket ()
{
  NS_LOG_FUNCTION_NOARGS ();
  NS_LOG_LOGIC ("sending packet at " << Simulator::Now ());
  NS_ASSERT (m_sendEvent.IsExpired ());

  if (currentState != STATE_PROBING) {
	  // should not happen - show warning
	  std::cout << "SendPacket event fired while not probing !!! [ERROR]" << std::endl;
  }

	probe_msg_t out;
	out.msg_type = MSG_PROBE;
	// dbg("MetricTestC","[%s] Sending PROBE message - counter %u\n", sim_time_string(),my_counter+1);

	// Set state field
	out.state = STATE_PROBING;

	// Increment counter
	out.counter = ++my_counter;

	out.rate = currentRate;

	out.sendTime = Simulator::Now().GetSeconds();


	for(unsigned int i=0; i< 5;i++) {
		out.source_status[i] = my_status[i];
	}

	//dbg("MetricTestC", "Setting sender features\n");
	int j = 0;
	for(unsigned int i=0; i < MAX_NODE_ID;i++)	{
		if(is_set(my_status[i/32], i%32)) {
			out.rate_data[j] = Neighbors[i].rate;
			out.rssi_data[j++] = (Neighbors[i].sum_rssi / Neighbors[i].rssi_samples);
//			dbg("MetricTestC",
//					"Setting RSSI and RATE DATA for %i, %hh [%u,%u] %hhu\n",
//					i,
//					out->rssi_data[j-1],
//					Neighbors[i].sum_rssi,
//					Neighbors[i].rssi_samples,
//					out->rate_data[j-1]);
		}

		if( j >= MAX_REPORTED_NEIGHBORS) {
			std::cout << "ERROR - Ups, too many neighbors" << std::endl;
			//dbg("ERRORS","[%s] ERROR - Ups, too many neighbors", sim_time_string());
		}
	}

	// For all messages, sent:
	out.last_update = last_update;
	out.source = GetNode()->GetId();
	out.n_samples = n_samples;

	// Create packet and set destination
	Ipv4Address destination;
	if (SEND_UNICAST) {
		destination = interfaces.GetAddress(currentDestination);		// unicast transmission: to node #currentDestination
	} else {
		destination = interfaces.GetAddress( GetNode()->GetId()).GetBroadcast();	// broadcast transmission
	}
	Ptr<Packet> packet = Create<Packet> (reinterpret_cast<const uint8_t*>(&out),sizeof(probe_msg_t));
	uint32_t sizeGap = PROBING_PACKET_SIZE - packet->GetSize();
	packet->AddPaddingAtEnd(sizeGap);

	// Send via socket
	sendSocket -> SendTo(packet, 0, InetSocketAddress(destination,portBase));
	m_txTrace (packet);	// support tracing of send packets
	NS_LOG_INFO ("[" << Simulator::Now().GetSeconds() << "] " << GetNode()->GetId()  << " (" << interfaces.GetAddress(GetNode()->GetId()) << "): Sent a packet to ..." << destination);

	// Schedule next send packet event
	m_sendEvent = Simulator::Schedule (Seconds(next_message_interval()), &MappleApplication::SendPacket, this);
}

/*
 *	Timer routine used to clean neighbor table
 *	someone might be gone or changed state
 *	and we did not get notified
 *	Look: MAX_NEIGHBOR_TTL (maximum time without receiving any message)
 *	Note: If cleaning is needed and we are sampling, we must reinitialize
 *	because our environment has 'changed', at least logically
 */
void MappleApplication::CleaningEvent (){
	double current_time = Simulator::Now().GetSeconds();

	//std::cout << current_time << " - Cleaning event, node " << GetNode()->GetId() << " at " << m_mobility->GetPosition() << std::endl;

	for(unsigned int i=0;i<MAX_NODE_ID;i++){
		if( Neighbors[i].state == STATE_PROBING)
		{
			if( (current_time - Neighbors[i].last_message_time) > MAX_NEIGHBOR_TTL){
				Neighbors[i].state = STATE_DISABLED;
				last_update = my_counter;
				if(currentState == STATE_PROBING){
					closeAllSamples();
				}
				clear_bit(my_status[i / 32],i % 32);
			}
		}
	}
	// schedule next cleaning event
	m_cleanEvent = Simulator::Schedule (Seconds(CLEANING_TIMER_PERIOD), &MappleApplication::CleaningEvent, this);
}



/*
 * Here is where nodes change states
 */
void MappleApplication::ChangeStateEvent()
{
	oldState = currentState;
	currentState = nextState();

	// If we are in the UNICAST mode, and we've randomly chosen the STATE_PROBING, check whether there is someone to send data to...
	if ( (SEND_UNICAST) && (currentState == STATE_PROBING)) {
		currentDestination = choose_destination();
		// if there are no neighbors, switch to STATE_DISABLED instead
		if (currentDestination < 0) {
			currentState = STATE_DISABLED;
		}
	}

	//cout << Simulator::Now().GetSeconds() << " - " << GetNode()->GetId() << " change state (from " << (int)oldState << " to " << (int)currentState << ") POS=" << m_mobility->GetPosition() << endl;

	// Debugging info

	//dbg("STATE_CHANGE","[%s] %hhu -> %hhu\n",sim_time_string(),oldState,currentState);


	// This routine handles all operations
	// required when node change states
	changeState(currentState);

	// Announce neighbors
	// Repeated announcements to
	// mitigate packet losses
	announceState(ANNOUNCE_TIMES);

	// Schedule next state change event
	if( currentState != STATE_MOVING) {
		// In order to avoid a state change while moving
		// NOTE: When node is moving, simulation environment signals that node stops moving
		Time nextStateChangeTime = Seconds(stateChangeInterval());
		m_stateChangeEvent = Simulator::Schedule (nextStateChangeTime, &MappleApplication::ChangeStateEvent, this);
		// Also, should add a waypoint for the mobility model: it is important that the waypoint event is executed later than the next state event...
		Time waypointTime = (nextStateChangeTime+Simulator::Now()+NanoSeconds(1));
		Vector waypointPosition = m_mobility->GetPosition();
		m_mobility->AddWaypoint (Waypoint(waypointTime, waypointPosition));
#ifdef SAVE_WAYPOINTS
		fprintf(fp_mobility,"[%f] %u (%f,%f)\n",waypointTime.GetSeconds(),GetNode()->GetId(),waypointPosition.x,waypointPosition.y);
#endif
	}
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Other events handlers
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////

/*
 * Called when a packet is received from the MAC layer
 * (one of the core functions here, handles all incoming packets...)
 */
void MappleApplication::HandleRead (Ptr<Socket> socket)
{
	NS_LOG_FUNCTION (this << socket);
	Ptr<Packet> packet;
	Address fromAddress;

	while (packet = socket->RecvFrom (fromAddress))
	{
		if (packet->GetSize () == 0)
		{ //EOF
			break;
		}

		// Get RSSI of received message
		double rssi = 0;
		double snr = 0;
		PhySignalTag pst;
		if(packet->PeekPacketTag(pst)){		// the RSSI value is stored in a packet tag added my (modified) 'YansWifiPhy::EndReceive' function
			rssi = pst.GetPwr();
			snr = pst.GetSnr();	// As a 'bonus', also the SNR value is obtained here...
		} else {
			cout << "A serious error - no packet tag with the RSSI value available !!!"<< endl;
		}



		// Ignore all messages if node is moving
		if( currentState == STATE_MOVING)
		{
			// update estimators first
			if (InetSocketAddress::IsMatchingType (fromAddress)) {
				// Get packet data
				uint8_t buf[packet->GetSize()];
				packet->CopyData(buf,packet->GetSize());
				probe_msg_t* in = (probe_msg_t*) buf;

				updateEstimatorsOnReceive (in, rssi, snr);
			}

			m_rxTrace (packet, fromAddress);	// and only support tracing of the received packets...
			continue;
		}

		if (InetSocketAddress::IsMatchingType (fromAddress))
		{
			m_totalRx += packet->GetSize ();
			Ipv4Address address = InetSocketAddress::ConvertFrom (fromAddress).GetIpv4();
			//          NS_LOG_INFO (GetNode()->GetId()  << " (" << interfaces.GetAddress(GetNode()->GetId()) << "): Received " << packet->GetSize () << " bytes from " <<
			//                       address << " [" << address << "]"
			//                                   << " total Rx " << m_totalRx);


			// Get packet data
			uint8_t buf[packet->GetSize()];
			packet->CopyData(buf,packet->GetSize());
			probe_msg_t* in = (probe_msg_t*) buf;

			// Process packet data
			double rat = 0;
			double ht = 0;
			MacDelaysTag mdt;
			if(packet->PeekPacketTag(mdt)){		//
				rat = mdt.GetRequestAccessTime();
				ht = mdt.GetHandshakeTime();	//
			} else {
				//cout << "A serious error - no packet tag with MAC delays values available !!!"<< endl;
			}
			uint32_t qs = 0;
			MacQueueTag mqt;
			if(packet->PeekPacketTag(mqt)){		//
				qs = mqt.GetQueueSize();
			} else {
				//cout << "A serious error - no packet tag with MAC queue values available !!!"<< endl;
			}
			updateEstimatorsOnReceiveBeforeProcessing(in, rssi, snr);
			ProcessPacketData(in,address,rssi,snr,rat,ht,qs);
			updateEstimatorsOnReceiveAfterProcessing(in, rssi, snr);

		}
		m_rxTrace (packet, fromAddress);		// Support tracing of the received packets
	}
}

/*
 * Update estimators periodically
 */
void MappleApplication::updateEstimatorsPeriodic() {
	for(unsigned int i=0;i<MAX_NODE_ID;i++)	{
		Neighbors[i].wmemwa_prr.periodic_update();
	}
	Simulator::Schedule (Seconds(Neighbors[0].wmemwa_prr.time_interval), &MappleApplication::updateEstimatorsPeriodic, this);
}

/*
 * Update estimators each time a (probe) packet is received (before processing, i.e., not considering the packet that is just received, but only the previously accumulated events)
 *
 */
void MappleApplication::updateEstimatorsOnReceiveBeforeProcessing (probe_msg_t* in, double rssi, double snir) {
	unsigned int from = in->source;
	neighbor_entry_t* neighbor_entry =  &(Neighbors[from]);

	if ( in->msg_type != MSG_PROBE)
		return;

	// prr estimators
	neighbor_entry->ewma_prr.update((in->counter - neighbor_entry->ewma_prr.last_sequence_number - 1),0,in->counter);
	neighbor_entry->wmemwa_prr.update((in->counter - neighbor_entry->wmemwa_prr.last_sequence_number - 1),0,in->counter);
}

/*
 * Update estimators each time a (probe) packet is received (after processing, i.e., considering the packet that is just received, but not the previously accumulated events)
 */
void MappleApplication::updateEstimatorsOnReceiveAfterProcessing (probe_msg_t* in, double rssi, double snir) {
	unsigned int from = in->source;
	neighbor_entry_t* neighbor_entry =  &(Neighbors[from]);

	if ( in->msg_type != MSG_PROBE)
		return;

	// single value estimators
	neighbor_entry->last_rssi = rssi;
	neighbor_entry->last_snir = snir;

	// moving window estimators
	neighbor_entry->mw_rssi.update(rssi);
	neighbor_entry->mw_snir.update(snir);

	// ewma estimators
	neighbor_entry->ewma_rssi.update(rssi);
	neighbor_entry->ewma_snir.update(snir);

	// prr estimators
	neighbor_entry->ewma_prr.update(0,1,in->counter);
	neighbor_entry->wmemwa_prr.update(0,1,in->counter);
}

/*
 * Actual packet processing
 */
void MappleApplication::ProcessPacketData(probe_msg_t* in, Ipv4Address address, double rssi, double snr, double rat, double ht, uint32_t qs) {

	/////////////////////////////////////////////////////////////////////////////////////////////////
	// Perform for each received packet:
	/////////////////////////////////////////////////////////////////////////////////////////////////
	unsigned int from = in->source;
	double current_time = Simulator::Now().GetSeconds();
	Neighbors[from].last_message_time = current_time;



	//			dbg("MetricTestC",
	//			    "[%s] Received packet from %u with rssi %hhi\n",sim_time_string(),from,rssi);
	//
	//			// Debugging info
	//			dbg("MetricTestC",
	//			    "[%s] Received packet from %u Neighbor status %d\n",
	//			    sim_time_string(),
	//			    from,
	//			    Neighbors[from].state);

	// Rssi statistics collection
	// Just compute average
	if(Neighbors[from].rssi_samples == 0)
	{
		//				dbg("MetricTestC","First rssi sample from %i rssi %i sum_rssi %u\n",
		//				    from,
		//				    (int32_t)rssi,
		//				    Neighbors[from].sum_rssi);
		Neighbors[from].sum_rssi = rssi;
		Neighbors[from].sum_snr = snr;
	}else
	{
		Neighbors[from].sum_rssi += rssi;
		Neighbors[from].sum_snr += snr;
	}
	Neighbors[from].rssi_samples += 1;

	//			dbg("MetricTestC","[%s] Updating rssi of %i sum_rssi %i rssi_samples %u\n",
	//			    sim_time_string(),
	//			    from,
	//			    Neighbors[from].sum_rssi,
	//			    Neighbors[from].rssi_samples);

	// Now, handle different message types individually /////////////////////////////////////////////////
	if ( in->msg_type == MSG_PROBE)
	{
		/////////////////////////////////////////////////////////////////////////////////////////////////
		// handle probe message
		/////////////////////////////////////////////////////////////////////////////////////////////////
		NS_LOG_INFO (GetNode()->GetId()  << "Received MSG_PROBE from " <<
				address << " [source=" << (int)in->source << " , state=" << (int)in->state << " , last_update=" << (int)in->last_update << " , rate=" << (int)in->rate << " , RSSI=" << rssi << ", SNR=" << snr);

		// CASE: Received PROBE from non announced neighbor
		if(Neighbors[from].state != in->state && in->state == STATE_PROBING)
		{
			//					dbg("MetricTestC", "[%s] PROBE MSG RCV for not announced neighbor\n", sim_time_string());

			// New guy in the neighborhood probing
			// Restart all samples
			if( currentState == STATE_PROBING)
				closeAllSamples();

			// Our neighbor table has changed, so log last update
			last_update = my_counter;

			// Update neighbor table (include new neighbor)
			if( in->state == STATE_PROBING)
			{
				set_bit(my_status[from/32],from % 32);
			}else
			{
				clear_bit(my_status[from / 32],from % 32);
			}
		}

		// If am PROBING I am also collecting samples, so process PROBE message
		if( currentState == STATE_PROBING)
		{
			//dbg("MetricTestC","[%s] PROBING!\n",sim_time_string());
			//report_received(in);

			// Handle first sample
			if(samples[from].received == 0)
			{
				//						dbg("MetricTestC",
				//						    "[%s] First sample from %u counter %u\n",
				//						    sim_time_string(),
				//						    from,in->counter);

				Neighbors[from].last_update = in->last_update;
				samples[from].received = 1;
				samples[from].avgTravelTime = (current_time - in->sendTime);
				samples[from].queueSize = qs;
				samples[from].avgSendTime = (current_time - rat);	// calculated starting from the moment when dca-txop requested for channel access for this packet...
				samples[from].avgSendTimeWithoutBackoff = (current_time - ht);	// calculated starting from the moment when packet was dequeued from the MAC queue

				samples[from].last_sample_time = current_time;
				samples[from].first_counter = in->counter;
				samples[from].sender_rate = in->rate;
				memcpy(samples[from].sender_status,
						in->source_status,sizeof(uint32_t)*5);
				samples[from].receiver_status[0] = my_status[0];
				samples[from].receiver_status[1] = my_status[1];
				samples[from].receiver_status[2] = my_status[2];
				samples[from].receiver_status[3] = my_status[3];
				samples[from].receiver_status[4] = my_status[4];

				//						dbg("MetricTestC","mystatus %d %d %d %d %d\n",
				//						    my_status[0],my_status[1],my_status[2],
				//						    my_status[3],my_status[4]);
				//						dbg("MetricTestC","recstatus %d %d %d %d %d\n",
				//						    samples[from].receiver_status[0],
				//						    samples[from].receiver_status[1],
				//						    samples[from].receiver_status[2],
				//						    samples[from].receiver_status[3],
				//						    samples[from].receiver_status[4]);

				memcpy(samples[from].sender_rssi_data,
						in->rssi_data,sizeof(float)*MAX_REPORTED_NEIGHBORS);
				memcpy(samples[from].sender_rate_data,
						in->rate_data, sizeof(uint8_t)*MAX_REPORTED_NEIGHBORS);
			}else
			{
				// If sender features have changed since last probe packet
				// close previous and deal with this as a first sample
				if(Neighbors[from].last_update != in->last_update)
				{
					//							dbg("MetricTestC","[%s] last update does not match\n",
					//							    sim_time_string());

					closeSample(from);

					//							dbg("MetricTestC",
					//							    "[%s] First sample from %u counter %u\n",
					//							    sim_time_string(),
					//							    from,in->counter);

					// Initialize new sample
					samples[from].received = 1;
					samples[from].avgTravelTime = (current_time - in->sendTime);
					samples[from].queueSize = qs;
					samples[from].avgSendTime = (current_time - rat);	// calculated starting from the moment when dca-txop requested for channel access for this packet...
					samples[from].avgSendTimeWithoutBackoff = (current_time - ht);	// calculated starting from the moment when packet was dequeued from the MAC queue
					samples[from].last_sample_time = current_time;
					samples[from].first_counter = in->counter;
					samples[from].sender_rate = in->rate;
					memcpy(samples[from].sender_status,
							in->source_status,sizeof(uint32_t)*5);
					memcpy(samples[from].receiver_status,
							&my_status[0],sizeof(uint32_t)*5);
					memcpy(samples[from].sender_rssi_data,
							in->rssi_data,sizeof(float)*MAX_REPORTED_NEIGHBORS);
					memcpy(samples[from].sender_rate_data,
							in->rate_data, sizeof(uint8_t)*MAX_REPORTED_NEIGHBORS);


				}
				// Otherwise, keep sampling
				else
				{
					//							dbg("MetricTestC",
					//							    "[%s] Sample n %u from %u counter %u\n",
					//							    sim_time_string(),
					//							    samples[from].received+1,from,in->counter);
					samples[from].received+=1;
					samples[from].avgTravelTime += (current_time - in->sendTime);
					samples[from].queueSize += qs;
					samples[from].avgSendTime += (current_time - rat);	// calculated starting from the moment when dca-txop requested for channel access for this packet...
					samples[from].avgSendTimeWithoutBackoff += (current_time - ht);	// calculated starting from the moment when packet was dequeued from the MAC queue
					samples[from].last_sample_time = current_time;
					samples[from].last_counter = in->counter;
					samples[from].sender_rate = in->rate;


					memcpy(samples[from].sender_status,
							in->source_status,sizeof(uint32_t)*5);
					memcpy(samples[from].receiver_status,
							&my_status[0],sizeof(uint32_t)*5);

					samples[from].receiver_status[0] = my_status[0];
					samples[from].receiver_status[1] = my_status[1];
					samples[from].receiver_status[2] = my_status[2];
					samples[from].receiver_status[3] = my_status[3];
					samples[from].receiver_status[4] = my_status[4];

					//							dbg("MetricTestC","mystatus %d %d %d %d %d\n",
					//							    my_status[0],my_status[1],my_status[2],
					//							    my_status[3],my_status[4]);
					//							dbg("MetricTestC","recstatus %d %d %d %d %d\n",
					//							    samples[from].receiver_status[0],
					//							    samples[from].receiver_status[1],
					//							    samples[from].receiver_status[2],
					//							    samples[from].receiver_status[3],
					//							    samples[from].receiver_status[4]);

					memcpy(samples[from].sender_rate_data,
							in->rate_data, sizeof(uint8_t)*MAX_REPORTED_NEIGHBORS);

					memcpy(samples[from].sender_rssi_data,
							in->rssi_data,sizeof(float)*MAX_REPORTED_NEIGHBORS);
				}
			}

			////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
			// Add a subsample here
			////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

			// calculate distance between the sender and the receiver
			Vector myPosition = m_mobility->GetPosition();
			Vector senderPosition = nodes.Get(in->source)->GetObject<MobilityModel>()->GetPosition();
			double dist = CalculateDistance(senderPosition,myPosition);

			subsample_t subsample;

			subsample.timestamp = current_time;

			subsample.distance = dist;
			subsample.rrsi = rssi;
			subsample.sinr = snr;

			//MAPPLE_estimator;
			subsample.singleRSSI_estimator = Neighbors[from].last_rssi;
			subsample.singleSNIR_estimator = Neighbors[from].last_snir;
			subsample.mwRSSI_estimator = Neighbors[from].mw_rssi.current_sum / Neighbors[from].mw_rssi.current_size;
			subsample.mwSNIR_estimator = Neighbors[from].mw_snir.current_sum / Neighbors[from].mw_snir.current_size;
			subsample.ewmaRSSI_estimator = Neighbors[from].ewma_rssi.current_value;
			subsample.ewmaSNIR_estimator = Neighbors[from].ewma_snir.current_value;;
			subsample.ewmaPRR_estimator = Neighbors[from].ewma_prr.current_value;
			subsample.wmewmaPRR_estimator = Neighbors[from].wmemwa_prr.current_value;

			samples[from].subsamples.push_back(subsample);

		}

		Neighbors[from].state = in->state;
		Neighbors[from].rate = in->rate;
		Neighbors[from].last_update = in->last_update;

	} else if ( in->msg_type == MSG_ANNOUNCE) {
		/////////////////////////////////////////////////////////////////////////////////////////////////
		// Handle announce message
		/////////////////////////////////////////////////////////////////////////////////////////////////
		NS_LOG_INFO (GetNode()->GetId()  << " Received MSG_ANNOUNCE from " <<
				address << " [source=" << (int)in->source << " , state=" << (int)in->state << " , last_update=" << (int)in->last_update << " , rate=" << (int)in->rate << " , RSSI=" << rssi << ", SNR=" << snr);

		//dbg("MetricTestC","[%s] MSG_ANNOUNCE from %u\n", sim_time_string(),from);
		//report_received(in);

		// If the neighbor has changed state from/to PROBING
		// My neighbor table has changed
		if( (Neighbors[from].state == STATE_PROBING && in->state != STATE_PROBING)
				|| (Neighbors[from].state != STATE_PROBING && in->state == STATE_PROBING))
		{
			// If am probing, restart everything
			if(currentState == STATE_PROBING)
				closeAllSamples();

			// Log last update of neighbor table
			last_update = my_counter;

			// Update it
			if( in->state == STATE_PROBING)
			{
				//						dbg("MetricTestC",
				//						    "[%s] NODE %u has state PROBING - Setting bit - cntState %u\n",
				//						    sim_time_string(),from, my_status[from/32]);

				set_bit(my_status[from/32],from % 32);

				//						dbg("MetricTestC","[%s] bit set %u\n",
				//						    sim_time_string(),my_status[from/32]);

			}else
			{
				clear_bit(my_status[from / 32],from % 32);
			}
		} else if ( Neighbors[from].state == STATE_PROBING && in->state == STATE_PROBING && Neighbors[from].rate != in->rate ) {
			// Here, the neighbor changed its state from STATE_PROBING to STATE_PROBING (so, did not change it in fact), but it changed its outgoing rate. Thus, the state effectively changed...

			// If am probing, restart everything
			if(currentState == STATE_PROBING)
				closeAllSamples();

			// Log last update of neighbor table
			last_update = my_counter;

			//std::cout << current_time << " Received state change: probe-probe, but different rates, node " << GetNode()->GetId() << " from " << from << std::endl;
		}

		Neighbors[from].state = in->state;
		Neighbors[from].last_update = in->last_update;
		Neighbors[from].rate = in->rate;

		// If neighbor is announcing moving state
		// restart RSSI measures
		if(in->state == STATE_MOVING)
		{
			Neighbors[from].sum_rssi = 0;
			Neighbors[from].sum_snr = 0;
			Neighbors[from].rssi_samples = 0;
		}
	}
}

/**
 * Called for packets successfully received by the device, in promiscuous mode *
 *
 * \param device a pointer to the net device which is calling this callback
 * \param packet the packet received
 * \param protocol the 16 bit protocol number associated with this packet.
 *        This protocol number is expected to be the same protocol number
 *        given to the Send method by the user on the sender side.
 * \param sender the address of the sender
 * \param receiver the address of the receiver
 * \param packetType type of packet received (broadcast/multicast/unicast/otherhost)
 * \returns true if the callback could handle the packet successfully, false
 *          otherwise.
 */
bool MappleApplication::HandlePromiscuous(Ptr<NetDevice> device, Ptr<const Packet> packet, uint16_t protocol,
        const Address & sender, const Address & receiver, enum NetDevice::PacketType packetType) {

	// Here, we deal only with packets addressed to other hosts
	if (packetType != NetDevice::PACKET_OTHERHOST)
		return true;

	// We are only interested in Ipv4 packets, and not the packets from lower layers...
	if (protocol != Ipv4L3Protocol::PROT_NUMBER)
		return true;

	Ipv4Header header;
	packet->PeekHeader(header);

	//Ipv4Address dst = header.GetDestination ();
	Ipv4Address origin = header.GetSource ();

	//cout << "Promisc[" << Simulator::Now().GetSeconds() << "] " << GetNode()->GetId()  << " from " << origin << " to " << dst << " type " << (int) packetType << endl;

	// Get RSSI of received message
	double rssi = 0;
	double snr = 0;
	PhySignalTag pst;
	if(packet->PeekPacketTag(pst)){		// the RSSI value is stored in a packet tag added my (modified) 'YansWifiPhy::EndReceive' function
		rssi = pst.GetPwr();
		snr = pst.GetSnr();	// As a 'bonus', also the SNR value is obtained here...
	} else {
		cout << "A serious error - no packet tag with the RSSI value available !!!"<< endl;
	}

	// Get packet data
	Ptr<Packet> p2 = packet->Copy();	// need a copy to remove headers (cannont operate on 'packet' as it is declared constant...)
	Ipv4Header ipHeader;
	p2->RemoveHeader(ipHeader);			// remove IP header
	LlcSnapHeader macHeader;
	p2->RemoveHeader(macHeader);		// remove MAC layer header

	uint32_t s = p2->GetSize();			// get data, finally
	uint8_t buf[s];
	p2->CopyData(buf,p2->GetSize());
	probe_msg_t* in = (probe_msg_t*) buf;

	// Process packet data
	double rat = 0;
	double ht = 0;
	MacDelaysTag mdt;
	if(packet->PeekPacketTag(mdt)){		// the RSSI value is stored in a packet tag added my (modified) 'YansWifiPhy::EndReceive' function
		rat = mdt.GetRequestAccessTime();
		ht = mdt.GetHandshakeTime();	// As a 'bonus', also the SNR value is obtained here...
	} else {
		//cout << "A serious error - no packet tag with MAC delays values available !!!"<< endl;
	}
	uint32_t qs = 0;
	MacQueueTag mqt;
	if(packet->PeekPacketTag(mqt)){		//
		qs = mqt.GetQueueSize();
	} else {
		//cout << "A serious error - no packet tag with MAC queue values available !!!"<< endl;
	}

	// // Ignore all messages if node is moving (i.e., update estimators only, do not process the data)
	if( currentState == STATE_MOVING) {
		updateEstimatorsOnReceive (in, rssi, snr);
		return true;
	}
	updateEstimatorsOnReceiveBeforeProcessing(in, rssi, snr);
	ProcessPacketData(in,origin,rssi,snr,rat,ht,qs);
	updateEstimatorsOnReceiveAfterProcessing(in, rssi, snr);

	return true;
}

/*
 * not supported here
 */
void MappleApplication::HandlePeerClose (Ptr<Socket> socket)
{
  NS_LOG_INFO ("MappleApplication, peerClose");
}

/*
 * not supperted here
 */
void MappleApplication::HandlePeerError (Ptr<Socket> socket)
{
  NS_LOG_INFO ("MappleApplication, peerError");
}

/*
 * called when a new socket is accepted (not used for now...)
 */
void MappleApplication::HandleAccept (Ptr<Socket> s, const Address& from)
{
  NS_LOG_FUNCTION (this << s << from);
  s->SetRecvCallback (MakeCallback (&MappleApplication::HandleRead, this));
  m_socketList.push_back (s);
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Other helpful functions
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////

/*
 * Handle the state change
 */
void MappleApplication::changeState(uint8_t newState)
{
	//dbg_clear("STATE_CHANGE","STATE %s %u ",sim_time_string(),TOS_NODE_ID);
	if(newState == STATE_DISABLED)
	{
		//dbg_clear("STATE_CHANGE","DISABLED\n");

		if(oldState == STATE_PROBING)
		{
			closeAllSamples();
			stopProbing();
		}
	}
	else if(newState == STATE_PROBING)
	{
		//dbg_clear("STATE_CHANGE","PROBING\n");

		// This might seem ackward, but if
		// we are using random state transitions
		// we could end up here.
		// Better be safe and reinitialize anyway
		if(oldState == STATE_PROBING)
		{
			closeAllSamples();
		}

		initializeSampling();
		startProbing();

	}else if(newState == STATE_MOVING)
	{
		//dbg_clear("STATE_CHANGE","MOVING\n");

		if(oldState == STATE_PROBING)
		{
			closeAllSamples();
			stopProbing();
		}

		Move();
	}
}

/*
 * Announce state change to neighbors
 * (repeat 'announcementsNumber' times)
 */
void MappleApplication::announceState(uint8_t announcementsNumber)
{
	// Set data - we use same message structure for announcements.
	// Just send message type
	probe_msg_t out;
	out.msg_type = MSG_ANNOUNCE;
	out.state = currentState;
	//out.last_update = my_counter;
	out.rate = currentRate;

	out.last_update = last_update;
	out.source = GetNode()->GetId();
	out.n_samples = n_samples;

	// Set destination
	Ipv4Address destination;
	destination = interfaces.GetAddress( GetNode()->GetId()).GetBroadcast();	// broadcast transmission

	//std::cout << Simulator::Now().GetSeconds() << " - " << GetNode()->GetId() << ": state=" << (int)currentState << " , pos=" << m_mobility->GetPosition() << std::endl;

	// Repeat 'announcementsNumber' times
	for (unsigned int i=0; i < announcementsNumber; i++) {
		Ptr<Packet> packet = Create<Packet> (reinterpret_cast<const uint8_t*>(&out),sizeof(probe_msg_t));
		uint32_t sizeGap = PROBING_PACKET_SIZE - packet->GetSize();
		packet->AddPaddingAtEnd(sizeGap);
		// Send via socket
		sendSocket -> SendTo(packet, 0, InetSocketAddress(destination,portBase));
		m_txTrace (packet);	// support tracing of send packets
		NS_LOG_INFO ("[" << Simulator::Now().GetSeconds() << "] " << GetNode()->GetId()  << " (" << interfaces.GetAddress(GetNode()->GetId()) << "): Sent a packet to ..." << destination);
	}

}

/*
 * In order to start probing,
 * we select rate, and start the MessageTimer
 */
void MappleApplication::startProbing(){
	double next_interval;

	currentRate = next_rate();

	//	if(currentRate == RATE_LOW)
//	{
//		dbg("MetricTestC","RATE LOW\n");
//	}else if(currentRate == RATE_MED)
//	{
//		dbg("MetricTestC","RATE MED\n");
//	}else if(currentRate == RATE_HIGH)
//	{
//		dbg("MetricTestC","RATE HIGH\n");
//	}

	// Set probe flag
	probe = true;

	// Schedule a send event...
	if (m_sendEvent.IsRunning ()) {	// Just in case, check whether there are no pending send events and cancel them...
		Simulator::Cancel (m_sendEvent);
	}
	next_interval = next_message_interval();
	m_sendEvent = Simulator::Schedule (Seconds(next_interval), &MappleApplication::SendPacket, this);

}

/*
 *  Initialize sample table
 */
void MappleApplication::initializeSampling()
{
	for(unsigned int i=0;i < MAX_NODE_ID;i++)
	{
		samples[i].received = 0;
		samples[i].avgTravelTime = 0;
		samples[i].queueSize = 0;
		samples[i].avgSendTime = 0;
		samples[i].avgSendTimeWithoutBackoff = 0;
		memset(samples[i].sender_status,0,sizeof(uint32_t)*5);
		memset(samples[i].receiver_status,0,sizeof(uint32_t)*5);
	}

}

/*
 * StopProbing ->
 *	- stop current message timer, disable probe (flag)
 *   and clean msgqueue (messages not yet sent)
 */
void MappleApplication::stopProbing(){
	//dbg("MetricTestC","[%s] Stop probing\n",sim_time_string());

	// Stop packet generation timer
	Simulator::Cancel (m_sendEvent);

	// Set PROBE flag false
	probe = false;

	// Clean the MAC queue
	clearMacQueue();
}

/*
 * Chooses the destination for unicast transmissions, when in UNICAST probing mode
 */
int MappleApplication::choose_destination() {
	uint32_t myId;
	std::vector<uint32_t> neighborIds;

	myId = GetNode()->GetId();

	// check all neighbors
	for(unsigned int i=0;i<MAX_NODE_ID;i++){
		if( i == myId)
			continue;
		if( (Simulator::Now().GetSeconds() - Neighbors[i].last_message_time) <= MAX_NEIGHBOR_TTL){
			neighborIds.push_back(i);
		}
	}
	if (neighborIds.empty())
		return -1;
	return neighborIds[uniformVariable.GetInteger(0,neighborIds.size()-1)];	// uniform [0,neighborIds.size()]
}

/*
 * // Move the node
 */
void MappleApplication::Move()
{
	// dbg("MetricTestC","[%s] I must move!\n",sim_time_string());

	// clear the state variables (invalidate them)
	for(unsigned int i=0; i<5;i++)
	{
		my_status[i] = 0;
	}
	for(unsigned int i=0;i<MAX_NODE_ID;i++)
	{
		Neighbors[i].state = STATE_NULL;
		Neighbors[i].sum_rssi = 0;
		Neighbors[i].sum_snr = 0;
		Neighbors[i].rssi_samples = 0;
	}

	// Perform the move, actually...
	Vector destination = getNextDestination();
	double distance = CalculateDistance(destination,m_mobility->GetPosition());
	double velocity = SPEED;
	double travelTime = distance / velocity;
	///cout << "Next dest=" << destination << " , travel_time=" << travelTime  << endl;
	Time waypointTime = Seconds(Simulator::Now().GetSeconds() + travelTime);

	// Before we add a waypoint, we need to schedule the 'state change' event (tricky: it should be fired at the same simulation time as a 'waypoint' event, but before this event...)
	m_stateChangeEvent = Simulator::Schedule (Seconds(travelTime), &MappleApplication::ChangeStateEvent, this);

	// Add a waypoint (the 'AddWaypoint' method also schedules a 'waypoint' event, i.e. 'WaypointMobilityModel::Update' event)
	waypointTime = waypointTime+NanoSeconds(1);
	Waypoint waypoint(waypointTime, destination);
	m_mobility->AddWaypoint (waypoint);

#ifdef SAVE_WAYPOINTS
		fprintf(fp_mobility,"[%f] %u (%f,%f)\n",waypointTime.GetSeconds(),GetNode()->GetId(),destination.x,destination.y);
#endif

}

/*
 * Clear the MAC interface queue
 */
void MappleApplication::clearMacQueue(){

	macQueue->Flush();			// flush the queue

}

/*
 * Determines the time interval to call the state change
 * uniform random (min, max)
 */
double  MappleApplication::stateChangeInterval()
{
	return uniformVariable.GetValue(min_state_change_time,max_state_change_time);
}

/*
 *  This one determines time between messages
 *  depending on node's current rate
 */
double MappleApplication::next_message_interval()
{
	double nextInterval = 0.0;
	if( currentRate == RATE_LOW)
	{
		nextInterval = next_gaussian(RATE_LOW_MU,RATE_LOW_SIGMA);
	}else if(currentRate == RATE_MED)
	{
		nextInterval = next_gaussian(RATE_MED_MU, RATE_MED_SIGMA);
	}else if(currentRate == RATE_HIGH)
	{
		nextInterval = next_gaussian(RATE_HIGH_MU, RATE_HIGH_SIGMA);
	}
	if (nextInterval < 0.0)
		return 0.0;
	return nextInterval;
}

/*
 * Determines randomly PROBE rate
 * Just random uniform variable
 */
uint8_t MappleApplication::next_rate(){
	return uniformVariable.GetInteger(1,3);	// uniform [1,3]
}


/*
 * A node can be in any of these 3 states:
 *    Disabled
 *    Probing
 *    MOVE
 *   This function determines the state transitions
 *   Transitions can be deterministic of random
 *   depending on RANDOM_STATE_CHANGE flag
 */
uint8_t MappleApplication::nextState(){

	if( RANDOM_STATE_CHANGE)
	{
		double randomValue = uniformVariable.GetValue();
		if(randomValue < P_MOVE)
		{
			if(NO_MOVE)
				return STATE_DISABLED;	// remain disabled instead of moving, when moving is off...
			return STATE_MOVING;
		}else{
			if(randomValue < P_MOVE + P_DISABLED){
				return STATE_DISABLED;
			}else{
				return STATE_PROBING;
			}
		}
	} else {
//#ifndef VALIDATION
		// Deterministic behaviour
		// Idle -> Probe -> Move -> Idle

		if(NO_MOVE){
			if(currentState == STATE_DISABLED)
				return STATE_PROBING;
			if(currentState == STATE_MOVING)
				return STATE_PROBING;
			if(currentState == STATE_PROBING)
				return STATE_DISABLED;

		}else{
			if(currentState == STATE_DISABLED)
				return STATE_MOVING;
			if(currentState == STATE_MOVING)
				return STATE_PROBING;
			if(currentState == STATE_PROBING)
				return STATE_DISABLED;
		}
//#else
//		// This is just for debugging
//		// do not change state -
//		return currentState;
//#endif
	}
	return currentState;	// just for compiler not to complain about not returning any value
}

/*
 * Draw a normally distributed random value (given the mean and the std deviation of the normal distribution)
 */
double MappleApplication::next_gaussian(double mu, double sigma) {
	return normalVariable.GetValue()*sigma+mu;
}

/*
 * Output sample
 * Increment n_samples
 */
void MappleApplication::print_sample(uint8_t id)
{
	int j;
	double rssi, snr;
	double dist;
	uint8_t rate;

	n_samples +=1;

	// calculate distance between the sender and the receiver
	Vector myPosition = m_mobility->GetPosition();
	Vector senderPosition = nodes.Get(id)->GetObject<MobilityModel>()->GetPosition();
	dist = CalculateDistance(senderPosition,myPosition);


	//dbg("MetricTestC","Printing sample...\n");

	// Start writing the raw sample
	fprintf(fp,"SAMPLE_%u %u %f source %u dist %.1f rcv %u range %u ",
			++sampleID,
			GetNode()->GetId(),
			samples[id].last_sample_time,
			id,
			dist,
			samples[id].received,
			samples[id].last_counter - samples[id].first_counter + 1);

#ifdef SAVE_SAMPLES_GEO
	// Start writing sample coordinates
	fprintf(fp_samples_geo,"SAMPLE_%u %u(%.1f,%.1f) %u(%.1f,%.1f) REC [",sampleID,GetNode()->GetId(),myPosition.x,myPosition.y,id,senderPosition.x,senderPosition.y);
#endif

	rssi = ((Neighbors[id].sum_rssi / Neighbors[id].rssi_samples));
	snr = ((Neighbors[id].sum_snr / Neighbors[id].rssi_samples));
	rate = Neighbors[id].rate;

	// Continue writing the raw sample
	fprintf(fp,"rate %hhu",rate);
	fprintf(fp," rssi %f ", rssi);
	fprintf(fp," snr %f ", snr);
	fprintf(fp," rec_status [");

#ifdef SAVE_PROCESSED_SAMPLES
	// prepare counters
	uint8_t Fs[3][NUMBER_OF_DIST_INTERVALS];
	uint8_t Fr[3][NUMBER_OF_DIST_INTERVALS];
	uint16_t Vs = 0;
	uint16_t Vr = 0;
	uint8_t index;
	// initialize counters
	for (unsigned int i=0; i < 3; i++) {
		for (unsigned int k=0; k < NUMBER_OF_DIST_INTERVALS; k++) {
			Fr[i][k]=0;
			Fs[i][k]=0;
		}
	}
//	// Start writing processed sample
//	fprintf(fp_processed,"SAMPLE_%u %f td=%f d=%f 1:%f 2:%f 3:%d 4:%d ",
//			sampleID,
//			(1.0*samples[id].received/(samples[id].last_counter - samples[id].first_counter + 1)),
//			samples[id].avgTravelTime/samples[id].received,
//			samples[id].avgSendTime/samples[id].received,
//			dist,
//			rssi,
//			currentRate,
//			rate);
	fprintf(fp_processed,"SAMPLE_%u %f 1:%f 2:%f 3:%d 4:%d ",
			sampleID,
			(1.0*samples[id].received/(samples[id].last_counter - samples[id].first_counter + 1)),
			dist,
			rssi,
			currentRate,
			rate);
	// update the class counters
	double tmpVal = (1.0*samples[id].received/(samples[id].last_counter - samples[id].first_counter + 1));
	if (tmpVal < 0.1)
		classCounter[0] += 1;
	else if (tmpVal < 0.2)
		classCounter[1] += 1;
	else if (tmpVal < 0.3)
		classCounter[2] += 1;
	else if (tmpVal < 0.4)
		classCounter[3] += 1;
	else if (tmpVal < 0.5)
		classCounter[4] += 1;
	else if (tmpVal < 0.6)
		classCounter[5] += 1;
	else if (tmpVal < 0.7)
		classCounter[6] += 1;
	else if (tmpVal < 0.8)
		classCounter[7] += 1;
	else if (tmpVal < 0.9)
		classCounter[8] += 1;
	else if (tmpVal <= 1.0)
		classCounter[9] += 1;

#ifdef VALIDATION
	int feature_index = 0;
	double target_label = (1.0*samples[id].received/(samples[id].last_counter - samples[id].first_counter + 1));
	if (useDistance) {
		testfeatures[feature_index].index = feature_index+1;
		testfeatures[feature_index++].value = dist;
	}
	if (useRSSI) {
		testfeatures[feature_index].index = feature_index+1;
		testfeatures[feature_index++].value = rssi;
	}
	testfeatures[feature_index].index = feature_index+1;
	testfeatures[feature_index++].value = currentRate;
	testfeatures[feature_index].index = feature_index+1;
	testfeatures[feature_index++].value = rate;
#endif
#endif

	// Write receiver status
	for(unsigned int i = 0; i < MAX_NODE_ID; i++)
	{
		if( i == id)
			continue;
		if(is_set(samples[id].receiver_status[i/32],i%32)){
			rssi = (Neighbors[i].sum_rssi / Neighbors[i].rssi_samples);	// not used...
			rate = Neighbors[i].rate;
			Vector position = nodes.Get(i)->GetObject<MobilityModel>()->GetPosition();
			dist = CalculateDistance(position ,myPosition);
			// continue writing the raw sample
			fprintf(fp," %i %.1f %hhu",i,dist,rate);
#ifdef SAVE_SAMPLES_GEO
			// continue writing sample's coordinates
			fprintf(fp_samples_geo," %i(%.1f,%.1f)",i,position.x,position.y);
#endif
#ifdef SAVE_PROCESSED_SAMPLES
			// prepare values for the processed sample
			Vr++;
			if (dist <= INTERVAL_LIMIT_1)
				index = 0;
			else
				index = 1;
			Fr[rate-1][index] += 1;
#endif
		}
	}

#ifdef SAVE_PROCESSED_SAMPLES
	// continue writing the processed sample
	fprintf(fp_processed,"5:%hhu ",Vr);
#ifdef VALIDATION
	testfeatures[feature_index].index = feature_index+1;
	testfeatures[feature_index++].value = Vr;
#endif

	j=6;
	for (unsigned int i=0; i < 3; i++){
		for (unsigned int k=0; k < NUMBER_OF_DIST_INTERVALS; k++) {
			fprintf(fp_processed,"%d:%hhu ",j++,Fr[i][k]);
#ifdef VALIDATION
			testfeatures[feature_index].index = feature_index+1;
			testfeatures[feature_index++].value = Fr[i][k];
#endif
		}
	}
#endif

	// continue writing the raw sample
	fprintf(fp," ] sender_status [");

#ifdef SAVE_SAMPLES_GEO
	// continue writing the sample's coordinates
	fprintf(fp_samples_geo," ] SENDER [");
#endif

	// Write senders status
	j=0;
	for(unsigned int i = 0; i < MAX_NODE_ID; i++)
	{
		if(is_set(samples[id].sender_status[i/32],i%32)){
			rssi = samples[id].sender_rssi_data[j];	// not used...
			rate = samples[id].sender_rate_data[j];
			Vector position = nodes.Get(i)->GetObject<MobilityModel>()->GetPosition();
			dist = CalculateDistance( position ,senderPosition);
			j+=1;
			// continue writing the raw sample
			fprintf(fp," %u %.1f %hhu",i,dist,rate);
#ifdef SAVE_SAMPLES_GEO
			// continue writing the sample's coordinates
			fprintf(fp_samples_geo," %i(%.1f,%.1f)",i,position.x,position.y);
#endif
#ifdef SAVE_PROCESSED_SAMPLES
			// prepare values for the processed sample
			Vs++;
			if (dist <= INTERVAL_LIMIT_1)
				index = 0;
			else
				index = 1;
			Fs[rate-1][index] += 1;
#endif
		}
	}

	// finish writing the raw sample
	fprintf(fp," ]");
	fprintf(fp," rec_rate %d",currentRate);
	fprintf(fp,"\n");

#ifdef SAVE_PROCESSED_SAMPLES
	// finish writing the processed sample
	fprintf(fp_processed,"12:%hhu ",Vs);
#ifdef VALIDATION
	testfeatures[feature_index].index = feature_index+1;
	testfeatures[feature_index++].value = Vs;
#endif
	j=13;
	for (unsigned int i=0; i < 3; i++) {
		for (unsigned int k=0; k < NUMBER_OF_DIST_INTERVALS; k++) {
			fprintf(fp_processed,"%d:%hhu ",j++,Fs[i][k]);
#ifdef VALIDATION
			testfeatures[feature_index].index = feature_index+1;
			testfeatures[feature_index++].value = Fs[i][k];
#endif
		}
	}
	fprintf(fp_processed,"19:%f",snr);
#ifdef VALIDATION
	if (useSNR) {
		testfeatures[feature_index].index = feature_index+1;
		testfeatures[feature_index++].value = snr;
	}
	// ...and print the predicted value at the end of the line...
	testfeatures[feature_index].index = -1; // char to terminate feature vector
	// scaling limits
	double lower = -1;
	double upper = +1;
	if (modelName == "model_new") {
		lower = 3;
		upper = -3;
		// scale data
		for (int i=0; i<nooffeatures; i++)
		{
			testfeatures[i].value = lower + (upper-lower) * (testfeatures[i].value-featuremin[i]) / (featuremax[i]-featuremin[i]);
			if (testfeatures[i].value < upper)
				testfeatures[i].value = upper;
			else if (testfeatures[i].value > lower)
				testfeatures[i].value = lower;
		}
	} else  {
		lower = -1;
		upper = +1;
		// scale data
		for (int i=0; i<nooffeatures; i++)
		{
			testfeatures[i].value = lower + (upper-lower) * (testfeatures[i].value-featuremin[i]) / (featuremax[i]-featuremin[i]);
			if (testfeatures[i].value < lower)
				testfeatures[i].value = lower;
			else if (testfeatures[i].value > upper)
				testfeatures[i].value = upper;
		}
	}

	double predict_label = svm_predict(model,testfeatures);

	if (predict_label > 1.0)
		predict_label = 1.0;
	else if (predict_label < 0.0)
		predict_label=0.0;

	double error = abs(target_label - predict_label);
	errSum += (error*error);
	errCounter++;
	fprintf(fp_processed," predict:%f err:%f cumMSE: %f\n",predict_label,error,errSum/errCounter);
#else
	fprintf(fp_processed,"\n");
#endif
#endif

#ifdef SAVE_SAMPLES_GEO
	// finish writing the sample's coordinates
	fprintf(fp_samples_geo," ]\n");
#endif

	// write subsamples
	for (unsigned int i = 0; i < samples[id].subsamples.size(); i++) {
		// Here, define the maximum number of gatheres subsamples (corresponding to the considered sessions length in correctly received packets)
		if (i > 25)
			break;

		subsample_t subsample = samples[id].subsamples[i];

		//fprintf(fp,"\t[%f]d=%f\trssi=%f\tsinr=%f\tsingleRSSI=%f\tsingleSINR=%f\tmwRSSI=%f\tmwSINR=%f\tewmaRSSI=%f\tewmaSNIR=%f\tewmaPRR=%f\twmewmaPRR=%f\tMAPPLE=%f\treal=%f\n",subsample.timestamp,subsample.distance,subsample.rrsi, subsample.sinr, subsample.singleRSSI_estimator,subsample.singleSNIR_estimator,subsample.mwRSSI_estimator,subsample.mwSNIR_estimator,subsample.ewmaRSSI_estimator,subsample.ewmaSNIR_estimator,subsample.ewmaPRR_estimator,subsample.wmewmaPRR_estimator,predict_label,target_label);
		fprintf(fp_samples_geo,"\t[SAMPLE_%u]\tsingleSINR=%f\tmwRSSI=%f\tmwSINR=%f\tewmaRSSI=%f\tewmaSNIR=%f\tewmaPRR=%f\twmewmaPRR=%f\tMAPPLE=%f\treal=%f\n",sampleID,subsample.singleSNIR_estimator,subsample.mwRSSI_estimator,subsample.mwSNIR_estimator,subsample.ewmaRSSI_estimator,subsample.ewmaSNIR_estimator,subsample.ewmaPRR_estimator,subsample.wmewmaPRR_estimator,predict_label,target_label);

	}

	// STOP THE SIMULATION AFTER GATHERING 10,000 samples:
	if (sampleID >= 10000) {
		for (unsigned int i=0; i < 10; i++) {
			printf("class %u = %u\n",(i+1),classCounter[i]);
		}
		Simulator::Stop ();
	}
}


/*
 * Closing a sample
 * We consider it valid, therefore print it
 * only if we have collected stats from
 * at least MIN_SAMPLE_RECEIVED packets
*/
void MappleApplication::closeSample(uint8_t id)
{
	// Save samle, if valid
	if( samples[id].received >= MIN_SAMPLE_RECEIVED)
	{
		print_sample(id);
	}

	// Clear sample data (for a new sample)
	memset(samples[id].sender_status,0x0,sizeof(uint32_t)*5);
	for(unsigned int i=0;i < 5; i++){
		samples[id].sender_status[i] = 0;
		samples[id].receiver_status[i] = 0;
	}
	samples[id].received = 0;
	samples[id].avgTravelTime = 0;
	samples[id].queueSize = 0;
	samples[id].avgSendTime = 0;
	samples[id].avgSendTimeWithoutBackoff = 0;
	samples[id].queueSize = 0;

	samples[id].subsamples.clear();
}

/*
 * Just call close for all
 */
void MappleApplication::closeAllSamples()
{
	for(unsigned int i=0;i < MAX_NODE_ID; i++)
	{
		if( samples[i].received > 0)
		{
			closeSample(i);
		}
	}
}

/*
 * MoveNodeStrategy2 implementation:
 * 		We start with d = MIN_DISTANCE, and try MAX_TRIALS to get
 * 		a random location (by choosing angle and distance <= d)
 * 		to apply the chosen policy
 * 		After MAX_TRIALS, increment d by STEP_FACTOR
 *		if we reach d = MAX_DISTANCE and still nothing, just return last random
 */
Vector MappleApplication::getNextDestination() {
	int criteria = 0;

	Vector myPosition = m_mobility->GetPosition();

	// Choose a random criteria
    if (uniformVariable.GetValue() < P_GO_MORE )
    	criteria = 2;
    else
    	criteria = 1;

	// Number of neighbors in current location
    int neighbors = numberNeighbors(myPosition, ASSUMED_COMMUNICATION_RANGE);


    int trials = MAX_TRIALS;


    double d = 0.0;
    while (true) {
		// Distance to move
        double rd = (MIN_DISTANCE + d)*uniformVariable.GetValue();

        // Angle in radians
        double ra = MAPPLE_PI_VALUE * uniformVariable.GetValue(0,360.0) / 180.0;

		// NEW POSITION
        Vector newPosition(myPosition.x + rd*cos(ra) , myPosition.y + rd*sin(ra) , myPosition.z);

		// If position is valid
        if ( (newPosition.x >= 0) && (newPosition.x <= AREA_MAX_X) &&
        	 (newPosition.y >=0) && (newPosition.y <= AREA_MAX_Y) ) {
			// Number of neighbors in new position
            int newNeighbors = numberNeighbors(newPosition, ASSUMED_COMMUNICATION_RANGE);
				if (criteria == 1) { // go less
					// If got less neighbors, OK
                    if (newNeighbors < neighbors) {
                    	//cout << GetNode()->GetId() << ": position=" << myPosition << "(" <<  neighbors << " neighbors) , new_position=" << newPosition << "(" << newNeighbors << ") GO_LESS " << endl;
                        return newPosition;
                    }
				} else if (criteria == 2) {	// go more
					// If got more neighbors, ok
                    if (neighbors < newNeighbors) {
                    	//cout << GetNode()->GetId() << ": position=" << myPosition << "(" <<  neighbors << " neighbors) , new_position=" << newPosition << "(" << newNeighbors << ") GO_MORE " << endl;
                        return newPosition;
                    }
				}
        }

        trials -= 1;

		// If we ran out of trials, reset and try with longer distance
        if (trials == 0) {
        	trials = MAX_TRIALS;
            d += STEP_FACTOR;
        }

		// We couldn't do it, return last random location (limited to valid area)
        if (MIN_DISTANCE + d > MAX_DISTANCE) {
        	if (newPosition.x < 0)
        		newPosition.x = 0;
        	if (newPosition.x > AREA_MAX_X)
        		newPosition.x = AREA_MAX_X;
        	if (newPosition.y < 0)
        		newPosition.y = 0;
        	if (newPosition.y > AREA_MAX_Y)
        	    newPosition.y = AREA_MAX_Y;
        	if (CalculateDistance(newPosition, myPosition) < 0.000001) {
        		printf("small distance!!!\n");
        		// move a liittle bit any way...
        		newPosition.x += 1;
        		if (newPosition.x >= AREA_MAX_X)
        			newPosition.x -= 2;
        		newPosition.y += 1;
        		if (newPosition.y >= AREA_MAX_Y)
        			newPosition.y -= 2;
        	}
        	return newPosition;
        }
    }

    // Shouldn't get here
    return Vector();
}

/*
 * how many neighbors are we going to have in the given position, assuming the given range of transmission?
 */
int MappleApplication::numberNeighbors(Vector position, double range) {
	unsigned int myId = GetNode()->GetId();
	int numberN = 0;
	for (unsigned int i = 0; i < nodes.GetN(); i++) {
		if (i == myId)
			continue;
		if (CalculateDistance( nodes.Get(i)->GetObject<MobilityModel>()->GetPosition(),position) <= range)
			numberN++;
	}
	return numberN;
}


} // Namespace ns3

