/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2009 IITP RAS
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * This is an example script for AODV manet routing protocol.
 *
 * Authors: Pavel Boyko <boyko@iitp.ru>
 */

#include "ns3/aodv-module.h"
#include "ns3/dsdv-helper.h"
#include "ns3/olsr-helper.h"
#include "ns3/pfr-helper.h"
#include "ns3/core-module.h"
#include "ns3/network-module.h"
#include "ns3/internet-module.h"
#include "ns3/mobility-module.h"
#include "ns3/ns2-mobility-helper.h"
#include "ns3/point-to-point-module.h"
#include "ns3/wifi-module.h"
#include "ns3/v4ping-helper.h"
#include "ns3/on-off-helper.h"
#include "ns3/onoff-application.h"
#include "ns3/packet-sink-helper.h"
#include "ns3/data-rate.h"
#include <iostream>
#include <cmath>
#include <string>

#include <stdlib.h>
#include <stdio.h>
#include <sys/times.h>
#include <unistd.h>

// To enable visualizations...
#include "ns3/pyviz.h"

// for gettimeofday()
#include <sys/time.h>

#include "CpuTime.h"

using namespace ns3;

/**
 * \brief Test script.
 *
 * This script creates 1-dimensional grid topology and then ping last node from the first one:
 *
 * [10.0.0.1] <-- step --> [10.0.0.2] <-- step --> [10.0.0.3] <-- step --> [10.0.04]
 *
 * ping 10.0.0.4
 */
class AodvExample
{
public:
  AodvExample ();
  /// Configure script parameters, \return true on successful configuration
  bool Configure (int argc, char **argv);
  /// Run simulation
  void Run ();
  /// Report results
  void Report (std::ostream & os);

  std::string outFilePath;

private:
  ///\name parameters
  //\{
  /// Number of nodes
  uint32_t size;
  /// Distance between nodes, meters
  double step;
  /// Simulation time, seconds
  double totalTime;
  /// Write per-device PCAP traces if true
  bool pcap;
  /// Print routes if true
  bool printRoutes;
  //\}

  int seed;

  uint32_t numGenerators;



  ///\name network
  //\{
  NodeContainer nodes;
  NetDeviceContainer devices;
  Ipv4InterfaceContainer interfaces;
  //\}

private:
  void CreateNodes ();
  void CreateDevices ();
  void InstallInternetStack ();
  void InstallApplications ();
};

void CourseChange (std::string context, Ptr<const MobilityModel> model)
{
  Vector position = model->GetPosition ();
  position.x = position.x;
  NS_LOG_UNCOND (context << "[" << Simulator::Now().GetSeconds() << "] " <<
    " x = " << position.x << ", y = " << position.y << ", z = " << position.z);
}

int main (int argc, char **argv)
{

  // To enable visualizations ()...
  { PyViz v; }

  AodvExample test;
  if (!test.Configure (argc, argv))
    NS_FATAL_ERROR ("Configuration failed. Aborted.");

  test.Run ();
  test.Report (std::cout);
  return 0;
}

//-----------------------------------------------------------------------------
AodvExample::AodvExample () :
  size (2),
  step (100),
  totalTime (100),
  pcap (true),
  printRoutes (true)
{
}

bool
AodvExample::Configure (int argc, char **argv)
{
  // Enable AODV logs by default. Comment this if too noisy
  //LogComponentEnable("AodvRoutingProtocol", LogLevel(LOG_LEVEL_LOGIC | LOG_PREFIX_TIME));

  seed = 12345;
  //seed = time(NULL);	// random seed
  SeedManager::SetSeed (seed);
  CommandLine cmd;

  cmd.AddValue ("pcap", "Write PCAP traces.", pcap);
  cmd.AddValue ("printRoutes", "Print routing table dumps.", printRoutes);
  cmd.AddValue ("size", "Number of nodes.", size);
  cmd.AddValue ("time", "Simulation time, s.", totalTime);
  cmd.AddValue ("step", "Grid step, m", step);
  cmd.AddValue ("out", "Output file path",outFilePath);
  cmd.AddValue ("num_gen", "Number of generators",numGenerators);


  cmd.Parse (argc, argv);
  return true;
}

void
AodvExample::Run ()
{
//  Config::SetDefault ("ns3::WifiRemoteStationManager::RtsCtsThreshold", UintegerValue (1)); // enable rts cts all the time.
  CreateNodes ();
  CreateDevices ();
  InstallInternetStack ();
  InstallApplications ();

//  std::ostringstream oss;
//  oss <<
//    "/NodeList/" << nodes.Get (size-1)->GetId () <<
//    "/$ns3::MobilityModel/CourseChange";
////  oss <<
////    "/NodeList/*/$ns3::MobilityModel/CourseChange";
//
//  Config::Connect (oss.str (), MakeCallback (&CourseChange));

  std::cout << "Starting simulation for " << totalTime << " s ...\n";

  LogComponentEnable("PacketSink", LogLevel(LOG_LEVEL_INFO | LOG_PREFIX_FUNC | LOG_PREFIX_TIME));
  LogComponentEnable("OnOffApplication", LogLevel(LOG_LEVEL_INFO | LOG_PREFIX_FUNC | LOG_PREFIX_TIME));
  Simulator::Stop (Seconds (totalTime));

  // Time statistics
  FILE* fp = fopen(outFilePath.c_str(),"a");
  CpuTime *ct = new CpuTime();
  printf("Time - go!\n");

  ct->start();
  Simulator::Run ();
  ct->end();

  printf("Time - stop!\n");

  // save time statistics etc.
  fprintf(fp,"%d\t%f\t%f\t%f\t%f\t%d\n",size,ct->cpu_time_elapsed(),ct->user_time_elapsed(),ct->system_time_elapsed(),ct->total_time_elapsed(),seed);
  fclose(fp);

  printf("Durata della simulazione (CPU): %g secondi\n", ct->cpu_time_elapsed());
  printf("Durata della simulazione (total): %g secondi\n", ct->total_time_elapsed());

  Simulator::Destroy ();

}

void
AodvExample::Report (std::ostream &)
{
}

void
AodvExample::CreateNodes ()
{
  std::cout << "Creating " << (unsigned)size << " nodes " << step << " m apart.\n";
  nodes.Create (size);
  // Name nodes
  for (uint32_t i = 0; i < size; ++i)
    {
      std::ostringstream os;
      os << "node-" << i;
      Names::Add (os.str (), nodes.Get (i));
    }

//  // Create static grid
  MobilityHelper mobility;
  mobility.SetPositionAllocator ("ns3::GridPositionAllocator",
                                 "MinX", DoubleValue (50.0),
                                 "MinY", DoubleValue (50.0),
                                 "DeltaX", DoubleValue (step),
                                 "DeltaY", DoubleValue (step),
                                 "GridWidth", UintegerValue (3),
                                 "LayoutType", StringValue ("RowFirst"));
  //mobility.SetMobilityModel ("ns3::ConstantPositionMobilityModel");

  // Using RandomWaypoint mobility model
  ObjectFactory pos;
  pos.SetTypeId ("ns3::RandomRectanglePositionAllocator");
  pos.Set ("X", RandomVariableValue (UniformVariable (0.0, 350)));
  pos.Set ("Y", RandomVariableValue (UniformVariable (0.0, 350)));
  Ptr<PositionAllocator> ptrPositionAlloc = pos.Create ()->GetObject<PositionAllocator> ();
  mobility.SetMobilityModel ("ns3::RandomWaypointMobilityModel",
							 "Speed",RandomVariableValue(ConstantVariable(10.0)),
							 "Pause",RandomVariableValue(ConstantVariable(10.0)),
							 "PositionAllocator",PointerValue(ptrPositionAlloc));
  mobility.Install (nodes);

//  // import mobility from ns2 scenario file
//  std::string fileScn("/ns3/ns-allinone-3.12.1/ns-3.12.1/scratch/ARGOS_movements.scn");
//  Ns2MobilityHelper mobility = Ns2MobilityHelper::Ns2MobilityHelper(fileScn);
//  mobility.Install (nodes.Begin(),nodes.End());

}

void
AodvExample::CreateDevices ()
{
  NqosWifiMacHelper wifiMac = NqosWifiMacHelper::Default ();
  wifiMac.SetType ("ns3::AdhocWifiMac");
  YansWifiPhyHelper wifiPhy = YansWifiPhyHelper::Default ();
  //YansWifiChannelHelper wifiChannel = YansWifiChannelHelper::Default ();
  YansWifiChannelHelper wifiChannel;
  wifiChannel.SetPropagationDelay ("ns3::ConstantSpeedPropagationDelayModel");
  wifiChannel.AddPropagationLoss ("ns3::TwoRayGroundPropagationLossModel",
								  "HeightAboveZ",DoubleValue(0.3));
  wifiPhy.SetChannel (wifiChannel.Create ());
  WifiHelper wifi = WifiHelper::Default ();
  wifi.SetRemoteStationManager ("ns3::ConstantRateWifiManager", "DataMode", StringValue ("OfdmRate6Mbps"), "RtsCtsThreshold", UintegerValue (0));
  devices = wifi.Install (wifiPhy, wifiMac, nodes);

//  if (pcap)
//    {
//      wifiPhy.EnablePcapAll (std::string ("aodv"));
//    }
}

void
AodvExample::InstallInternetStack ()
{
  AodvHelper aodv;
  DsdvHelper dsdv;
  OlsrHelper olsr;
   PfrHelper pfr;
   pfr.Set("BroadcastProbability", DoubleValue(1.0));
  // you can configure AODV attributes here using aodv.Set(name, value)
  InternetStackHelper stack;
  stack.SetRoutingHelper (aodv); // has effect on the next Install ()
  stack.Install (nodes);
  Ipv4AddressHelper address;
  address.SetBase ("10.0.0.0", "255.0.0.0");
  interfaces = address.Assign (devices);

//  if (printRoutes)
//    {
//      Ptr<OutputStreamWrapper> routingStream = Create<OutputStreamWrapper> ("aodv.routes", std::ios::out);
//      aodv.PrintRoutingTableAllAt (Seconds (8), routingStream);
//    }
}

void
AodvExample::InstallApplications ()
{
//  V4PingHelper ping (interfaces.GetAddress (size - 1));
//  ping.SetAttribute ("Verbose", BooleanValue (true));
//
//  ApplicationContainer p = ping.Install (nodes.Get (0));
//  p.Start (Seconds (0));
//  p.Stop (Seconds (totalTime) - Seconds (0.001));


  int port = 9;
  for (uint32_t i = 0; i < numGenerators; i++) {
	  // Node i to node i+numGenerators
	  OnOffHelper onOffHelper1 ("ns3::UdpSocketFactory", InetSocketAddress(interfaces.GetAddress (i+numGenerators), port));
	  onOffHelper1.SetAttribute("DataRate", DataRateValue(DataRate(8)));
	  onOffHelper1.SetAttribute("PacketSize", UintegerValue(1));
	  onOffHelper1.SetAttribute("OnTime", RandomVariableValue(ConstantVariable(10000)));
	  onOffHelper1.SetAttribute("OffTime",  RandomVariableValue(ConstantVariable(0)));
	  ApplicationContainer app1 = onOffHelper1.Install(nodes.Get (i));
	  app1.Start (Seconds (0.01 * UniformVariable ().GetInteger (0, 100)));

	  PacketSinkHelper sinkHelper1("ns3::UdpSocketFactory", InetSocketAddress(Ipv4Address::GetAny (), port));
	  ApplicationContainer app11 = sinkHelper1.Install(nodes.Get(i+numGenerators));
	  app11.Start(Seconds (0));
  }

  // Show nodes initial positions
  Ptr<Node> node = nodes.Get (0);
  Ptr<MobilityModel> mob = node->GetObject<MobilityModel> ();
  for(uint32_t i = 0; i < size; i++) {
	  node = nodes.Get (i);
	  mob = node->GetObject<MobilityModel> ();
	  std::ostringstream oss;
	  oss << "Node" << i << " initial position:\t";
	  Simulator::Schedule (Seconds (0.001), &CourseChange, oss.str() , mob);
  }

  // move one node away
  //node = nodes.Get (size/2);
  //mob = node->GetObject<MobilityModel> ();
  //Simulator::Schedule (Seconds (totalTime/3), &MobilityModel::SetPosition, mob, Vector (122, 0, 0));
}

void CpuTime::start()
{
  gettimeofday(&total_time_start, NULL);
  times(&cpu_time_start);
}

void CpuTime::end()
{
  times(&cpu_time_end);

  gettimeofday(&total_time_end, NULL);

  cpu_time = (cpu_time_end.tms_utime - cpu_time_start.tms_utime) +
    (cpu_time_end.tms_stime - cpu_time_start.tms_stime);
  cpu_time /= (double)sysconf(_SC_CLK_TCK);

  user_time = (cpu_time_end.tms_utime - cpu_time_start.tms_utime) / ((double)sysconf(_SC_CLK_TCK));
  system_time = (cpu_time_end.tms_stime - cpu_time_start.tms_stime) / ((double)sysconf(_SC_CLK_TCK));

  total_time = (total_time_end.tv_sec - total_time_start.tv_sec) +
    ((double)(total_time_end.tv_usec - total_time_start.tv_usec) / 1000000.0);

  /*
  printf("Cpu time elapsed: %f - Total time elapsed %f\n",
	 cpu_time, total_time);
  fflush(stdout);
  */
}

void CpuTime::end(FILE *log_fp)
{
  times(&cpu_time_end);

  gettimeofday(&total_time_end, NULL);

  cpu_time = (cpu_time_end.tms_utime - cpu_time_start.tms_utime) +
    (cpu_time_end.tms_stime - cpu_time_start.tms_stime);
  cpu_time /= (double)sysconf(_SC_CLK_TCK);

  total_time = (total_time_end.tv_sec - total_time_start.tv_sec) +
    ((double)(total_time_end.tv_usec - total_time_start.tv_usec) / 1000000.0);

  /*
  printf("Cpu time elapsed: %f - Total time elapsed %f\n",
	 cpu_time, total_time);
  fflush(stdout);
  */
  fprintf(log_fp, "Cpu time elapsed: %f - Total time elapsed %f\n",
	  cpu_time, total_time);
  fflush(log_fp);
}

void CpuTime::end(FILE *log_fp, char *msg)
{
  times(&cpu_time_end);

  gettimeofday(&total_time_end, NULL);

  cpu_time = (cpu_time_end.tms_utime - cpu_time_start.tms_utime) +
    (cpu_time_end.tms_stime - cpu_time_start.tms_stime);
  cpu_time /= (double)sysconf(_SC_CLK_TCK);

  total_time = (total_time_end.tv_sec - total_time_start.tv_sec) +
    ((double)(total_time_end.tv_usec - total_time_start.tv_usec) / 1000000.0);

  fprintf(log_fp, "%s: Cpu time: %g - Total time: %g\n",
	  msg, cpu_time, total_time);
}

double CpuTime::cpu_time_elapsed()
{
//  times(&cpu_time_end);
//
//  cpu_time = (cpu_time_end.tms_utime - cpu_time_start.tms_utime) +
//    (cpu_time_end.tms_stime - cpu_time_start.tms_stime);
//  cpu_time /= (double)sysconf(_SC_CLK_TCK);

  return cpu_time;
}

double CpuTime::user_time_elapsed() {
	return user_time;
}

double CpuTime::system_time_elapsed() {
	return system_time;
}

double CpuTime::total_time_elapsed()
{
//  gettimeofday(&total_time_end, NULL);
//
//  total_time = (total_time_end.tv_sec - total_time_start.tv_sec) +
//    ((double)(total_time_end.tv_usec - total_time_start.tv_usec) / 1000000.0);

  return total_time;
}

